
import {CSS3DObject } from './3rdparty/threejs/examples/jsm/renderers/CSS3DRenderer.js';
import { color } from './other/dat.gui.module.js';

/**
 * LOADED AS AN IMPORT INTO THE ACE.
 */
var C = {};

C.CSS3DObject = CSS3DObject;

C.CSS3_NEURON_LBL = 0;
C.CSS3_RELON_LBL = 1;

C._targetMeshes = [];
C._imgs = [];
C._circuitsMap = [];   //map 
C._aCircuits = [];  //strd array


C.placements = [];
C.unit = 1200;
C.targetSize = 60;
C._maxTgenSpokes = 165; //max 5 rotations of 33

//3rd base, 2nd base, 1st base, home plate for 4 possible points.
C.placements = [{x:C.unit, y:0, z:0}, {x:-C.unit, y:0, z:0}, {x:0, y:0, z:0}, {x:0, y:0, z:C.unit}, {x:0, y:0, z:-C.unit}];
C.ta = null;

C.cam_dist = 10400;
C.cam_y = 1200;

C.css3dLbls = [];
C.clickableMeshes = [];
C.chaseableClassTnids = [];
C.chaseableClassNids = [];
C.tnid_reln_skips = [];//these relns will not be shown or used in the circuit
C.alt_ids = [];

C.YELLOW = {color:0xfefe17};
C.RED = {color:0xff0000};
C.BLUE = {color:0x55aaff};
C.ORANGE = {color:0xff7300}
C.PURPLE = {color:0xab07ff};
C.PINK = {color: 0xfe17df};
C.GREEN = {color:0x00FF00};
C.WHITEBLUE = {color:0xF5FBFF}; //used to denote new ghosts not yet chased.
C.BLACK = {color:0x171717};

C.colors = [C.YELLOW, C.RED, C.BLUE, C.ORANGE, C.PURPLE, C.PINK, C.GREEN, C.BLACK, C.WHITEBLUE];

C.ACE = null;   //ptr back

C.cam_points = [];
C.cam_points_way_out = [];

C.homeOutterZoom = .33;
C.homeInnerZoom = 3;
C.closeUpZoom = 5;
C.circuitTopViewZoom = 0.33;

C.materials = [];

//tracks new things in a multi run of external sources --- before we have ghosts to match
//to prevent dupe creations in the same run.
C.trackerMap = null;


C.hasCircuit = function() {
    if (C._aCircuits.length) return true;
    return false;
}


/**
 * 
 * @param {threeanne} $threeanne
 * @param {*} $unitSize 
 * @param {*} $targetSize 
 */
C.init=function($threeanne, $unitSize=1200, $targetSize=60, $chaseableClassTnids=[], $alt_ids=[]) {
    C.ta=$threeanne;
    C.ACE = C.ta.ACE;

    C.trackerMap = new tracker_map();
    
    C.unit = $unitSize;
    C.targetSize = $targetSize;
    C.chaseableClassTnids = $chaseableClassTnids;
    C.chaseableClassNids = MEMG.getNidsListFromTnidsList(C.chaseableClassTnids);
    C.tnid_reln_skips = [MEM.inst.temp_nid, MEM.class.temp_nid, 
                        MEM.gender.temp_nid, MEM.primaryNeuronImg.temp_nid, MEM.neuronImg.temp_nid,
                        MEM.firstName.temp_nid, MEM.middleName.temp_nid];

    if ($alt_ids.length) {
        C.alt_ids = $alt_ids;
    }

    //alt ids this app is interested in displaying for display neurons.
    //mostly relevant to persons, but we can add any unique ids for companies etc.
    else {
        C.alt_ids = [MEM.wikipart.temp_nid, MEM.bioguide_id.temp_nid, MEM.thomas_id.temp_nid, MEM.youtube_id.temp_nid, MEM.govtrack.temp_nid,
            MEM.icpsr.temp_nid, MEM.icpsr_prez.temp_nid, MEM.opensecrets.temp_nid, MEM.fec.temp_nid, MEM.votesmart.temp_nid, MEM.lis.temp_nid,
            MEM.ballotpedia.temp_nid, MEM.maplight.temp_nid, MEM.cspan.temp_nid, MEM.google_entity_id.temp_nid,
            MEM.house_history.temp_nid, MEM.wikidata.temp_nid, MEM.TMDB_id.temp_nid, MEM.facebook_id.temp_nid, 
            MEM.twitter_id.temp_nid, MEM.instagram_id.temp_nid, MEM.IMDB_id.temp_nid, MEM.TMDB_id.temp_nid];
    }
    
    //gen the main outter cam base points to view the circuits.
    var radius = C.cam_dist/2;
    var radiusOUT = C.cam_dist;

    var angleStep = (Math.PI * 2) / 8;
    for (var i=0; i < 8; i++) {
        var bumpi = i;//-.15;
        //outer point
        var y = C.ACE._defCamY;
        var x = (Math.cos(angleStep*bumpi) * radius);//+position.x;
        var z = (Math.sin(angleStep*bumpi) * radius);//+position.z;
        var v3 =  new C.ta.THREE.Vector3(x,y,z);
        C.cam_points.push(v3);

        var yo = C.ACE._defCamY;
        var xo = (Math.cos(angleStep*bumpi) * radiusOUT);//+position.x;
        var zo = (Math.sin(angleStep*bumpi) * radiusOUT);//+position.z;
        var v3o =  new C.ta.THREE.Vector3(xo,yo,zo);
        C.cam_points_way_out.push(v3o);
    }

    C.initMatsAndColorsObj();
}


/**
 * Each color ets a wireframe, solid, and solid alpha material
 * We will reuse use to optmize.
 */
C.initMatsAndColorsObj = function() {
    for (var c=0; c < C.colors.length; c++) {   
        C.colors[c].mat_wf = new C.ta.THREE.MeshBasicMaterial( { 
            wireframe:true,
            color:C.colors[c].color,
            envMap:null
        });

        C.colors[c].mat_solid = new C.ta.THREE.MeshStandardMaterial( { 
            roughness: 0, 
            metalness: 0,
            color:C.colors[c].color,
            blending:C.ta.THREE.NoBlending,
            envMap:null
        });

        //ANX.white_material_low_alpha = new THREE.MeshBasicMaterial({ color: 0xffffff, wireframe:true, transparent:true, opacity:.2 });

        C.colors[c].mat_solid_alpha = new C.ta.THREE.MeshStandardMaterial( { 
            transparent:true,
            opacity:.2,
            roughness: 0,
            color:C.colors[c].color,
            blending:C.ta.THREE.NoBlending,
            depthWrite:true,
            depthTest:true,
            side:C.ta.THREE.DoubleSide,
            envMap:null
        });

        C.colors[c].mat_solid_alpha2 = new C.ta.THREE.MeshStandardMaterial( { 
            transparent:true,
            opacity:.1,
            color:C.colors[c].color,
            blending:C.ta.THREE.NoBlending,
            depthWrite:true,
            depthTest:true,
            side:C.ta.THREE.DoubleSide,
            envMap:null
        });
    }
}

/**
 * 
 * @param {*} $nObj 
 * @param {*} $tgen 
 */
C.setColorMaterialWF = function($nObj, $tgen) {
    if ($tgen.mesh==null) return;
    var colorObj = C.getNeuronColor($nObj, $tgen);
    $tgen.mesh.material = colorObj.mat_wf;
    
    $tgen.mesh.colorObj = colorObj;  //for easy swapping
    
    colorObj.mat_wf.needsUpdate = true;
}
C.setColorMaterialAlpha = function($nObj, $tgen) {
    if ($tgen.mesh==null) return;
    var colorObj = C.getNeuronColor($nObj, $tgen);
    $tgen.mesh.material = colorObj.mat_solid_alpha;
    
    $tgen.mesh.colorObj = colorObj;  //for easy swapping
    
    colorObj.mat_solid_alpha.needsUpdate = true;
}
C.setColorMaterialSolid = function($nObj, $tgen) {
    if ($tgen.mesh==null) return;
    var colorObj = C.getNeuronColor($nObj, $tgen);
    $tgen.mesh.material = colorObj.mat_solid;

    $tgen.mesh.colorObj = colorObj;  //for easy swapping
    
    colorObj.mat_solid.needsUpdate = true;
}

C.setColorMaterialManual = function($mesh, $mat) {
    $mesh.material = $mat;
    $mat.needsUpdate = true;
}


/**
 * If the neuron is PENDING HANAI, we want to turn it pink.
 * We will only bother looking at pending hanai status for the active user's neurons.
 */
C.getNeuronColor=function($nObj, $tgen){  //this.bDUPEDOWNSTREAM
    if ($nObj.ghost==true) return C.BLUE;
    //else if (C.compCircINST!=null && C.compCircINST.match_tnids.indexOf($nObj.temp_nid)!=-1) return C.GREEN;
    else if ($tgen.bDUPEDOWNSTREAM==true) return C.BLACK; //this is a dupe of another tgen in this branch line already.
    else if ($nObj.bes.length && $nObj.bes[0]==MEM.ec_nonliving.temp_nid) {
        //not chasing classes in this app.
        return C.BLACK;
    }
    else if (ANNE.isANNE($nObj.owned_by)==true) return C.YELLOW; 
    else {
        if ($nObj.owned_by==ANNE.USER._userNObj.nid) {
            var pending_hanai = ANNE.HANAI.getPendingHanai($nObj);
            if (pending_hanai!=null) return C.PINK;
            return C.RED;
        }
        else return C.BLACK;
    }
}


/**
 * a color for if we are the sponsor of the neuron and/or relon?  hmm.
 * 
 * @param {*} $rel 
 * @returns 
 */
C.getRelonColor=function($rel){
    if ($rel.ghost==true) return C.BLUE.color;
    var fromNOBJ = MEMG.getNeuronByTempNid($rel.f);
    if (fromNOBJ==null) return C.BLACK.color;
    if ($rel.relon_sponsor==ANNE.USER._userNObj.nid || 
           fromNOBJ.owned_by==ANNE.USER._userNObj.nid) return C.RED.color;//C.ORANGE;  
    if ($rel.status_text!=undefined && $rel.status_text=="p" && $rel.relon_sponsor==ANNE.USER._userNObj.nid) {
        return C.PINK.color;
    }
    if (ANNE.isANNE(fromNOBJ.owned_by)==true) return C.YELLOW.color;
    return C.BLACK.color;
}



/**
 * Let animation run while we do other things.
 * @param {*} $nObj 
 * @returns 
 */
C.buildBasicTargetCircuit =function($nObj=null) {
    if (ANNE.UTILS.isset($nObj)==false) {
        C.ACE.processingDone();
        return;
    }

    var circ = null;
    if (C._circuitsMap[$nObj.nid]!=undefined) {
        return C._circuitsMap[$nObj.nid];
    }

    else {
        circ = new neuron_circuit($nObj);
        circ.c_index = C._aCircuits.length;
        circ.circ_id = circ.c_index;
        C._circuitsMap[$nObj.nid] = circ;
        C._aCircuits.push(circ);

        //var colorObj = C.getNeuronColor($nObj, circ.root_tgen);
        var geometry = new C.ta.THREE.BoxGeometry( C.targetSize, C.targetSize, C.targetSize, 10, 10, 10 );
        var mesh = new C.ta.THREE.Mesh( geometry, C.GREEN.mat_wf);
        C.ta.scene.add(mesh);

        mesh.box_size = C.targetSize;

        var plObj = C.placements[circ.c_index];
        mesh.position.x = plObj.x;
        mesh.position.z = plObj.z;
        mesh.position.y = 0;
        
        C._targetMeshes.push(mesh);
        C.clickableMeshes.push(mesh);
    
        circ.root_tgen.setMesh(mesh);
        circ.root_tgen.anne_chaseable=true;
    
        C.ACE.shutDownSearch();
        
        C.__doNeuronLbl(mesh, "anne-neuron", circ, circ.root_tgen, $nObj, true, "", (C.targetSize)+10);

        ANNE.playSound("swoosh");
        
        //the root location of the circuit.
        //C.ACE.controls.setOrbitPoint( plObj.x, plObj.y, plObj.z );
        C.camToCircuit(circ.c_index);
        
        C.ta.gsap.to( mesh.position, {
            ease:"sine-in",
            duration: .5,
            x:plObj.x,
            z:plObj.z
        } );
        return circ;
    }
}



/**
 * Turn back the line to its old color and size.
 * @param {*} $tgen 
 * @returns 
 */
C.deEmphasizePath=function($tgen) {
    var safety = 20;
    while ($tgen!=null && safety-->0) {
        for (var l=0; l < $tgen.in_bound_lines.length; l++) {
            var line = $tgen.in_bound_lines[l];
            if (line.relon!=null && line.relon.txid==$tgen.initiating_rel.txid) {
                if (line.old_color!=undefined) line.material.color.setHex(line.old_color);
                if (line.old_opacity!=undefined) line.material.opacity = line.old_opacity;
                line.material.linewidth=1;
            }
        }

        if ($tgen.parent_tgen==null || $tgen.gen==0) break;
        $tgen = $tgen.parent_tgen;
    }
}


/**
 * 
 * @param {string} $msg 
 */
C.addToConsole=function($msg) {
    C.ACE.addToConsole($msg);
}


/**
 * Get the relons of the alt IDs we are interested in.
 * @param {*} $outs 
 */
C.getIds = function($outs){
    return [];
}


/**
 * Do a better job at guess the query string.
 * We often have a long-name for an ANNE neuron and we may have a better guess
 * for using external sources if we have hit one of them and stored their short-hand nomenclatures.
 * We don't want to query wikispooks for Elon Reeve Musk (it will return as no good).
 * If we already hit DBPEDIA (or have the ref for it, wikipedia, ballotpedia) we can try those
 * instead.  They are more likely to be the same as what wikispooks uses.
 */
C.getFixerQueryStr = function($nObj, $rezPartNOBJ) {
    var x_relon = MEMG.getRelonFromWith($nObj.outs, TYPE_SYS, $rezPartNOBJ.nid);
    if (x_relon==null) {
        x_relon = MEMG.getRelonFromWith($nObj.outs, TYPE_HAS, MEM.ballotpedia.nid);
    }

    if (x_relon!=null) return x_relon.to_s.replaceAll(" ", "_");

    if ($nObj.bes[0]==MEM.person.temp_nid && $nObj.text.split(" ").length>2 
            && $nObj.aka!=undefined && $nObj.aka!=null && $nObj.aka.length) {
        return $nObj.aka[0].replaceAll(" ", "_");
    }
    return "";
}



C.chaseExternalData = async($tgen, $chaseNOBJ=null, $val="", $callbackdone=null, $parentTGEN=null)=> {
    var search_override_str = "";
    var test_str = $val;

    //we are only going to automatically (by default) chase wikipedia data and/or wikispooks data
    //if the flag is set in the gui config, if the class is of the right class
    //and if we haven't already done it begore.

    //If the class is the right class we can always reinvoke the wikipedia/wikispooks query with the 
    //launch urls in the neuron display.  (at anytime)
    //we want to avoid useless external queries for large chases or refreshes by limiting this to 
    //the right times.
    var bAUTOWP = true;
    var bAUTOWS = true;
    
    if ($chaseNOBJ!=null) {
        if (C.ACE.isWikispooksChaseable($chaseNOBJ)==true) {
            bAUTOWS = true;
        }
        
        if (C.ACE.isWikipediaChaseable($chaseNOBJ)==true) {
            var wp_relon = MEMG.getRelonFromWithR($chaseNOBJ.outs, TYPE_SYS, MEM.wikipart.temp_nid);
            if (wp_relon!=null) bAUTOWP=false;
        }
        else bAUTOWP=false;
    }
   
    //wikipedia first (wider net)
    if (bAUTOWP==true && ANNE.SETTINGS.bDOWIKIPEDIA==true) {

        //resolved known already (not naked text search)
        if ($chaseNOBJ!=null) {
            var fixer_str = C.getFixerQueryStr($chaseNOBJ, MEM.wikipart);
            if (fixer_str!="") {
                search_override_str = fixer_str;
                test_str = fixer_str;
            }
        }
        
        //no string search?
        if (test_str==null || test_str=="") {
            //if we have no known neuron.. we can really do much with this.
            if ($chaseNOBJ==null) {
                if ($callbackdone!=null && typeof $callbackdone=="function") {
                    return $callbackdone($chaseNOBJ);
                }
            }
            //set the test_str to the name of the neuron.
            else test_str = $chaseNOBJ.text;
        }
        
        //no optimization skip.. continue fwd.
        if (search_override_str!="") C.ACE.addToConsole("getting <b>wikipedia</b> new data for <b>"+search_override_str+"</b>");
        else C.ACE.addToConsole("getting <b>wikipedia</b> new data for <b>"+test_str+"</b>");
        
        //if we have the chaseNOBJ already, let's use it!
        if ($chaseNOBJ!=null && $tgen!=null) await C.forceChaseWIKIPEDIA($tgen);
        else {
            var wke = await ANNE.exploreWiki(test_str, C.trackerMap, null, C.ACE.addToConsole);
            if (wke && wke._wiki!=null && wke._wiki.resolvedNOBJ!=null) {
                var newNOBJ = wke._wiki.resolvedNOBJ;
                newNOBJ.bWIKIPEDIADONE=true;
                C.ACE.addToConsole("acquired data from wikipedia");
var bAUTOWS=false;
                if (bAUTOWS==true && ANNE.SETTINGS.bDOWIKISPOOKS==true) {
                    //pass val here to wikispooks, their custom relon may or may not be present.
                    return C.doWikispooks(newNOBJ, test_str, $callbackdone, tgen);
                }
                else if ($callbackdone!=null && typeof $callbackdone=="function") {
                    return $callbackdone(newNOBJ);
                }
            }
        }
    }

    //we have chases turned off, just use ANNNE + local data.
    else {
        if ($chaseNOBJ!=null) {
            C.ACE.addToConsole("ANNE neuron found for <b>"+test_str+"</b>.  Returning known data.");
            if ($callbackdone!=null && typeof $callbackdone=="function") {
                return $callbackdone($chaseNOBJ);
            }
        }
        else {
            C.ACE.addToConsole("no neuron found in memory for <b>"+test_str+"</b> external queries turned off.");
            if ($callbackdone!=null && typeof $callbackdone=="function") {
                return $callbackdone(null);
            }
        }
    }

    return;
}


/**
 * No need to await -- we want things to update in the GUI (if possible)
 * GUI will be blocked until we finish. 
 * @param {anne_neuron} $chaseNOBJ 
 * @param {*} $val 
 * @param {function} $callbackdone 
 * @param {neuron_gen} $parentTGEN
 */
C.doWikispooks = function($chaseNOBJ, $val, $callbackdone, $parentTGEN=null) {
    
    if(ANNE.MEM.isANNE(ANNE.USER._userNObj.nid)==false) {
        alert("coming soon");
        return;
    }

    var search_override_str = "";

    if ($chaseNOBJ!=null) {
        var fixer_str = C.getFixerQueryStr($chaseNOBJ, MEM.wikispooks_respart);
        if (fixer_str!=null && fixer_str!="") {
            search_override_str = fixer_str;
        }
    }

    //($CIRCUIT_MOD=null, $trackerMap = null, $altConsole=null, $resolveAmbigsCb=null)
    var wikispooksdata = new wikispooks(C, C.trackerMap, null, C.ACE.resolveAmbigs);
    wikispooksdata.setParentTGEN($parentTGEN);

    if (search_override_str!="") C.ACE.addToConsole("getting <b>wikispooks</b> new data for <b>"+search_override_str+"</b>");
    else C.ACE.addToConsole("getting <b>wikispooks</b> new data for <b>"+$val+"</b>");
    wikispooksdata.getData($chaseNOBJ, $val, search_override_str, async()=>{

        //lastBLock is just used as a flag here.
        //await ANNE.HISTORY_LOCAL.manuallyUpdatePullHistory(pObj, 1);

        ANNE.GHOSTDATA.ANNEDataFromExternalGeneralized(wikispooksdata, MEM.wikispooks_respart, ($nObj, $wikispooks)=> {
            if ($nObj!=null) {
                $nObj.bWIKISPOOKSDONE=true;
            }
            if ($callbackdone!=null && typeof $callbackdone=="function") {
                $callbackdone($nObj);
            }
        }, C.ACE.addToConsole);
    });
}



/**
 * A non-target subgen (gen > 0) neuron has been clicked in the gui.
 * 
 * @param {neuron_gen} $tgen 
 * @returns 
 */
C.toggleSubGen = async($tgen, $bANIMATETRANSITION=false)=> {
    C.ACE.setLastAction();
    var nObj = MEMG.getNeuronByTempNid($tgen.tnid);

    //alert("TEST");
    
    //do nothing.
    if ($tgen.bANIMATING==true) {
        ANNE.playSound("error_soft");
        return;
    }

    //first click of this neuron, set bOPEN if chaseable.
    if ($tgen.bOPEN==false && $tgen.bCHASED==false) { 
        C.closeAllGen($tgen, $tgen.circ, $tgen.gen);
        
        $tgen.bOPEN=true;
       
        //delayed creation of mesh... (until needed)
        //but we now default create tgens before they are opened on the screen.
        if ($tgen.mesh==null) {
            C.addSubGen($tgen, $tgen.parent_tgen.tgens.length, $tgen.initiating_rel); 
        }

        //first zoom out for the camera wide.
        //we want to do this when links are clicked on..
        //this helps regain user bearings
        if ($bANIMATETRANSITION==true) {
            C.zoomOutFromCircuit2($tgen.circ.c_index, $tgen.mesh);
        }

        if ($tgen.anne_chaseable==true) {
            C.ACE.processingStart();//only put processing up if we going to process.
        
            //chase ANNE for this neuron, then blow out one more gen for it.
            C.ACE.addToConsole("querying ANNODE for <b>"+nObj.text+"</b>.");
            
            //chase here, not in blowout of default tgens.
            //await CHASE.queryForOuts_not(nObj, true, [TYPE_AWARENESS]);
            await ANNE.CHASE.queryForOutsAndIns(nObj, true, [0,1,2,3,4,5,6,7,8,9]);

            //chase txids that are pending collab from this?
            ANNE.MEM_UDF.getAndyTxidsFrom(nObj.temp_nid);

            //chase external data.  When done come back here to finish things off with the 
            //final callback.

            //we're only chasing externals auto if this is a ghost neuron or if it has
            //less than 3 outs.  
            if(C.ACE.isValidExternalChase(nObj)==true && (nObj.ghost==true || nObj.outs.length<=2)) {
                var parent_tgen = $tgen.parent_tgen;
                if (parent_tgen==undefined) parent_tgen = null;
                C.chaseExternalData($tgen, nObj, "", async()=>{
                    //extend the circuit one gen beyond this one. 
                    //this gives the setHalos something to make halos with.  :)
                    await $tgen.buildCircuitLineOneMoreGen(C, $tgen.circ);
        
                    //start animation of halos
                    C.setHalos($tgen);
                    C.ACE.showNeuronDisplay(nObj, $tgen);
                    C.ACE.processingDone();
                    $tgen.bCHASED=true;
                }, parent_tgen);
            }

            else {
                //extend the circuit one gen beyond this one. 
                //this gives the setHalos something to make halos with.  :)
                await $tgen.buildCircuitLineOneMoreGen(C, $tgen.circ);
                
                //start animation of halos
                C.setHalos($tgen);
                C.ACE.showNeuronDisplay(nObj, $tgen);
                C.ACE.processingDone();
                $tgen.bCHASED=true;
            }
        }

        else {
            await CHASE.queryForOuts_not(nObj, true, [TYPE_AWARENESS]);
            $tgen.bCHASED=true;
            C.setHalos($tgen);
            C.ACE.showNeuronDisplay(nObj, $tgen);
            C.ACE.processingDone();
        }
    }

    //close the open neuron
    //zoom out to mid view
    else if ($tgen.bOPEN==true) {
        $tgen.bOPEN=false;
        C.removeHalos($tgen);
        C.ACE.hideNeuronDisplay();
    }


    //open the closed neuron
    else {
        C.closeAllGen($tgen, $tgen.circ, $tgen.gen);
        $tgen.bOPEN=true;

        await C.refreshHanais(nObj.outs);

        C.setHalos($tgen);
        C.zoomOutFromCircuit($tgen.circ.c_index);
        C.ACE.showNeuronDisplay(nObj, $tgen);
    }
}


/**
 * When we oepn a tgen we want to close other tgens at this level.
 * This helps us explore a graph more easily through any-given-path.
 * @param {*} $circ 
 * @param {*} $gen 
 */
C.closeAllGen = function($tgen, $circ, $gen) {

    //i want to fade out all neurons (with alpha mat)
    //i want to fade out all relons (with individual alpha. they all have their own mat)
    
    for (var t=0; t<$circ.tgens.length; t++) {
        if ($circ.tgens[t].bOPEN==true && $circ.tgens[t].gen==$gen) {
            $circ.tgens[t].bOPEN=false;
            C.removeHalos($circ.tgens[t]);
        }
    }

    C.fadeAllInPath($tgen);
}


/**
 * This turns from top to children on and continues to cascade down...
 * (ie, outs.. so this turns on any given "branch" of our circuit tree)
 * @param {*} $tgen 
 */
C.fadeAllInPath=function($tgen) {
    
    var parentTgen = $tgen.parent_tgen;
    
    while(parentTgen!=null) {
   
        for (var t=0; t < parentTgen.tgens.length; t++) {
            var tgen = parentTgen.tgens[t];
            if (tgen.tnid==$tgen.tnid) continue;
            if (tgen.mesh.colorObj==undefined) continue;
            C.setColorMaterialManual(tgen.mesh, tgen.mesh.colorObj.mat_wf);

            for (var l=0; l < tgen.in_bound_lines.length; l++) {
                var line = tgen.in_bound_lines[l];
                if (line.relon!=null && line.relon.txid==tgen.initiating_rel.txid) {
                    line.old_opacity=line.material.opacity;
                    line.material.opacity=.05;
                    line.material.needsUpdate=true;
                }
            }
            
        }
        
        parentTgen = parentTgen.parent_tgen;
        
        //fade box material of parent, and all children (other than $tgen)
        //fade all relons to .1 alpha
    }

}


/**
 * @param {neuron_gen} $tgen 
 * @returns 
 */
C.autoOpen = async($tgen)=> {
    var nObj = MEMG.getNeuronByTempNid($tgen.tnid);

    //first click of this neuron, set bOPEN if chaseable.
    if ($tgen.bOPEN==false) {
        $tgen.bOPEN=true;

        //delayed creation of mesh... (until needed)
        //but we now default create tgens before they are opened on the screen.
        if ($tgen.mesh==null) {
            C.addSubGen($tgen, $tgen.parent_tgen.tgens.length, $tgen.initiating_rel); 
        }
        
        //audit color of neuron
        if ($tgen.gen>0) C.setColorMaterialSolid(nObj, $tgen);
        C.setHalos($tgen, null);
    }

    else {
        $tgen.bOPEN=true;
        if ($tgen.gen>0) C.setColorMaterialSolid(nObj, $tgen);
    }
}



/**
 * turn off neuron gui
 */
C.turnOffNeuronGui=function($tgen){
    if ($tgen==undefined || $tgen==null) {
        console.warn("FIX THIS turnOffNueronGuid tgen=null");
        return;
    }
    
    if($tgen.primaryNeuronLbl==null 
       || $tgen.primaryNeuronLbl.element==undefined 
       || $tgen.primaryNeuronLbl.element==null) return;
    
    ANNE.A.addClass($tgen.primaryNeuronLbl.element, "faded-neuron");
}


C.removeCss3dLbl=function($lbl) {
    for (var cl=0; cl < C.css3dLbls.length; cl++) {
        if (C.css3dLbls[cl]==$lbl) {
            C.css3dLbls.splice(cl,1);
            return;
        }
    }
}



/**
 * stupid hack fix since I am doing something wrongly.
 */
C.turnOnNeuronGui=function($tgen) {
    if ($tgen.primaryNeuronLbl==null) return;
    if(($tgen.primaryNeuronLbl.element==undefined || $tgen.primaryNeuronLbl.element==null)) return;
    ANNE.A.show($tgen.primaryNeuronLbl.element);
    ANNE.A.removeClass($tgen.primaryNeuronLbl.element, "faded-neuron");
}




/**
 * This is a naked search using the ANNODE API.
 * This is not an NLP parse here.
 * 
 * It presumes we DO NOT have all the possible info so it first looks up the string.
 * After that is will then proceed to search local memory for the neuron.
 * If no neuron, it will make one.
 * 
 * It will then look at our external sources to extract maximum info.
 * Wikispooks, Wikileaks, TMDB
 * 
 * @param {*} $searchVal 
 * @param {*} $beClassNids 
 * @param {*} $lets 
 * @param {*} $bAPPLYWVTOBELINE 
 * @param {*} $inbeline 
 * @param {*} $bEXACT 
 * @returns 
 */
C.makeSearch=function($searchVal, $beClassNids=[], $lets=[], $bAPPLYWVTOBELINE=false, $bEXACT=true){
    var wvs_live = C.getWorldViewNidList();
    if (wvs_live==undefined) wvs_live = [];
    if (wvs_live.indexOf(USER._userNObj.nid)==-1) wvs_live.push(USER._userNObj.nid);
    if (wvs_live.indexOf(ANNE.MEM.ANNE_ACCTID)==-1) wvs_live.push(ANNE.MEM.ANNE_ACCTID);
    
    var bes_tnids = [];
    if ($beClassNids.length) {
        for (var b=0; b < $beClassNids.length; b++) {
            var be = $beClassNids[b];
            var beNOBJ = MEMG.getNeuronByNid(be);
            bes_tnids.push(beNOBJ.temp_nid);
        }
    }

    //we are lettign in other worldviews here (in addition to our own and ANNE)
    if ($lets.length) {
        for (var w=0; w < $lets.length; w++) {
            var wv = $lets[w];
            if (wvs_live.indexOf(wv)==-1) wvs_live.push(wv);
        }
    }

    var inbeline = null;
    if ($bAPPLYWVTOBELINE==true) {
        if ($beClassNids.length==1) {
            inbeline = $beClassNids[0]; //invoke the BEM search on this line.
        }
    }

    return { 
            text:$searchVal, 
            wvs:wvs_live, 
            bes:$beClassNids,
            bes_tnids:bes_tnids,
            exact:$bEXACT,
            applywvtobeline:$bAPPLYWVTOBELINE,
            inbeline:inbeline,
            chaseIt:true
    };
}


/**
 * 
 * @returns 
 */
C.getWorldViewNidList = function() {
    if (ANNE.USER._bWORLDVIEWBIPASS==false) return MEM.getWorldViewNidList();
    else return [];
}

/**
 * 
 * @param {*} $val 
 * @param {*} $beNObj 
 * @param {*} $bAPPLYTOBELINES 
 * @param {*} $be_nids_override - if set it will add to the beNObj 
 * @returns 
 */
C.searchToResolve=async($val, $beNObj, $bAPPLYTOBELINES=false, $be_nids_override=[])=> {
    if (C.trackerMap.hasClassVal($val, $beNObj.nid)==true) {
        return [ C.trackerMap.getClassVal($val, $beNObj.nid)];
    }

    var bes = $be_nids_override;
    if (bes.length==0) bes = [$beNObj.nid];
    else if (bes.indexOf($beNObj.nid)==-1) {
        bes.push($beNObj.nid);
    }
    
    var searchObj = C.makeSearch($val, bes, [], $bAPPLYTOBELINES);
    searchObj.thresh = 0.96;

    //this returns all matching relons from memory (not just the ANNODE search)
    var bDOPRE = true; //ok to end by finding match
    var aMatches = await CHASE.anne_global_search(searchObj, bDOPRE);
    console.log(aMatches);
    return aMatches;
}


/**
 */
C.manualLaunchTop = async($nObj)=> {
    await ANNE.API.queryForCircuitOutsAndIns($nObj.nid, true);
    C.ACE.addToConsole("manual launching neuron building basic target circuit");
    var circ = C.buildBasicTargetCircuit($nObj);
    
    var afile = await IMAGES.getAnyValidImgAFILE($nObj);
    if (afile!=null) {
        await ANNE.u3js.imagePlane(circ.root_tgen.primaryNeuronLbl, afile, 44, new C.ta.THREE.Vector3(0, 30, 0));
    }
    if ($nObj.ghost==false) {
        C.toggleTargetNeuron(circ.root_tgen);         
        return;
    }
    else {
        C.chaseExternalData(circ.root_tgen, $nObj, $nObj.text, async()=>{
            C.toggleTargetNeuron(circ.root_tgen);
        }, null);
    }
    return;
}



/**
 * Search can be throttled by BE class.
 * 
 * We can do other much more complciated searches if/when need be.
 * @param {*} $val 
 */
C.nakedSearch=async($val)=>{
    console.log("nakeSearch "+$val);
    C.ACE.processingStart("searching A.N.N.E. for <b>"+$val+"</b>");
    C.ACE.setLastAction();

    var bes = C.chaseableClassNids;
    var searchObj = C.makeSearch($val, bes);
    searchObj.thresh = 0.9;

    //this returns all matching relons from memory (not just the ANNODE search)
    var bDOPRE = true; //ok to end by finding match
    var aMatches = await CHASE.anne_global_search(searchObj, bDOPRE);
    console.log(aMatches);

    if (aMatches.length>1) {

        //get max info on these so we can let the user decide.
        //use function optimized for multi neurons
        if(aMatches.length>1) {
            await ANNE.CHASE.queryForOutsAndInsFroms(aMatches, true, [0,1,2,3,4,5,6,7,8,9]);
        }

        //use function for one neuron
        else {
            await ANNE.CHASE.queryForOutsAndIns(aMatches[0], true, [0,1,2,3,4,5,6,7,8,9]);
        }

        //we could do simple user input here into the console :)
        C.ACE.addToConsole("ambiguous matches, please differentiate.");
        C.ACE.disambigNakedChoice(aMatches, $val);
        return;
    }

    //we only have one, lets run with it.
    else if (aMatches.length==1) {
        var nObj = aMatches[0];
        
        await ANNE.API.queryForCircuitOutsAndIns(nObj.nid, true);

        C.ACE.addToConsole("found neuron building basic target circuit");

        var circ = C.buildBasicTargetCircuit(nObj);

        var afile = await IMAGES.getAnyValidImgAFILE(nObj);
        if (afile!=null) {
            await ANNE.u3js.imagePlane(circ.root_tgen.primaryNeuronLbl, afile, 44, new C.ta.THREE.Vector3(0, 30, 0));
        }
    
        //not chasing non-ghosts as a default.
        //only chase ghosts for external data as a default (if flags are on)
        if (nObj.ghost==false) {
            C.toggleTargetNeuron(circ.root_tgen);         
            return;
        }
        
        else {
            C.chaseExternalData(circ.root_tgen, nObj, $val, async()=>{
                //don't set this here, set it on first pass to indicate circuit built
                C.toggleTargetNeuron(circ.root_tgen);
            }, null);
        }
        return;
    }

    //we have no matches, we need to do external sources first.
    else {
        try{
            C.chaseExternalData(null, null, $val, 
            async($nObj)=>{
                //circ.root_tgen.bCHASED=true; //don't set this here, set it on first pass to indicate circuit built
                if ($nObj==null) {
                    C.ACE.processingDone();
                    return;
                }
                var circ = C.buildBasicTargetCircuit($nObj);
                if (circ==null) {
                    C.ACE.processingDone();
                    return;
                }
                C.toggleTargetNeuron(circ.root_tgen);
            }, null);
        }
        catch($e) {
            console.log($e);
        }
        
        return;
    }
}



/**
 * This disambig is for nake search disambig by the user.
 * @param {*} $nObj
 */
C.userDisambiguatedContinue = async($nObj, $val)=> {
    await ANNE.CHASE.queryForOutsFroms_not([$nObj.nid], true);
    C.ACE.addToConsole("userDisambiguatedContinue building basic target circuit");
    var circ = C.buildBasicTargetCircuit($nObj);
    //do wikispooks query.  CHECK FOR EXISTING WIKISPOOKS RELONS indicating we've already chased this.
    //third param is callback to run after all external props are processed..
    //and all new relons are conceptualized.
    C.chaseExternalData(circ.root_tgen, $nObj, $val, async()=>{
        //circ.root_tgen.bCHASED=true;  //don't set this here, set it on first pass to indicate circuit built
        C.toggleTargetNeuron(circ.root_tgen);
    }, null);
}



/**
 * Here we should pan/zoom into where the tgen
 * @param {*} $circ 
 * @param {*} $nid 
 * @returns 
 */
C.toggleByCircAndNid=function($circ, $nid, $bONLYTOGGLEIFCLOSED=false){
    C.ACE.setLastAction();
    var targetNOBJ = MEMG.getNeuronByNid($nid);
    if (targetNOBJ==null) return;
    ANNE.A.playSound("beep");
    for (var t=0; t < $circ.tgens.length; t++) {
        if ($circ.tgens[t].tnid==targetNOBJ.temp_nid) {
            if ($circ.tgens[t].bANIMATING==true) {
                continue;  //no toggle of animating tgens.
            }
            if ($bONLYTOGGLEIFCLOSED==true && $circ.tgens[t].bOPEN==true) {
                continue;
            }

            //only toggle if the PARENT is open
            if ($circ.tgens[t].parent_tgen.bOPEN==false) continue;
            
            C.toggleAny($circ.tgens[t]);
            return;
        }
    }
    return;
}


/**
 */
C.toggleAny=async($tgen)=> {
    if ($tgen.bANIMATING==true) {
        ANNE.playSound("error_soft");
        return; //no toggle of animating tgens. 
    }
    if ($tgen.gen==0) return await C.toggleTargetNeuron($tgen, true);

    var bANIMATETRANSITION=true;
    return await C.toggleSubGen($tgen, bANIMATETRANSITION);
}


/**
 * clicking a target circuit (or default opening)
 *
 * @param {*} $tgen 
 * @param {*} $bINIT 
 */
C.toggleTargetNeuron=async($tgen, $bSHOWDISPLAY=true)=> {
    C.ACE.setLastAction();
    if ($tgen.bANIMATING==true) {
        ANNE.playSound("error_soft");
        return;
    }
    
    var y = 0;
    var nObj = MEMG.getNeuronByTempNid($tgen.tnid);    

    if ($tgen.bCHASED==false && $tgen.bOPEN==false) {
        y = C.targetSize;
        $tgen.bOPEN=true;
        $tgen.bCHASED=true;
        await $tgen.circ.buildGens(C, $tgen.circ.root_tgen, 1);
    }
        
    else if ($tgen.bOPEN==true) {
        y = 0;
        $tgen.bOPEN=false;
        await C.ACE.hideNeuronDisplay(true);
    }

    else if ($tgen.bOPEN==false) {
        y = C.targetSize;
        $tgen.bOPEN=true;
        C.camToCircuit($tgen.circ.circ_id); //we want to make sure we're looking at this circuit.
    }

    //lift or lower the circuit if it is clicked (or when it is added for default, lift)
    C.ta.gsap.to($tgen.mesh.position, {
        ease:"power1-in",
        duration: .5,
        y: y,
        onUpdate: function() {
        },
        onComplete: async()=> {
            //set the halos for this gen.
            if ( $tgen.bOPEN==true) {
                if ($bSHOWDISPLAY==true) await C.ACE.showNeuronDisplay(nObj, $tgen);

                C.setHalos($tgen); 
                C.ACE.processingDone(true);

                if (ANNE.SETTINGS.soundOk()==true) {
                    ANNE.playSound("low_drop");
                }
            }
            
            //remove halos and hide.
            else {
                //C.zoomOutFromCircuit($tgen.circ.c_index);
                C.removeHalos($tgen);
                C.ACE.processingDone(false);
            }
        }
    } );
}


/**
 * Here we just want to close a circuit and its children.
 * We don't want to do anything with the display neuron area.
 * 
 * Don't set animating flag here.
 * @param {*} $circ 
 * @returns 
 */
C.closeOnlyIfOpen = function($circ) {
    console.log("%c(closeOnlyIfOpen) %c"+$circ.root_tgen.audit, ANNE.TALK.bggreen, ANNE.TALK.bgyellow);
    
    if ($circ.root_tgen.bOPEN==false) return;
    C.ta.gsap.to($circ.root_tgen.mesh.position, {
        ease:"power1-in",
        duration: .1,
        y: 0,
        onComplete: async()=> {
            $circ.root_tgen.bOPEN=false;
            C.removeHalos($circ.root_tgen, .1);   //go fast.
        }
    } );
}




/**
 */
C.removeMesh=function($mesh){
    C.removeByMESHIDFromArray(C._imgs, $mesh.uuid);
    C.removeByMESHIDFromArray(C._targetMeshes, $mesh.uuid);
    C.removeByMESHIDFromArray(C.clickableMeshes, $mesh.uuid);
}



/**
 * 
 * @param {*} $mesh 
 * @param {*} $class 
 * @param {*} $circ 
 * @param {*} $tgen 
 * @param {*} $nObj 
 * @returns 
 */
C.__doNeuronLbl=function($mesh, $class='anne-neuron', $circ, $tgen, $nObj, $bDOIMG=true, $nameOverride="", $Y=10) {
    if ($tgen.tnid!=$nObj.temp_nid) {
        return;
    }
    var neuronLblDiv = document.createElement( 'div' );
    neuronLblDiv.className = $class;
    if ($nameOverride!="") neuronLblDiv.textContent = $nameOverride;
    else neuronLblDiv.textContent = $tgen.getName(null);
    neuronLblDiv.style.marginTop = '-1em';
    neuronLblDiv.classList.add("show");
 
    const primaryNeuronLbl = new CSS3DObject( neuronLblDiv );
    primaryNeuronLbl.lbl_type=C.CSS3_NEURON_LBL;
    primaryNeuronLbl.circ_id = $circ.circ_id;
    primaryNeuronLbl.tgen = $tgen;
    primaryNeuronLbl.position.set( 0, $Y, 0 );
    $mesh.add( primaryNeuronLbl );
    C.css3dLbls.push(primaryNeuronLbl);
    $tgen.primaryNeuronLbl = primaryNeuronLbl;

    return primaryNeuronLbl;
}


/**
 * A relon changed its to value.
 * We need to alter the tgen.
 */
C.updateSingleTO=async($tgenFrom, $fromNid, $oldToValue, $txid)=> {
    var circ = $tgenFrom.circ;
    var fromNOBJ = MEMG.getNeuronByNid($fromNid);
    var oldNOBJ = MEMG.getNeuronByNid($oldToValue);
    var oldTgen = C.getTgen(circ, oldNOBJ.temp_nid, $tgenFrom.tnid);
        
    //remove the out relon so our extend will redo this relon
    //with the new TO NOBJ.
    //this will clean up all arrays and the other tgen too.
    oldTgen.removeInRelon($txid, $tgenFrom);   //will remove the tgen if no ins left.
    oldTgen.parent_tgen.removeTgen(oldTgen, C);
    oldTgen.mesh = null;

    //this should pick up our new relon.
    await $tgenFrom.extendNeuron(C, $tgenFrom.circ, fromNOBJ, true);
}



/**
 * 
 * @param {*} $tgen 
 * @returns 
 */
C.autoOpenTgens = async($tgen)=>{
    for (var t=0; t < $tgen.tgens.length;  t++) {
        var tgen = $tgen.tgens[t];
        if (tgen.anne_chaseable==true) {
            await C.toggleSubGen(tgen);
        }
    }
    return;
}


/**
 * All boxes start alpha..
 * color specific to ownership/sponsorship/status.
 * 
 * Green turns on solid when box is selected.
 * @param {*} $gen 
 * @param {*} $totalNidsInHalo 
 * @returns 
 */
C.getBoxInfos = function($gen, $totalNidsInHalo) {
    var thumb_size = 48;
    thumb_size-=(8*$gen);
    thumb_size = Math.min(12, thumb_size);
    var bDORELONS = true;
    var bUSEALPHA = true;

    var size = 20 - (4*$gen);  
    if ($totalNidsInHalo > C._maxTgenSpokes) {
        size = 4;
        thumb_size = 0;
        bDORELONS = false;
    }
    return {box_size:size, thumb_size:thumb_size, bDORELONS:bDORELONS, bUSEALPHA:bUSEALPHA};
}



/**
 * Add a neuron to the circuit.
 * 
 * This is called for ANY of the gens that pass through 
 * 
 * If this neuron is also in another circuit we are going to be converging
 * the object and placing it in the center of the circuit diagram.
 * This process wil be animated so we can keep both mesh objects (as they are connected to their circuit) 
 * @param {anne_circuit} $circ 
 * @param {neuron_gen} $tgen 
 */
C.addSubGen = function($tgen, $totalInThisHalo=1, $initiating_rel=null, $bNOLBLS=false) {
    var nObj = MEMG.getNeuronByTempNid($tgen.tnid);
    var colorObj = C.getNeuronColor(nObj, $tgen);
    if ($totalInThisHalo>C._maxTgenSpokes) {
        $bNOLBLS=true;
    }
    var sizeObj = C.getBoxInfos($tgen.gen, $totalInThisHalo);
    var box_size = sizeObj.box_size;
    
    $tgen.anne_chaseable=true;    //with no tnids set, all are chaseable
    //if (ANNE.SETTINGS.debugExternal()==true) console.log("neuron is chaseable (can be extended)");

    var geometry = new C.ta.THREE.BoxGeometry(box_size, box_size, box_size);

    if (sizeObj.bUSEALPHA==true) {
        $tgen.mesh = new C.ta.THREE.Mesh( geometry, colorObj.mat_solid_alpha);
    }
    else {
        $tgen.mesh = new C.ta.THREE.Mesh( geometry, colorObj.mat_solid_alpha);
    }
    $tgen.mesh.box_size = box_size;
    
    var y = $tgen.getParentY();

    if ($tgen.anne_chaseable==true) {
        C.clickableMeshes.push($tgen.mesh);
    }

    //elevate gens upward
    //I would like to tilt them out too, in a fly-while type thing.
    //I think i can accomplish that by tilting the parent_tgen away from ITS parent tgen.
    if ($tgen.gen>1) y = y+(200);

    $tgen.mesh.position.y = y;//C.targetSize/2;
    $tgen.setMesh($tgen.mesh); //lolz

    if ($bNOLBLS==false && $tgen.primaryNeuronLbl==null) {
        C.__doNeuronLbl($tgen.mesh, "anne-neuron", $tgen.circ, $tgen, nObj, false, $tgen.getName($initiating_rel), (box_size)+10);
    }

    $tgen.mesh.tgen_ptr = $tgen;

    if ($tgen.parent_tgen!=null && $tgen.parent_tgen.mesh==null) {
        console.log("WTH");
        console.log("tgen parent "+$tgen.parent_tgen.audit+" of gen "+$tgen.parent_tgen.gen+" does not have a mesh yet");
    }
    $tgen.parent_tgen.mesh.add($tgen.mesh);

    return;
}


/**
 * @param {*} $tgen 
 */
C.forceChaseWIKIPEDIA=async($tgen)=>{
    var nObj = MEMG.getNeuronByTempNid($tgen.tnid);
    var test_str = nObj.text;

    C.ACE.processingStart("checking wikipedia for new data");
    
    var pObj = ANNE.HISTORY_LOCAL.getLastBlockFromPullHistory("anne", 'wikipedia', test_str);
        pObj.lastBlock = 0;//optimize later


    var rel = ANNE.MEMG.getSysRelon(nObj.outs, ANNE.MEM.wikipart.temp_nid);
    var query = "";
    
    if(rel!=null) query = rel.to_s;
    else query = nObj.text.replaceAll(" ", "_");
    
    ANNE.A.playSound("processing");
    
    var bDONE = false;

    //person - tested
    if (nObj.bes[0]==MEM.person.temp_nid) {
        var wiki = new wikipediaMakePerson(query, nObj, C.trackerMap, ANNE.superConsolePrint);
        await wiki.resolveAndGet();
        ANNE.A.stopSound("processing");
        bDONE=true;
    }
    
    else if (nObj.bes[0]==MEM.service_term.temp_nid) {
        var props = ["service_term"];
        for (var i=1; i <=18; i++) {
            props.push("service_term"+i);
        }
    
        var targetNOBJ = null;
        await nObj.processAsClumpNeuron();
        if (nObj["role_actor"]==undefined) {
            console.warn("we can't do a target search without role_actor");
            return;
        }
    
        targetNOBJ = ANNE.MEMG.getNeuronByNid(nObj["role_actor"][0].to_s);
        
        var targets = {props:props, only_one:true, actual_nOBJ:nObj, found:false};
    
        //we query the ACTOR of the service term...
        //this will set the proper search string to get the info from wikipedia.
    
        var rel_sys = ANNE.MEMG.getSysRelon(targetNOBJ.outs, ANNE.MEM.wikipart.temp_nid);
        var query = "";
        if (rel_sys!=null) query=rel_sys.to_s;
        else query = targetNOBJ.text;
        
        var wiki = new wikipediaMakePerson(query, targetNOBJ, null, ANNE.superConsolePrint);
        await wiki.resolveAndGet([], targets);
    
        ANNE.A.stopSound("processing");
        bDONE=true;
    }
    
    //company - untested/building
    else if (nObj.bes[0]==MEM.company.temp_nid) {
        var wiki = new wikipediaMakeCompany(query, nObj, C.trackerMap, ANNE.superConsolePrint);
        await wiki.resolveAndGet();
        ANNE.A.stopSound("processing");
        bDONE=true;
    }
    
    //location - untested/unbuilt
    else if (nObj.bes[0]==MEM.location.temp_nid) {
        var wiki = new wikipediaMakeLocation(query, nObj, C.trackerMap, ANNE.superConsolePrint);
        var dataObj = await wiki.resolveAndGet();
        
    
        //hmmm.
        if (dataObj.result_type == "disambig") {
            console.log(dataObj);
    
            //if a US city in a US state, look for nObj.text, state in disambig
            //otherwise, look for nObj.text, country
            var consituentOfNOBJ = nObj.getConstituentOf();
            if (ANNE.isState(consituentOfNOBJ)==true) {
                var pagename = WIKIDATA.findPage(dataObj.pages, consituentOfNOBJ.text);
                if (pagename!=null) {
                    wiki = new wikipediaMakeLocation(pagename, nObj, C.trackerMap, ANNE.superConsolePrint);
                    dataObj = await wiki.resolveAndGet();
                }
            }
            else if (ANNE.isCountry(consituentOfNOBJ)==true) {
                var pagename = WIKIDATA.findPage(dataObj.pages, consituentOfNOBJ.text);
                if (pagename!=null) {
                    wiki = new wikipediaMakeLocation(pagename, nObj, C.trackerMap, ANNE.superConsolePrint);
                    dataObj = await wiki.resolveAndGet();
                }
            }
    
            //just do whatever, next-level-constituent.
            //we may consider doing next level constituent AGAIN!
            //ie, someknownplace, somehigherorder local, somehigherorder local
            else {
                var pagename = WIKIDATA.findPage(dataObj.pages, consituentOfNOBJ.text);
                if (pagename!=null) {
                    wiki = new wikipediaMakeLocation(pagename, nObj, C.trackerMap, ANNE.superConsolePrint);
                    dataObj = await wiki.resolveAndGet();
                }
            }
        }
        
        ANNE.A.stopSound("processing");
        bDONE=true;
    }
    
    //group - untested/unbuilt
    else if (nObj.bes[0]==MEM.group.temp_nid) {
        var wiki = new wikipediaMakeGroup(query, nObj, C.trackerMap, ANNE.superConsolePrint);
        await wiki.resolveAndGet();
        ANNE.A.stopSound("processing");
        bDONE=true;
    }
    
    //group event - untested/unbuilt
    else if (nObj.bes[0]==MEM.group_event.temp_nid) {
        var wiki = new wikipediaMakeGroupEvent(quer,nObj, C.trackerMap, ANNE.superConsolePrint);
        await wiki.resolveAndGet();
        ANNE.A.stopSound("processing");
        bDONE=true;
    }
    
    //name
    else if (nObj.bes[0]==MEM.name.temp_nid) {
        var wiki = new wikipediaMakeName(query,nObj, C.trackerMap, ANNE.superConsolePrint);
        await wiki.resolveAndGet();
        ANNE.A.stopSound("processing");
        bDONE=true;
    }
    
    else {
        C.ACE.showNeuronDisplay(nObj, $tgen);
        C.ACE.processingDone();
        $tgen.bCHASED=true;
        console.log("??");
    }

    if (bDONE==true) {
        C.ACE.addToConsole("acquired data from wikipedia");
        await $tgen.extendNeuron(C, $tgen.circ, nObj, true);
        //start animation of halos
        C.setHalos($tgen);
        C.ACE.showNeuronDisplay(nObj, $tgen);
        C.ACE.processingDone();
        $tgen.bCHASED=true;
    }
}


/**
 */
C.forceChaseTMDB=async($tgen)=>{
    var nObj = MEMG.getNeuronByTempNid($tgen.tnid);

    var tmdb_relon = MEMG.getRelonFromWithRToType(nObj.outs, TYPE_HAS, MEM.TMDB_id.temp_nid, TO_TYPE_NUMBER);
    var tmdb_id = null;
    if (tmdb_relon!=null) tmdb_id = tmdb_relon.to_s;
    
    //for now, we're only doing people for TMDB.
    if (nObj.bes[0]==MEM.person.temp_nid) {
        C.ACE.addToConsole("querying TMDB for person <b>"+nObj.text+"</b>");
        nObj.bTMDBDONE = true;
        
        var tmdbc = new tmdb_connector(this, nObj, "console_print", async($url)=>{
            console.log("add to thumbs");
            
            //C.ACE.addImgByUrlToThumbs($url);  (turning off extra img thumbs here)
            if ($tgen.bHASBOXIMG==false) {
                nObj = MEMG.getNeuronByTempNid($tgen.tnid);
                var afile = await IMAGES.getAnyValidImgAFILE(nObj);
                if (afile!=null) {
                    await ANNE.u3js.imagePlane($tgen.primaryNeuronLbl, afile, 44, new C.ta.THREE.Vector3(0, 30, 0));
                }
            }

            return C.ACE.closeConsole();
        },
        async($newGhostRels=[])=>{
            console.log("TMDB done", $newGhostRels);
            if ($newGhostRels.length) {
                await $tgen.extendNeuron(C, $tgen.circ, nObj, true);

                //start animation of halos
                C.setHalos($tgen);
                
                C.ACE.showNeuronDisplay(nObj, $tgen);
                C.ACE.processingDone();
            }
        });  //results go into the console.

        
        if(tmdb_id!=null) {
            return tmdbc.queryPersonWithKnownID(tmdb_id);
        }

        else {
            var shortName = nObj.shortName();

            //open up for more types?  ["movie", "company"] ?
            ANNE.CHASE.searchTMDBByString(shortName, tmdbc.gotPersonData, ["person"]);
            return;      
        }
    }
        
    else {
        var beNOBJ = nObj.getParent();
        C.ACE.addToConsole("unhandled TMDB query of <b>"+nObj.text+"</b> class <b>"+beNOBJ.text+"</b> neuron ");
        return;
    }
    
}


/**
 */
C.forceChaseWIKISPOOKS=async($tgen)=>{
    var nObj = MEMG.getNeuronByTempNid($tgen.tnid);
    var test_str = nObj.text;
    var search_override_str = ""; 
    var bDOALIASONFAIL=true;

    C.ACE.processingStart("checking Wikispooks for new data");
    
    //get the wikispooks respart, if we have it.
    var fixer_str = C.getFixerQueryStr(nObj, MEM.wikispooks_respart);
    if (fixer_str!="") {
        search_override_str = fixer_str;
        test_str = fixer_str;
    }
    
    var parent_tgen = $tgen.tgen_parent;
    if (parent_tgen==undefined) parent_tgen=null;

    var wikispooksdata = new wikispooks(C, null, null, C.ACE.resolveAmbigs);
    wikispooksdata.setParentTGEN($tgen.tgen_parent);

    wikispooksdata.getData(nObj, test_str, search_override_str, async()=>{
        //process new anne relons from the prop extractor of wikipedia.
        //when completed -- extend the neuron and set the new halos.
        ANNE.GHOSTDATA.ANNEDataFromExternalGeneralized(wikispooksdata, MEM.wikispooks_respart, async($nObj, $wikispooksdata)=>{
            nObj.bWIKISPOOKSDONE=true;
            if (wikispooksdata.status=="good" && nObj!=null) {
                C.ACE.addToConsole("acquired data from wikispooks");
                
                await $tgen.extendNeuron(C, $tgen.circ, nObj, true);

                //start animation of halos
                C.setHalos($tgen);
                
                C.ACE.showNeuronDisplay(nObj, $tgen);
                C.ACE.processingDone();
                $tgen.bCHASED=true;
            }
            
            else {
                C.ACE.showNeuronDisplay(nObj, $tgen);
                C.ACE.processingDone();
                $tgen.bCHASED=true;
            }
        }, C.ACE.addToConsole)
    }, bDOALIASONFAIL);
}


//we want to turn stuff back on.
//typically from turning off a deeper layer..
//so we roll back to the last turned on item.
//we turn  "off" items (fade them) in the gui when they are not
//the focus of attention.. there is so much stuff on screen this 
//helps with seeing what the user is looking at.
C.activateCircuitGuis=function($tgen) {
    C.turnOnAllInPath($tgen);
    if ($tgen.parent_tgen!=null) C.turnOnAllInPath($tgen.parent_tgen);
}


//called when we CLOSE a neuron (remove halos)
//we may have some labels/relons to turn back on if we are in the root view
// OR if there was another circuit line open.
C.turnOnStuffForCircAndAllOpenTgens=function($circ) {
    var opens=0;
    for (var cl=0; cl < C.css3dLbls.length; cl++) {
        if (C.css3dLbls[cl].circ_id!=$circ.circ_id) continue;
        
        if (C.css3dLbls[cl].lbl_type!=C.CSS3_RELON_LBL) continue;

        //if the tgen is open we want its relon lbls turned back on.
        if (C.css3dLbls[cl].tgen.bOPEN==true) {
            opens+1;
            ANNE.A.show(C.css3dLbls[cl].element);
        }        
    }


    if ($circ.root_tgen.bOPEN==true) {

        //also unfade out the neuron lbls so we are in the starting position.
        //C.turnOnAllInPath($circ.root_tgen);
    }
}



/**
 * This turns from top to children on and continues to cascade down...
 * (ie, outs.. so this turns on any given "branch" of our circuit tree)
 * @param {*} $tgen 
 */
C.turnOnAllInPath=function($tgen) {

    //this turns on solid outbound relons from the neuron we are opening (showing its next gen TOs)
    $tgen.turnOnSolidOutboundRelons(true);

    for (var c=0; c < $tgen.tgens.length; c++) {
        var lower_tgen = $tgen.tgens[c];
        if (lower_tgen.bOPEN==false) continue;

        //recurse down the line.
        if (lower_tgen!=undefined && lower_tgen!=null) {
            C.turnOnAllInPath(lower_tgen);
        }
    }
}


C.sizeUpMeshWithoutScale = function($mesh, $currentSize, $desiredSize) {
    var scaleFactor = $currentSize / $desiredSize;

    const positions = $mesh.geometry.attributes.position.array;
    const normals = $mesh.geometry.attributes.normal.array;
    const uvs = $mesh.geometry.attributes.uv.array;

    for (let i = 0; i < positions.length; i += 3) {
      positions[i] *= scaleFactor; // X coordinate
      positions[i + 1] *= scaleFactor; // Y coordinate
      positions[i + 2] *= scaleFactor; // Z coordinate
    }
    $mesh.geometry.verticesNeedUpdate = true;
}


/**
 * 
 * @param {*} $outs 
 * @returns 
 */
C.refreshHanais = async($outs)=> { 
    var queryable_nids = []; 
    for (var r=0; r < $outs.length; r++) {
        if ($outs[r].to_type!=TO_TYPE_ANNE) continue;
        var toNOBJ = MEMG.getNeuronByTempNid($outs[r].t);
        if (toNOBJ==null) continue;
        var pending_hanai = ANNE.HANAI.getPendingHanai(toNOBJ);
        if (pending_hanai!=null) {
            if (queryable_nids.indexOf(toNOBJ.nid)!=-1) continue;
            queryable_nids.push(toNOBJ.nid);
        }
    }

    if (queryable_nids.length==0) return;
    var bFORCEPULL=true;
    await ANNE.CHASE.queryForOutsFroms_not(queryable_nids, bFORCEPULL);
}



/**
 * place objects in a halo around the parent (but above)
 * going button up.
 * As more come in we can adjust them so they are evenly distributed
 * 
 * @param {*} $tgen 
 * @param {*} $oneTimeCallback 

 * @returns 
 */
C.setHalos = function($tgen, $oneTimeCallback=null) {
    if ($tgen.mesh==null) {
        console.warn("NO MESH");
        return;
    }

    var yBump = 140;

    var nids = $tgen.tgens.length;  //number of next gen neurons in this circuit.
    
    var nObj = MEMG.getNeuronByTempNid($tgen.tnid);


    //spiral boxes MUST BE ALPHA'D!!!! GET THIS RIGHT SMOKEY
    if (nids>C._maxTgenSpokes) {
        C.doSpiral($tgen);
        return;
    }
    
    var tgens = $tgen.tgens;
    
    if (nids==0) {
        C.setColorMaterialManual($tgen.mesh, C.GREEN.mat_solid);
        $tgen.mesh.colorObj = C.GREEN;
        return;
    }

    //if this item is a member of a spiral, move it to the CENTER POINT when clicked.
    if ($tgen.gen>0 && $tgen.parent_tgen.bHASPIRAL!=undefined) {
        $tgen.homePosition = new C.ta.THREE.Vector3();
        $tgen.homePosition.copy($tgen.mesh.position);
        yBump = 40;//no need to extend too high.

        if ($tgen.primaryNeuronLbl==null) {
            C.__doNeuronLbl($tgen.mesh, "anne-neuron", $tgen.circ, $tgen, nObj, false, nObj.bestNameHTML(), ($tgen.box_size)+10);
        }

        //maxY here.
        C.ta.gsap.to($tgen.mesh.position, {
            ease:"sine-in",
            duration: .5,
            x:0,
            z:0,
            y:$tgen.parent_tgen.maxY-20
        });
    }

    //update the name every time we display it.
    if ($tgen.primaryNeuronLbl!=null) {
        $tgen.primaryNeuronLbl.element.innerHTML = nObj.bestNameHTML();
    }
    

    var maxInOneRot = 33 - (5*$tgen.gen);
    
    var radius = 120;
    if ($tgen.gen<=0) radius = Math.max(150, Math.min(nids, 12)*20);
    else radius = Math.max(120, Math.min(nids, 10)*20);

    //radius=radius-($tgen.gen*40); //higher gens, tighter circ
    
    if (nids<=5) radius = 75;
    else if (nids<=7) radius = 95;
    else if (nids<=12) radius = 115;

    var sizeObj = C.getBoxInfos($tgen.gen, nids);
    var box_size = sizeObj.box_size;

    //ensure we have the right color/size set.
    if ($tgen.gen>0) {
        
        //active color is green now.
        C.setColorMaterialManual($tgen.mesh, C.GREEN.mat_solid);
        $tgen.mesh.colorObj = C.GREEN;

        //mesh size needs to change basd on number of relons in this tgen.
        if (box_size!=$tgen.mesh.box_size) {
            C.sizeUpMeshWithoutScale($tgen.mesh, $tgen.box_size, box_size);
            $tgen.box_size = box_size;
        }
    }
    else {
        C.setColorMaterialManual($tgen.mesh, C.GREEN.mat_wf);
        $tgen.mesh.colorObj = C.GREEN;
    }
    
    var angleStep = (Math.PI * 2) / nids;
    var yDelta = 1;

    var nidsPerRot = nids;

    //we need more than one rotation..
    if (nids > maxInOneRot) {
        var rots = Math.max(2, Math.ceil(nids/33));
        nidsPerRot = Math.floor(nids/rots);
        angleStep = (Math.PI * 2) / nidsPerRot; 
        yBump=280;
        yDelta = 50/maxInOneRot;    //max y up 400 (so we can see it with our cam settings)
    }
    else if (nids>10) radius+=20;


    if (nids==0) {
        $tgen.bANIMATING=false;
        if ($oneTimeCallback!=null) return $oneTimeCallback();
        else return;
    }


    for (var i=0; i < tgens.length; i++) {
        const tgen = tgens[i];
        let nObj2 = MEMG.getNeuronByTempNid(tgen.tnid);
        
        if (tgen.mesh!=null) {
            
            C.setColorMaterialSolid(nObj2, tgen);
        }
                 
        if (tgen.mesh==null) {
            C.addSubGen(tgen, tgen.parent_tgen.tgens.length, tgen.initiating_rel);
        }

        //only draw them ONE TIME (adding them ot the lines arrays)
        if(tgen.primaryRelon==undefined || tgen.primaryRelon==null) {
            C.drawRelons(tgen, false);
        }
        else {
            var color = C.getRelonColor(tgen.ins[0]);
            tgen.primaryRelon.material.color.set(color);
        }

        //update the name every time we display it.
        if (tgen.primaryNeuronLbl!=null) {
            tgen.primaryNeuronLbl.element.innerHTML = nObj2.bestNameHTML();
        }
        
        var y = yBump + (i*yDelta);
        
        var thisRot = Math.floor(y / nidsPerRot);
        var rotoffset = thisRot * .25;

        //last item to be processed (used on callbacks)
        var final_tnid = -1;
        if (($tgen.tgens.length-1) == i) {
            final_tnid = tgen.tnid;
        }

        var x = (Math.cos(angleStep*(i+rotoffset)) * radius);
        var z = (Math.sin(angleStep*(i+rotoffset)) * radius);

        tgen.mesh.visible=true;

        (function($tgenEnclosed) {

            C.drawRelons($tgenEnclosed);

            $tgenEnclosed.bANIMATING=true;
            C.ta.gsap.to( $tgenEnclosed.mesh.position, {
                ease:"sine-in",
                duration: 1.5,
                x:x,
                z:z,
                y:y,
                onUpdate: function() {
                    C.updateRelons($tgenEnclosed.in_bound_lines, $tgenEnclosed.mesh);
                },

                //draw relon labels only at the end.
                onComplete: function() {
                    if ($tgenEnclosed.primaryNeuronLbl!=null) {
                        ANNE.A.removeClass($tgenEnclosed.primaryNeuronLbl.element, "faded-neuron");
                        ANNE.A.show($tgenEnclosed.primaryNeuronLbl.element);
                    }

                    $tgenEnclosed.bANIMATING=false;
                    $tgen.bANIMATING=false;

                    //non-max mode got to here..
                    if (final_tnid==$tgenEnclosed.tnid && $oneTimeCallback!=null) {
                        console.log("ontime last entry callback");
                        $tgen.circ.sub_gens_open+=1;
                        $oneTimeCallback();
                    }
                }
            } );
        })(tgen);
    }
}


C.doMaxDisplayLathe = function($tgen) {
    const points = [];
    for ( let i = 0; i < 10; i ++ ) {
        points.push( new THREE.Vector2( Math.sin( i * 0.2 ) * 10 + 5, ( i - 5 ) * 2 ) );
    }
    const geometry = new THREE.LatheGeometry( points, 30);
    const material = new THREE.MeshBasicMaterial( { color: 0x00ffff, } );
    const lathe = new THREE.Mesh( geometry, material );
    lathe.scale.set(10, 10, 10)
    $tgen.mesh.add(lathe);

    var vector3Points = geometry.vertices;
    console.log(vector3Points);
}


/**
 * display a spiral of cubes -- this is way too many things!
 * could do a pt cloud but will optimize that later.
 * @param {*} $tgens 
 */
C.doSpiral=function($tgen) {
    const YBUF = 70;

    $tgen.bHASPIRAL = true;

    var ti = 0;
    var totalLoops = 1;
    var itemsPerLoop = 5;
    var itemsHandled = 5;
    var total_tgens = $tgen.tgens.length;
    var startingRadius = 30;

    $tgen.maxY = 0;

    while(itemsHandled<total_tgens) {
        itemsHandled+= (6+totalLoops);
        totalLoops+=1;
    }

    var sizeObj = C.getBoxInfos($tgen.gen, total_tgens);
    var box_size = sizeObj.box_size;
    

    for (var loop = 0; loop < totalLoops; loop++) {
        
        if (ti>=total_tgens) break;
        
        const currentRadius = startingRadius + (loop * 3.1);
        const angleIncrement = (2 * Math.PI) / (6 + loop);

        for (let i = 0; i < itemsPerLoop + loop; i++) {
            var tgen = $tgen.tgens[ti];
            
            var nObj = MEMG.getNeuronByTempNid(tgen.tnid)
            
            //mesh size needs to change basd on number of relons in this tgen.\
            //this is only relevant if the tgen had below the limit and rendered on the screen...
            //and then we query for more and it find more than the limit.
            //those big ones need bo sized down to little ones for the spiral.
            if (box_size!=tgen.mesh.box_size) {
                C.sizeUpMeshWithoutScale(tgen.mesh, tgen.box_size, box_size);
                $tgen.box_size = box_size;
            }
            
            if (tgen.mesh==null) {
                var bNOLBLS = true;
                C.addSubGen(tgen, tgen.parent_tgen.tgens.length, tgen.initiating_rel, bNOLBLS);
            }

            C.setColorMaterialAlpha(nObj, tgen);

            tgen.mesh.visible=true;

            const angle = i * angleIncrement;
            const x = currentRadius * Math.cos(angle);
            const z = currentRadius * Math.sin(angle);
            const y = YBUF+(loop * 6);

            if (y> $tgen.maxY) $tgen.maxY = y;

            tgen.mesh.scale.set(1, 1, 1);
            tgen.mesh.position.set(x, y, z);
            ti+=1;
            if (ti>=total_tgens) break;
        }
    }
}


/**
 * $tgen has been closed.  Remove the halos below it.
 * Reactivate the level above this as the active level.
 * @param {*} $tgen 
 * @param {*} $gen 
 */
C.removeHalos = function($tgen, $time_factor=.7, $relativeDepth=0) {

    //item is no longer selected.
    //turn it back to alpha.
    var nObj = MEMG.getNeuronByTempNid($tgen.tnid);

    if($tgen.gen>0) {
        C.setColorMaterialAlpha(nObj, $tgen);
    }
    else {
        C.setColorMaterialWF(nObj, $tgen);
    }
    
    if ($tgen.tgens.length==0) {
        $tgen.bANIMATING=false;
        return;
    }
    var bONETIME = false;
    var bDORELONLBLS = true;

    //if this was a center focus tgen..
    if ($tgen.gen>0 && $tgen.homePosition!=undefined && $tgen.parent_tgen.bHASPIRAL!=undefined) {
        C.ta.gsap.to($tgen.mesh.position, {
            ease:"sine-out",
            duration: .3,
            x:$tgen.homePosition.x,
            z:$tgen.homePosition.z,
            y:$tgen.homePosition.y
        });
    }


    for (var i=0; i < $tgen.tgens.length; i++) {
        const tgen = $tgen.tgens[i];
        if (tgen.mesh==null) continue;
        
        //handle higher nests first, pass a faster param
        if (tgen.tgens.length) {
            //var nestedNOBJ = MEMG.getNeuronByTempNid(tgen.tnid);
            //console.log("this has nested lvl "+nestedNOBJ.text+" that needs to be shut too!");
            C.removeHalos(tgen, ($time_factor/1.5), $relativeDepth+1); //deeper the open nest, faster animate the close
        }

        tgen.hideInboundLineLbls();

                   
        //use enclosed func here to prevent dramas.
        (function($tgenEnclosed) {
            $tgenEnclosed.bANIMATING=true;
            C.ta.gsap.to( $tgenEnclosed.mesh.position, {
                ease:"power2.InOut",
                duration: 1.5*$time_factor,
                x:0,
                z:0,
                y:0,
                onUpdate: function() {
                    if (bDORELONLBLS==true) C.updateRelons($tgenEnclosed.in_bound_lines, $tgenEnclosed.mesh);
                },

                onComplete: function() {
                    if ($tgenEnclosed.primaryNeuronLbl) ANNE.A.hide($tgenEnclosed.primaryNeuronLbl.element);
                    
                    $tgenEnclosed.bOPEN=false;

                    if (bONETIME==false)  {
                        C.turnOnStuffForCircAndAllOpenTgens($tgen.circ); 
                        bONETIME=true;
                        $tgen.bANIMATING=false;
                        $tgen.circ.sub_gens_open-=1;
                    }
                    $tgenEnclosed.mesh.visible=false;
                    $tgenEnclosed.bANIMATING = false;
                }
            } );
        })(tgen);
    }
}


/**
 * Draw the line & lbl.
 * We need to render ALL the "ins".
 * The first (primary) will be from the FROM (this is the TO)
 * In the circuit, it is a parent tgen.
 * We will, however, at times...  link into a neuron from another branch of the circuit.
 * @param {*} $tgen 
 */
C.drawRelons = function($tgen, $bMAXDISPLAY=false) {
    for (var i=0; i < $tgen.ins.length; i++){
        var in_rel = $tgen.ins[i];
        const points = [];
        var start=new C.ta.THREE.Vector3(0,0,0);
        points.push(start);
        points.push($tgen.mesh.position);
        let geometry = new THREE.BufferGeometry().setFromPoints(points);

        var color = C.getRelonColor(in_rel);

        if ($bMAXDISPLAY==false) {
            var line = new THREE.Line(
                geometry,
                new THREE.LineBasicMaterial({ color: color, transparent:true, opacity:.5 })
            );
        }
        else {
            var line = new THREE.Line(
                geometry,
                new THREE.LineBasicMaterial({ color: color, transparent:true, opacity:.1 })
            ); 
        }
        geometry.setDrawRange( 0, 2 );  //$tgen.circ.color
        line.relon = in_rel;

        //add the line.
        //if we move the shape we are going to have to set the update 
        //line.geometry.verticesNeedUpdate = true
        $tgen.in_bound_lines.push(line);  //lines are stored on the inbound receiver of the line.  They are anchored in the FROM.
        $tgen.parent_tgen.mesh.add(line);
        
        //primary is the first inbound connecting TGEN.  (the parent gen connects to its "TOs", TOs being the TGEN here.
        //This tgen CAN, however, be connected to by other tgens in other locations of the circuit.
        if ($tgen.primaryRelon==undefined || $tgen.primaryRelon==null) {
            $tgen.primaryRelon = line;
        }

        $tgen.circ.lines.push(line);
        $tgen.parent_tgen.out_bound_lines.push(line);
    }
}

   
/**
 * Draws INBOUND lines (edges) to this neuron
 * @param {*} $lines 
 * @param {*} $meshTo 
 */
C.updateRelons = function($lines, $meshTo) {
    for (var l=0; l < $lines.length; l++) {
        var line = $lines[l];
        line.geometry.attributes.position.needsUpdate = true;
        const vStart = new THREE.Vector3(line.geometry.attributes.position[0],line.geometry.attributes.position[1],line.geometry.attributes.position[2])        
        var vEnd = new THREE.Vector3();
        vEnd.copy($meshTo.position);
        let newGeometry = new THREE.BufferGeometry().setFromPoints([vStart, vEnd]);
        line.geometry.dispose()
        line.geometry = newGeometry;
    }    
}


C.getCircuitCount = function() { return Object.keys(C._circuitsMap).length; }
C.setTnidRelnSkips = function($trs) { C.tnid_reln_skips = $trs; }

C.removeByMESHIDFromArray=function($arr, $meshID) {
    for (var a=0; a < $arr.length; a++) {
        if ($arr[a].uuid==$meshID) {
            $arr.splice(a, 1);
            return;
        }
    }
}


/**
 * TODO: RE-WRITE THIS.
 * THIS IS OVERLY COMPLICATED NOW THAT WE ONLY SHOW ONE CIRCUIT AT A TIME.
 * NO MORE COMPARE CIRCUIT IN THIS VIEW/APP.
 * 
 * znid map must be a list of old znids as the key mapped to a new nid.
 * This is only relevant for the top circuits.
 * We use this in case we updated a top neuron from ghost to real and thus 
 * we need to relaunch it again but now need the new nid to get the right neuron launched
 * 
 * if deleteNid is set -- we do not want to regen this one. 
 * 
 * WE DO NOT WANT EXTERNAL CHASES IN CIRCUIT REFRESH MODE
 */
C.refreshAllCircuits = async($oldZNidMap={}, $deleteNID=null)=> {
    var circs_to_reload = [];
    var opens_by_circ = {};

    //loop all circuits and nuke them.
    var keys = Object.keys(C._circuitsMap);
    
    for (var c_index=keys.length-1; c_index>=0; c_index--) {
        var circ = C.getCircById(c_index);
        
        //if this was a ghost and was updated, it is now nuked. :(
        var circOldNid = circ.init_nid;

        //if this was a ghost nid top circ... (that was updated)
        var newCircNid = $oldZNidMap[ circOldNid ];

        if (newCircNid!=undefined) {
            if ($deleteNID==null || $deleteNID!=circOldNid) {
                circs_to_reload.push(newCircNid);
            }
        }
        else {
            newCircNid = circ.nObj.nid;
            if ($deleteNID==null || $deleteNID!=circOldNid) {
                circs_to_reload.push(newCircNid);
            }
        }

        opens_by_circ[newCircNid] = [];

        //all tgens, all lvls.
        for (var t=0; t < circ.tgens.length; t++) {
            var tgen = circ.tgens[t];
            tgen.bANIMATING =false;
            tgen.bHASBOXIMG = false;

            //document a lower neuron was open
            if (tgen.init_ghost==false) {
                if (tgen.bOPEN==true) {
                    if ($deleteNID==null || $deleteNID!=tgen.init_nid) {
                        var parent_nid = null;
                        if (tgen.parent_tgen!=null) {
                            parent_nid = tgen.parent_tgen.init_nid;
                            if ($oldZNidMap[ tgen.parent_tgen.init_nid ]!=undefined) {
                                parent_nid = $oldZNidMap[ tgen.parent_tgen.init_nid ];
                            }
                        }
                        opens_by_circ[newCircNid].push({nid:tgen.init_nid, gen:tgen.gen, parent_nid:parent_nid});
                    }
                }
            }

            else {
                //document a lower ghost (now real) neuron was open
                if (tgen.bOPEN==true){
                    var newLowerNid = $oldZNidMap[ tgen.init_nid ];
                    if ($deleteNID==null || $deleteNID!=tgen.init_nid) {
                        var parent_nid = null;
                        if (tgen.parent_tgen!=null) {
                            parent_nid = tgen.parent_tgen.init_nid;
                            if ($oldZNidMap[ tgen.parent_tgen.init_nid ]!=undefined) {
                                parent_nid = $oldZNidMap[ tgen.parent_tgen.init_nid ];
                            }
                        }
                        opens_by_circ[newCircNid].push({nid:newLowerNid, gen:tgen.gen, parent_nid:parent_nid});
                    }
                }
            }

            //rmove top mesh, all other meshes under will get nuked too.
            if (tgen.gen==0) C.ta.scene.remove(tgen.mesh);

            for (var css=0; css < tgen.in_bound_lines.length; css++) {
                tgen.mesh.remove(tgen.in_bound_lines[css]);
            }
        }
        delete C._circuitsMap[circ.nid];
    }


    C.ACE.resetCSSLbls();
    
    C._circuitsMap = new Object();
    C.css3dLbls = [];

    C._targetMeshes = [];
    C.clickableMeshes = [];
    
    C._aCircuits = [];
    
    ANNE.A.empty("instruct_overlay");

    ANNE.ACE.showLeft3jsOverlay();
    ANNE.A.show("csscontainer");
    ANNE.A.show("container");
    ANNE.A.hide("anne_site");
    
    C.ACE.enable();

    //reload these circuits.  (untested)
    //open up the circuit.
    for (var c=0; c < circs_to_reload.length; c++) {
        var new_nid = circs_to_reload[c];
        var nObj = MEMG.getNeuronByNid(new_nid);

        var circ = C.buildBasicTargetCircuit(nObj);

        //hmm.. how to re-open these too?
        console.log("neurons that were open in this circ:", opens_by_circ[new_nid]);

        var bSHOWDISPLAY = false;
        var highestShowTgen = circ.root_tgen;
        var highestGen = 0;
        var highestShowNOBJ =circ.nObj;

        await C.toggleTargetNeuron(circ.root_tgen, bSHOWDISPLAY);
        
        //by circuit, array of {nid:newLowerNid, gen:tgen.gen} that were open.
        var arrcirc = opens_by_circ[new_nid];
        

        //loop by gen.. to open gen 1's..
        //then any gen 2's.. 3's... etc.
        for (var gen=1; gen < 10; gen++) {
            var bFOUNDGENOPEN = false;
            for (var a=0; a < arrcirc.length; a++) {
                var openObj = arrcirc[a];
                if (openObj.gen==gen) {
                    var NOBJ = MEMG.getNeuronByNid(openObj.nid);
                    if (NOBJ==null) continue;
                    var parentNOBJ = MEMG.getNeuronByNid(openObj.parent_nid);
                    if (parentNOBJ==null) continue;
                    var atgen = C.getTgen(circ, NOBJ.temp_nid, parentNOBJ.temp_nid);
                    if (atgen!=null) {
                        if (atgen.gen>highestGen) {
                            highestGen = atgen.gen;
                            highestShowTgen = atgen;
                            highestShowNOBJ = NOBJ;
                        }

                        atgen.bANIMATING=false;
                        await C.toggleSubGen(atgen);
                        bFOUNDGENOPEN =true;
                    }
                }
            }

            //can't open next gen if we had no prev gens open.
            if (bFOUNDGENOPEN==false) break;
        }

        //we only want to show one of the tgens on refresh.
        //it will either be a main target circ or a lower level value.
        //we want to show the highest gen (deepest into the circuit) tgen we had showing before.
        C.ACE.showNeuronDisplay(highestShowNOBJ, highestShowTgen);

        //build the compare circuit which is slot 2 (length 3) of the _aCircuits.
    }
}



/**
 * Get all matching tgens of a specific tnid.
 * NOTE: do not get multiple matching tgens of the ROOT tgen.
 * There are many backlinks in a circ pointing to the root tgen in various clumps,
 * and back relons.  These are not informative for matching a root circ.
 * @param {*} $circ_id 
 * @param {*} $tnid 
 * @returns 
 */
C.getAllCircTgensByTnid = function($circ_id, $tnid) {
    var matches = [];
    var circ = C._aCircuits[$circ_id];

    for (var c=0; c < circ.tgens.length; c++) {

        if (circ.tgens[c].tnid==$tnid) {
            //we don't need this one again.  It is the root tnid that shows up
            //in other places in the circ.  If the circ a happens to link directly to circuit b's root tgen..
            //it will be a match.  We don't want to highlight all occurances of self.  It is overkill.
            if ($tnid==circ.nObj.temp_nid && circ.tgens[c].gen>0) continue;

            if (circ.tgens[c].CIRC_DEAD_END==true) {
                console.log('skipping dead-end');
                continue;
            }
            
            if (circ.tgens[c].bDUPEDOWNSTREAM==true) {
                console.log("skipping dupe");
                continue;
            }

            //last attempt catch at dupes that shouldn't be in there.
            var test = circ.tgens[c].parent_tgen;
            var bSKIP=false;
            while (test!=null ) {
                if (test.tnid==$tnid) {
                    bSKIP=true;
                    test.bDUPEDOWNSTREAM=true;
                    break;
                }
                test = test.parent_tgen;
            }
            if (bSKIP==false) matches.push(circ.tgens[c]);
        }
    }
    return matches;
}


/**
* When all else fails, just bang it out logically.   Let someone else come in an optimize this nonsense.
*/
C.spaceAtoSpaceB=function($meshA, $meshB) {
    if ($meshA==undefined || $meshB==undefined) return new THREE.Vector3();;
    if ($meshA==null || $meshB==null) return new THREE.Vector3();;
    
    var worldPositionA = new THREE.Vector3();
    $meshA.localToWorld(worldPositionA);

    var worldPositionB = new THREE.Vector3();
    $meshB.localToWorld(worldPositionB);

    var diff = new THREE.Vector3();
    diff.x = worldPositionA.x - worldPositionB.x;
    diff.y = worldPositionA.y - worldPositionB.y;
    diff.z = worldPositionA.z - worldPositionB.z;
    return diff;
}   


C.__localToWorld=function($mesh) {
    var v = new C.ta.THREE.Vector3();
    v.copy($mesh.position);
    var v2 =  $mesh.localToWorld(v);
    return v2;
}

C.__worldToLocal=function($mesh) {
    var v = new C.ta.THREE.Vector3();
    v.copy($mesh.position);
    var v2 =  $mesh.worldToLocal(v);
    return v2;
}


/**
 * 
 * @param {*} $cid 
 * @returns 
 */
C.getCircById = function($cid) {
    var keys = Object.keys(C._circuitsMap);
    var cid = parseInt($cid);
    for (var k=0; k < keys.length; k++) {
        if (C._circuitsMap[keys[k]].circ_id==cid) return C._circuitsMap[keys[k]];
    }
    return null;
}


/**
 * Comp tgens are unique in the comp circuit so we can just pull by tnid.
 * @param {*} $tnid 
 * @returns 
 */
C.getCompTgen=function($tnid) {
    var circ = C._aCircuits[2];
    if (circ==null) return null;
    for (var t=0; t < circ.tgens.length; t++) {
        var tgen = circ.tgens[t];
        if (tgen.tnid==$tnid) return tgen;
    }
    return null;
}


C.getTgenByCircAndTgenId=function($circ_id, tgenid) {
    var circ = C.getCircById($circ_id);
    if (circ==null) return null;
    for (var t=0; t < circ.tgens.length; t++) {
        var tgen = circ.tgens[t];
        if (tgen.tgenid==tgenid) return tgen;
    }
    return null;
}


/**
 * 
 * @param {*} $circ 
 * @param {*} $tnid 
 * @returns 
 */
C.getTgen = function($circ, $tnid, $fromTnid) {
    for (var t=0; t < $circ.tgens.length; t++) {
        var tgen = $circ.tgens[t];
        if (tgen.tnid==$tnid) {
            if (tgen.parent_tgen.tnid==$fromTnid) return tgen;
        }
    }
    return null;
}


///// CAM TWEENS.

C.zoomOutFromCircuit2 = function($c_index, $localMesh=null) {
    C.camToCircuit($c_index, true, $localMesh);
}

/**
 * Move camera position to outter cam pt.
 * Set zoom to default home zoom.
 * Set lookat to circ root_tgen
 * @param {*} $c_index 
 */
C.zoomOutFromCircuit = function($c_index) {
    return;
    if ($c_index==0) C._moveCamToHomePt(4);
    else if ($c_index==1) C._moveCamToHomePt(0);
    else if ($c_index==2) C._moveCamToHomePt(2);
    else if ($c_index==3) C._moveCamToHomePt( 6);
}



/**
 * Move camera to a home pt and fix zoom to home outter zoom.
 * @param {*} $p_index 
 */
C._moveCamToHomePt = function($p_index) {
    return;
    var pt = C.cam_points[$p_index];
    //C.ACE.controls.setOrbitPoint( 0, 0, 0 ); 
    C._changeCameraPosition(pt, true );
    C.ACE.controls.zoomTo( C.homeOutterZoom );
    C._last_p_index = $p_index;
}


/**
 * Move camera to a inner home pt and fix zoom to home inner zoom.
 * 
 * too complicated, overriding this for now to just move cam to homeInnerzoom pt.
 * @param {*} $p_index 
 */
C._moveCamToInnerHomePt = async($p_index, $bDOZOOM=true)=> {
    return;
    var pt = C.cam_points[$p_index];
    //C.ACE.controls.setOrbitPoint( 0, 0, 0 ); 
    await C.ACE.controls.zoomTo( C.homeInnerZoom );

    C._changeCameraPosition(pt, true );
    C._last_p_index = $p_index;
}




C.zoomInToCircuit = async($c_index, $3dobj, $bZOOM=true, $bUSEWORLDSPACE=true)=> {
    return;
    if ($c_index==0) await C._moveCamToInnerHomePt(4, $bZOOM);
    else if ($c_index==1) await  C._moveCamToInnerHomePt(0, $bZOOM);
    else if ($c_index==2) await  C._moveCamToInnerHomePt(2, $bZOOM);
    else if ($c_index==3) await  C._moveCamToInnerHomePt(6, $bZOOM);
    if ($bUSEWORLDSPACE==true) {
        var worldPt = C.__localToWorld($3dobj);
        C.ACE.controls.setOrbitPoint( worldPt.x, worldPt.y, worldPt.z );
    }
    else {
        C.ACE.controls.setOrbitPoint( $3dobj.position.x, $3dobj.position.y, $3dobj.position.z ); 
    }
}


/**
 * Move the position of the camera..
 * It will stay looking at the target.
 * @param {*} $pt 
 * @param {*} $bANIMATE 
 */
C._changeCameraPosition=function($pt, $bANIMATE=true) {
    return;
    C.ACE.controls.setPosition($pt.x, $pt.y, $pt.z, $bANIMATE);
}


/**
 * Point the camera without moving it.
 * we have to convert local points to world for the camera to look at the right point.
 * @param {*} $3dobj 
 * @param {*} $bANIMATE 
 */
C.cameraLookat=function($3dobj, $bANIMATE=true) {
    return;
    var worldPt = C.__localToWorld($3dobj);
    //C.ACE.controls.setTarget( worldPt.x, worldPt.y, worldPt.z, $bANIMATE );
    C.ACE.controls.setTarget( worldPt.x, worldPt.y, worldPt.z, $bANIMATE );
}


/**
 * 
 * @param {*} $c_index 
 * @param {*} $lookatMesh 
 */
C._moveCameraToInnerHomePtwithY = function($c_index, $3dObj) {
    return;
    var p_index=0;
    if ($c_index==0 ) p_index = 4;
    else if ($c_index==1) p_index =  0;
    else if ($c_index==2) p_index =  2;
    else if ($c_index==3) p_index =  6;
    
    //orbit the circuit root.
    //C.ACE.controls.setOrbitPoint( C._targetMeshes[$c_index].position.x, C._targetMeshes[$c_index].position.y, C._targetMeshes[$c_index].position.z );
  
    //animate the lookat (setTarget) to the world coordinates of the mesh we want to look at
    C.cameraLookat($3dObj, true);

    //animate camera position to the near-side inner-cam pt
    var pt = C.cam_points_inner[p_index];
    C._changeCameraPosition( pt, true );

    C._last_p_index = p_index;
}


/**
 * 
 * @param {*} $lookat 
 */
C.zoomToCloseup = async($c_index, $lookatMesh=null)=>{
    //need world space here???
    return;
    
    //look at the right place, then zoom.
    if ($lookatMesh!=null) {

        
        //var meshworld = C.__localToWorld($lookat);

        C._moveCameraToInnerHomePtwithY($c_index, $lookatMesh);
    }

    //gentle zoom at the end.
    C.ACE.controls.zoomTo( C.closeUpZoom, true );
}



/**
 * outter point cam. (way outside look)
 * @param {*} $c_index 
 * @param {*} $bCAMTOLOWAFTER 
 * @returns 
 */
C.camToCircuit_outter = function($c_index=0, $bCAMTOLOWAFTER=true) {
    var pl = C.placements[$c_index];
    C.ACE.controls.smoothTime = .25;
    var camIndex = 0;
    var zoomLvl = C.homeOutterZoom;

    if ($c_index==0) camIndex = 0;
    else if ($c_index==1) camIndex = 4;
    else if ($c_index==2) camIndex = 2; //or 6

    C.ACE.controls.setTarget(pl.x, pl.y, pl.z, true); 
    C.ACE.controls.zoomTo( zoomLvl); 
    
    C.ACE.controls.smoothTime = .25;
    
    C.ta.gsap.to(C.ACE.camera.position, {
        ease:"power1-in",
        duration: .5,
        z: C.cam_points_way_out[camIndex].z,
        x: C.cam_points_way_out[camIndex].x,
        y: C.cam_y,
        onUpdate: function() {
        },
        onComplete: async()=> {
            C.ACE.controls.saveState();
            if ($bCAMTOLOWAFTER==true) C.camToLowView($c_index)
        }});
    return;
}


C.camToCircuit = function($c_index=0, $bCAMTOLOWAFTER=true, $localMesh=null) {
    var pl = C.placements[$c_index];
    C.ACE.controls.smoothTime = .25;
    var camIndex = 0;
    var zoomLvl = C.homeOutterZoom;

    if ($c_index==0) camIndex = 0;
    else if ($c_index==1) camIndex = 4;
    else if ($c_index==2) camIndex = 2; //or 6

    var localPt = new C.ta.THREE.Vector3(0,0,0);
    if ($localMesh!=null) {
        $localMesh.localToWorld(localPt);
        C.ACE.controls.setTarget(localPt.x, localPt.y, localPt.z, true); 
        //zoomLvl = C.homeInnerZoom;

         //test this.. we don't want to look "up" to an inner tgen
         //we want to be looking downwards or even. .from above.
        $bCAMTOLOWAFTER=false; 
        C.ACE.controls.smoothTime = .7;
        C.ACE.controls.zoomTo( zoomLvl, true );
    }
    else {
        C.ACE.controls.setTarget(pl.x, pl.y, pl.z, true); 
        C.ACE.controls.zoomTo( zoomLvl); 
    }
    
    //put camera in 
    C.ACE.controls.smoothTime = .25;
    
    C.ta.gsap.to(C.ACE.camera.position, {
        ease:"power1-in",
        duration: .5,
        z: C.cam_points[camIndex].z,
        x: C.cam_points[camIndex].x,
        y: C.cam_y,
        onUpdate: function() {
        },
        onComplete: async()=> {
            C.ACE.controls.saveState();
            if ($bCAMTOLOWAFTER==true) C.camToLowView($c_index)
        }});

    //console.log("camera position:",C.ACE.camera.position);
    return;
}


//pans camera down to flat view.
C.camToLowView = async($c_index)=> {
    C.ACE.controls.smoothTime = .7;
    C.ACE.controls.rotatePolarTo( C.ACE.controls.maxPolarAngle-.1, true);
    if ($c_index==0) {
        await C.ACE.controls.rotateAzimuthTo( 7.1, true);
        //C.ACE.controls.zoomTo( ); 
    }
    else if ($c_index==1) {
        
    }
    if ($c_index==2) {
        
    }
    return;
}



/**
 * Destroy it all.
 */
C.resetDisplay = function() {

    var root_nids = [];
    
    for (var c=0; c < C._aCircuits.length; c++) {
        var circ = C._aCircuits[c];
        var rootNOBJ = MEMG.getNeuronByTempNid(circ.root_tgen.tnid);
        root_nids.push(rootNOBJ.nid);

        //all tgens, all lvls.
        for (var t=0; t < circ.tgens.length; t++) {
            var tgen = circ.tgens[t];
            var nObj = MEMG.getNeuronByTempNid(tgen.tnid);
            nObj.bWIKISPOOKSDONE=false;
            nObj.bWIKIPEDIADONE=false;
            nObj.bTMDBDONE=false;
            if (tgen.gen==0) C.ta.scene.remove(tgen.mesh);

            for (var css=0; css < tgen.in_bound_lines.length; css++) {
                tgen.mesh.remove(tgen.in_bound_lines[css]);
            }

            //for the comp circ outbound lines.
            if (tgen.out_bound_lines!=undefined && tgen.out_bound_lines!=undefined) {
                for (var css=0; css < tgen.out_bound_lines.length; css++) {
                    tgen.mesh.remove(tgen.out_bound_lines[css]);
                }
            }
        }

        delete C._circuitsMap[circ.nid];
    }

    C.ACE.resetCSSLbls();

    C._circuitsMap = new Object();
    C._aCircuits = [];
    C.css3dLbls = [];
    C._targetMeshes = [];
    C.clickableMeshes = [];
    C._aCircuits = [];

    ANNE.A.empty("instruct_overlay");

    C.ACE.hideOnly();

    return root_nids;
}



export { C };