class history_slim  {
    constructor(){
        //tracks our space-wide FULL PULLS by the most recent block id in the results.
        this._aHistory = []; 

        //if we ever want to load memory from a snapshot or post-boot
        //we can set this number higher than 0 to optimize.
        this._ANNE_EARLIEST_BLOCK = 0;  

        //just a helpful heads up to us.
        this._GREATEST_LAST_KNOWN_BLOCK = this._ANNE_EARLIEST_BLOCK;
        this._GLKB_SOLID = 0;
        
        //have to make my own.  cant get the built in auto key figured out (don't want to waste any more time)
        this.pullid = 1;    

        this.bhid = "annechain_block_height";
    }

    getGLKB = function() {
        if (this._GLKB_SOLID > this._GREATEST_LAST_KNOWN_BLOCK) return this._GLKB_SOLID;
        return this._GREATEST_LAST_KNOWN_BLOCK;
    }

    //what to do if we have a LOWER HEIGHT than relons in our mem?
    //this means we have had a rollback..
    //i think it is ok to leave the relon in place.
    updateGLKB_SOLID = function($b) {
        this._GREATEST_LAST_KNOWN_BLOCK = $b;
        this._GLKB_SOLID = $b;
        ANNE.A.setHTML("annechain_block_height", "#"+ANNE.UTILS.commas($b));
    }

    updateGLKB_fromRelons = function($b) {
        if ($b<=this._GREATEST_LAST_KNOWN_BLOCK) return;

        this._GREATEST_LAST_KNOWN_BLOCK = $b;   
    }

    getNextPullID = function () {
        var pullid = HISTORY.pullid;
        HISTORY.pullid +=1;
        return pullid;
    }


    /**
     * Process the pull history (optimizer).
     * 
     * We store this in a DB in localStorage                                             
     * 
     * @author {Mr.Scatman}
     */
    processPullHistory = function ($aFullPullHistory) {
        var maxpullid = 0;
        if ($aFullPullHistory!=undefined && $aFullPullHistory!=null) {
            HISTORY._aHistory = $aFullPullHistory;
            for (var h=0; h < HISTORY._aHistory.length; h++) {
                var hObj = HISTORY._aHistory[h];
                if (hObj.pullid > maxpullid) maxpullid = hObj.pullid;
                hObj.bDIDIDB = false;
            }
        }

        HISTORY.pullid = maxpullid + 1;
    }
    

    /**
     * Get the last block we found results using this exact query.
     * 
     * Try to save processing/pull attempts by looking for this pull obj.  
     * We will still try to pull from the blockchain but with a min block  
     * number from the last known spot we already pulled in data for this  
     * type of pull.                                                       
     * 
     * @author {Mr.Scatman}
     */
    getLastBlockFromPullHistory = function($prefix, $functionName, $target, $extra=null) {
        var extra="";
        if ($extra!=null) extra=$extra;

        if (HISTORY._aHistory.length==0) {
            //console.log("%c NO HISTORY ---- U SHOULD SEE THIS ONCE for "+$functionName+" target:"+$target+" "+$extra, 'background: white; color: red');
            return HISTORY.__makeDefaultPullHistoryObj($prefix, $functionName, $target, $extra);
        }

        for (var h=0; h < HISTORY._aHistory.length; h++) {
            var phObj = HISTORY._aHistory[h];
            if (phObj.user_nid!=USER._userNObj.nid) continue;
            if (phObj.prefix!=$prefix || phObj.functionName!=$functionName || phObj.target!=$target) continue;
            if (extra!="") {
                if (phObj.extra!=$extra) {
                    continue;
                }
            }
            //console.log("%c history - we found a pull history object for function: "+$functionName+" target:"+$target, 'background: green; color: white');
            //console.log(phObj);
            return phObj;
        }

        //we had history in our  history object, but not for THIS entry.
        //console.warn("temp new HISTORY object (will save if we find results) for ", $prefix, $functionName, $target);
        return HISTORY.__makeDefaultPullHistoryObj($prefix, $functionName, $target, $extra);
    }




    /**
     * Update our history trackers for latest goodies found.
     * 
     * Look for more recent .time (block int number).   Update existing 
     * tracker records or push a new one. 
     * block index is only relevant within the scope of the searrch, not 
     * for the highest block we've seen elsewhere. 
     * 
     * @author {Mr.Scatman}
     * 
     */ 
    processResultsForNewPullHistory=async($arr, $phObj)=> {
        var localMaxBlock = $phObj.lastBlock;     
        var bUPDATINGEXISTINGRECORD = false;

        if ($arr==undefined || $arr==null) return;

        //process results for max time. 
        for (var r=0; r  < $arr.length; r++) {
            var obj = $arr[r];
            if (obj.time > localMaxBlock) localMaxBlock = obj.time;
        }
        
        //greatest of either the localmax (observed, or lastrun)
        //and the greatest last known block
        localMaxBlock = Math.max(localMaxBlock, HISTORY._GREATEST_LAST_KNOWN_BLOCK);
        
        //our results have a greater block than our previous history and we have an existing pull history object
        //we want to update this record with the new block. 
        if (localMaxBlock > $phObj.lastBlock  && $phObj.pullid>0) {
            bUPDATINGEXISTINGRECORD=true;
            if(localMaxBlock>=1500000000) {
                console.log("%c HISTORY localMAX BLOCK TOO BIG THIS SHOULD NEVER HAPPEN!!!", 'background: red; color: white');
            }
            else $phObj.lastBlock = localMaxBlock; 
        }    

        //this is a first time new entry, not new info.
        else if ($phObj.pullid==0) {
            bUPDATINGEXISTINGRECORD = false;
            if(localMaxBlock>=1500000000) {
                console.log("%c HISTORY localMAX BLOCK TOO BIG THIS SHOULD NEVER HAPPEN!!!", 'background: red; color: white');
            }
            else $phObj.lastBlock = localMaxBlock;
        }

        //assign it a real ID (next in line)
        if ($phObj.pullid==0) {
            $phObj.lastBlock = $phObj.lastBlock+1;
            $phObj.pullid = HISTORY.getNextPullID();    //this accepts the existing pullid and then incriments it by one
            if (ANNE.SETTINGS.bCONSOLE_DEBUG==true) console.log("%c ((inserting new history record with pullid:"+$phObj.pullid+" "+$phObj.functionName+" target:"+$phObj.functionName+"))", 'background: white; color: green');
            var done = await DB.___update($phObj, "fullPullHistory");    //insert a new record
        }

        //if we had new info, update it.
        else if (bUPDATINGEXISTINGRECORD) {
            $phObj.lastBlock = $phObj.lastBlock+1;
            console.log("%c ((updating existing record with new data))" , 'background: green; color: white');
            var done = DB.___update($phObj, "fullPullHistory");    //update the localDB record
        }

        //existing pull history object and no new results.
        //else console.warn("((no new info from block > "+$phObj.lastBlock+" info to add to pull history))")
        return;
    }


    /**
     * @author {Mr.Scatman}
     */ 
    __makeDefaultPullHistoryObj=function($prefix, $functionName, $target, $extra=null) {
        var phObj = new Object();
        phObj.user_nid = USER._userNObj.nid;
        phObj.prefix = $prefix;
        phObj.functionName = $functionName;
        phObj.target = $target;
        phObj.extra = $extra;
        phObj.lastBlock = 0;//HISTORY._GREATEST_LAST_KNOWN_BLOCK;
        phObj.pullid = 0;  //we will set this to the next one in line if we are going to insert the record
        phObj.bDONETHISSESSION = false;
        phObj.bDIDIDB = false;          //for RAM optimization.
        HISTORY._aHistory.push(phObj);
        return phObj;
    }
}

const HISTORY = new history_slim ();