const AXNPP_NO_MATCHES = "axnpp_no_matches";    //we have no valid solutions alive
const AXNPP_STILL_ALIVE = "axnpp_still_alive";  //we have parts matched with some left to match
const AXNPP_MATCH = "axnpp_match";              //we have a match
const AXNPP_NOT_A_MATCH = "axnpp_not_a_match";  //not a match.
const AXNPP_NO_MORE_PARTS = "axnpp_no_more_parts";
const AXNPP_DONE = "axnpp_done";


class anne_x_neural_pat_pool {
    constructor() {}


    /**
     * 
     * @param {anne_x_neural_pat} $axnp 
     * @param {*} $next_wd 
     * @returns 
     */
    nextWdPatMatch=function($axnp, $next_wd) {
        if ($axnp.bDONE==true) return AXNPP_DONE;

        //get the next piece in the pattern.
        var patPiece = $axnp.getNextPatPiece(); //incriments pp_index for the next pat piece
        if (patPiece==undefined || patPiece==null) {
            return AXNPP_NO_MORE_PARTS;
        }

        var getNeuron=function($patPiece, $thisIndex) {
            var _wdNOBJ = null;
            //loop our neurons (there may be more than one option.
            //if we have any options that fit the class of the pattern piece, continue.
            for (var n=0; n < $next_wd.tnids.length; n++) {
                var neuron = MEMG.getNeuronByTempNid($next_wd.tnids[n]);
                if (neuron.anyParentHasBeOf($patPiece.nObjClass.temp_nid)==false) {
                    continue;
                }
                else {
                    _wdNOBJ = neuron;
                    $next_wd.patMatchNObj = _wdNOBJ;
                    break;
                }
            }

            var manualIndex = $thisIndex+1;
            if (_wdNOBJ==null && $patPiece.optional==true) {
                if ((manualIndex < this.patPieces.length)) {
                    var patP =  this.patPieces[manualIndex];
                    return getNeuron(patP, manualIndex);
                }
            }
            return _wdNOBJ;
        }




        var wdNOBJ = getNeuron(patPiece, $axnp.pp_index-1);

        if (wdNOBJ==null) {
            $axnp.matchObj.score = 0;
            $axnp.matchObj.status=AXNPP_NOT_A_MATCH;
            return AXNPP_NOT_A_MATCH;
        }

        //$axnp.matchObj.score=1;
        $axnp.matchObj.pieces_found+=1;

        patPiece.setWdvalue(wdNOBJ);
        $axnp.setWdValue(wdNOBJ);

        //update the score of this pattern.
        if ($axnp.matchObj.pieces_found==$axnp.real_pieces) {
            $axnp.matchObj.status=AXNPP_MATCH;
            $axnp.matchObj.score=1;
            return AXNPP_MATCH;
        }
        else {
            $axnp.matchObj.score = $axnp.matchObj.pieces_found/$axnp.real_pieces;
            $axnp.matchObj.status=AXNPP_STILL_ALIVE;
            return AXNPP_STILL_ALIVE;
        }
    }



    /**
     * 
     * Score the pattern for match against the conceptualized role's nObj and mods.
     * @param {anne_x_neural_pat}
     * @param {array} - array of nObjs (including non space chars between words)
     */
    rolePatMatchWithWds=async($axnp, $role)=> {
        var roleParts=[];
        $axnp.pieces_found = 0;
        
        //pushing our observed roles and mods to roleparts so we can
        //look for these against the pattern for a match
        //roleParts.push($role.nObj);
        for (var w=0; w < $role.ph.l1str_nosignal_wds.length; w++) {
            var wdObj = $role.ph.l1str_nosignal_wds[w];

            //push the ender on the stack too
            if (wdObj.ender==true) {
                roleParts.push(wdObj.charObj_ender.charNOBJ);
            }

            else if (wdObj.type=="punct-break") {
                continue;
            }

            else if (wdObj.patMatchNObj!=null) {
                roleParts.push(wdObj.patMatchNObj);
            }
        }

        //no chance.
        if ($axnp.req_pieces > roleParts.length) {
            $axnp.matchObj.score=0;
            $axnp.matchObj.status="bad";
            $axnp.matchObj.reasons="different number of pieces";
            return;
        }

        var skips=0;
        for (var p=0; p < $axnp.patPieces.length; p++) {
            var pp = $axnp.patPieces[p];
            var rolePartNObj = roleParts[p-skips];
            //monitor the rolePartNObj from charObj_ender..
            //we don't want alpha/num or spaces here.  Just the ender/breakers
            console.log("%c pp "+pp.nObjClass.text+" %cvs mod %c '"+rolePartNObj.text+"'", TALK.bggreen, TALK.black, TALK.bggreen);
            if (rolePartNObj.anyParentHasBeOf(pp.nObjClass.temp_nid)==false) {
                if (pp.optional==null) {
                    $axnp.matchObj.score=0;
                    $axnp.matchObj.status="bad";
                    $axnp.matchObj.msg="pieces mismatch at p "+p;
                    return;
                }
                else {
                    skips+=1;
                    continue;
                }
            }
            else {
                console.log("%c pp mod match", TALK.bgpink);
            
                if (pp.max!=null) {
                    //evaluate max.
                }
                else if (pp.min!=null) {
                    //evaluate min
                }

                $axnp.matchObj.pieces_found+=1;
                pp.setValue(rolePartNObj, $role.ph);
            }
        }

        
        if ($axnp.matchObj.pieces_found>=$axnp.req_pieces) {
            $axnp.matchObj.status="good";
            //make new patResultingClass ---
            //OR find existing patResultingClass
            await $axnp.lookupOrMake($role);
        }
    }

}

class anne_x_neural_pat_piece {
    constructor($type, $nObjClass, $max=null, $min=null, $optional=false, $debug="") {
       this.order=$type;
       this.nObjClass=$nObjClass; //this is the value
       this.max=$max;
       this.min=$min;
        
       this.optional=$optional;
       if (this.optional==undefined || this.optional==null) this.optional=false;
       
       this.debug=$debug;
       this.value = null;
       this.reln = null;
       this.reln_text = "";
       this.relNOBJ = null;
       this.to_s = null;
       this.to_text = "";
       this.toNOBJ = null;
    
       this.new = true;

       this.charObj_ender = null;   //helps w/ pat matches.

       //for sequential processing.
       this.wdNOBJS = [];
    }


    printPiece = function() {
        console.log(this.debug);
    }

    setWdvalue = function($nObj) {
        this.to_s = $nObj.nid;
        this.to_text = $nObj.text;
        this.toNOBJ=$nObj;
        if ($nObj.bNEW!=undefined && $nObj.bNEW==true) {
            this.new=true;
        }
    }

    setValue = function($nObj, $ph) {
        this.to_s = $nObj.nid;
        this.to_text = $nObj.text;
        this.toNOBJ=$nObj;
        if ($nObj.bNEW!=undefined && $nObj.bNEW==true) {
            this.new=true;
        }

        //not sure why we need this one any more (could be dep)
        if ($ph!=null) {
            if ($ph.l1str_nosignal_wds.length) {
                var lastWdObj = $ph.l1str_nosignal_wds[$ph.l1str_nosignal_wds.length-1];
                if (lastWdObj.charObj_ender!=null) {
                    this.charObj_ender=lastWdObj.charObj_ender;
                }
            }
        }
    }
}


class anne_x_neural_pat {
    constructor($patNObj, $type="match", $partials=null) {
        this.patNObj = $patNObj;
        this.patPieces = [];
        this.real_pieces=0;
        this.req_pieces = 0;
        this.patResultingClass=null;
        this.statusObj = {status:"init", msg:""};
        this.matchObj = {status:"", score:0, msg:"", pieces_found:0, new:false};
        this.name = "";
        this.finalNOBJ = null;
        this.type=$type;
        this.find_relon_pat = {from_nid:"", f:"", reln:"", r:"", to_s:"", t:""};
        
        //for sequential processing mode.
        this.pp_index=0;    
        this.wdNOBJS = [];
        this.bDONE=false;

        if ($type=="match") {
            this.initMatchPatProcessing();
        }
        else if ($type=="name") {
            this.initNamePatProcessing($partials);
        }
    }

    setWdValue=function($wdNOBJ){
        this.wdNOBJS.push($wdNOBJ);
    }

    /**
     * For in-time sequential processing of pat pieces..
     * This is used to check for a pat piece by piece as we encounter it.
     * We do not have the pieces all at once.
     * @returns 
     */
    getNextPatPiece=function() {
        if (this.pp_index < this.patPieces.length) return this.patPieces[this.pp_index++];
        else this.bDONE=true;
        return null;
    }
    getNextPatPieceManual=function($index){
        if ($index < this.patPieces.length) return this.patPieces[$index];
        else this.bDONE=true;
        return null;
    }


    printPatPieces = function() {
        for (var p=0; p < this.patPieces.length; p++) {
            this.patPieces[p].printPiece();
        }   
    }

    /**
     */
    initMatchPatProcessing = function() {
        if (this.patNObj==undefined || this.patNObj==null) return;
        var brain_rels = MEMG.getBrainRels(this.patNObj.outs);
        this.processBrainRelsForPatMatch(brain_rels);

        //where do we do _optional vals?
    }

    initNamePatProcessing = function($partials) {
        if (this.patNObj==undefined || this.patNObj==null) return;
        var brain_rels = MEMG.getBrainRels(this.patNObj.outs);
        console.log("%c(processBrainRelsForName)", TALK.bgpink);
        brain_rels.sort(this.brainSort);
        var lastActive = null;
        var nameParts = [];
        for (var b=0; b < brain_rels.length; b++) {
            var brel = brain_rels[b];
            if (brel.r==MEM.eval.temp_nid) {
                var toNOBJ = MEMG.getNeuronByTempNid(brel.t);
                console.log(toNOBJ);
                for (var p=0; p < $partials.length; p++) {
                    if (toNOBJ.temp_nid==$partials[p].r) {
                        var realTONOBJ = $partials[p].toNOBJ;
                        if (realTONOBJ.value==undefined) {
                            var val = MEM.__MASTER_GET_VALUE(realTONOBJ);
                            nameParts.push(val);
                        }
                        else nameParts.push(realTONOBJ.value);
                    }
                }
            }
            else if (brel.r==MEM.__min_digits.temp_nid) {
                if (nameParts.length==0) {
                    return;
                }
                var formatted = UTILS.formatDigits(nameParts[ nameParts.length-1], brel.to_s);
                nameParts[ nameParts.length-1] = formatted;
            }
            else if (brel.r==MEM.ascii_eval.temp_nid) {
                var toNOBJ = MEMG.getNeuronByTempNid(brel.t);
                console.log(toNOBJ);
                var param_name = MEM.ascii_eval.PARAM_NAME; //s.b. char
                console.log(param_name);
                nameParts.push(toNOBJ.char);
            }
        }
        this.name = nameParts.join("");
    }


    /**
     * pat for matching
     */
     processBrainRelsForPatMatch=function($brs) {
        $brs.sort(this.brainSort);
        var lastActive = null;
 
        var setToNextActive=null;
        var setProp = "";

        for (var b=0; b < $brs.length; b++) {
            var brel = $brs[b];

            //ANNE.TALK.printRelon(brel);

            if (brel.r==MEM.__make_new.temp_nid) {
                //if (lastActive==null) {
                    var nOBJ = MEMG.getNeuronByTempNid(brel.t);
                    this.patResultingClass = nOBJ;
                    //console.log("pat resulting class: "+nOBJ.text );
                //}
            }
            else if (brel.r==MEM.relon_from.temp_nid) {
                var nOBJ = MEMG.getNeuronByTempNid(brel.t);
                if (nOBJ.temp_nid==MEM.__ACTIVE.temp_nid) {
                    //this.find_relon_pat.from = "";
                    //this.find_relon_pat.f = "";
                }
            }
            else if (brel.r==MEM.relon_reln.temp_nid) {
                var nOBJ = MEMG.getNeuronByTempNid(brel.t);
                //this.find_relon_pat.reln = nOBJ.nid;
                //this.find_relon_pat.r = nOBJ.temp_nid;
                
                lastActive.reln = nOBJ.nid;
                lastActive.reln_text = nOBJ.text;
                lastActive.relNOBJ = nOBJ;
            }

            else if (brel.r==MEM.relon_to.temp_nid) {
                var nOBJ = MEMG.getNeuronByTempNid(brel.t);
                //this.find_relon_pat.to = nOBJ.nid;
                //this.find_relon_pat.t = nOBJ.temp_nid;
                setToNextActive=lastActive;
                setProp="t";
            }
            
            //active neuron -- the next pat piece
            else if (brel.r==MEM.__ACTIVE.temp_nid) {
                if (brel.to_type!=TO_TYPE_ANNE) {
                    this.statusObj.msg = "malformed pat - active must be a TO_TYPE_ANNE"
                    this.status="bad";
                    return; 
                }
                var nOBJ = MEMG.getNeuronByTempNid(brel.t); //integer..
                var pp = new anne_x_neural_pat_piece (brel.type, nOBJ, null, null, false, nOBJ.text);
                this.patPieces.push(pp);
                this.real_pieces+=1;
                if (pp.optional==false) {
                    this.req_pieces+=1;
                }
                if (pp.new==true) this.matchObj.new=true;

                //if we had a setToNextActive..
                //nextActive must be equal to the patResultingClass..
                //if so, count it as the target (to_s)
                if (setToNextActive!=null && this.patResultingClass && nOBJ.temp_nid==this.patResultingClass.temp_nid) {
                    setToNextActive.to_s = nOBJ.nid;
                    setToNextActive.to_text = nOBJ.text;
                    setToNextActive.toNOBJ=nOBJ;
                    setToNextActive=null;
                }
                lastActive=pp;
            }

            else if (brel.r==MEM.__max.temp_nid) {
                if (lastActive==null) {
                    this.statusObj.msg = "malformed pat - can't have max before an lastActive item on the stack";
                    this.status="bad";
                    return;
                }
                if (brel.to_type!=TO_TYPE_NUMBER) {
                    this.statusObj.msg = "malformed pat - max must be a number";
                    this.status="bad";
                    return;
                }
                
                //this.enforceMax(this.pat

                lastActive.max = brel.to_s;
            }
            else if (brel.r==MEM.__min.temp_nid) {
                if (lastActive==null) {
                    this.statusObj.msg = "malformed pat - can't have min before an lastActive item on the stack";
                    this.status="bad";
                    return;
                }
                if (brel.to_type!=TO_TYPE_NUMBER) {
                    if (this.pat.length==0) {
                        this.statusObj.msg = "malformed pat - min must be a number";
                        this.status="bad";
                        return;
                    }
                }

                //enforceMIn
                lastActive.min = brel.to_s;
            }

            //adding this optional piece makes life a bit more complicated
            //but it needs to be done.
            else if (brel.r==MEM.__optional.temp_nid) {
                if (lastActive==null) {
                    this.statusObj.msg = "malformed pat - can't have min before an lastActive item on the stack";
                    this.status="bad";
                    return;
                }
                if (brel.to_type!=TO_TYPE_ANNE) {
                    this.statusObj.msg = "malformed pat - active must be a TO_TYPE_ANNE"
                    this.status="bad";
                    return; 
                }
                var nOBJ = MEMG.getNeuronByTempNid(brel.t); //integer..
                var pp = new anne_x_neural_pat_piece (brel.type, nOBJ, null, null, true, nOBJ.text);
                this.patPieces.push(pp);
                if (pp.optional==false) {
                    this.req_pieces+=1;
                }
                this.real_pieces+=1;
                if (pp.new==true) this.matchObj.new=true;   //?
            }
        }
    }


    static applyNamePatFromPartials = function($nObj, $partials, $role) {
        var namePatRelon = MEMG.getRelonByTempR($nObj.outs, MEM.name_pattern.temp_nid);
        
        //just go with what came in on the phrase/role
        if (namePatRelon==null) {
            return $role.text;
        }
        
        var namePat = MEMG.getNeuronByTempNid(namePatRelon.t);
        var axpt_name = new anne_x_neural_pat(namePat, "name", $partials);
        console.log("%c(applyNamePatFromPartials) "+axpt_name.name, TALK.poolTitle);
        return axpt_name.name;
    }


    /**
     * 
     * GOING TO DEP THIS.
     * 
     * JUST MAKE A NEURON.  LATER STEPS CAN PROPERLY MAKE THE NEURON FOR THE L1 
     * WITH WHATEVER WE KNOW ABOUT THE NEURON.
     * OUR L1 handling in ANNE should not be a core part of the NLP process.
     * IT is essentially L2 with respect to parsing how we store the neuron in ANNE.
     * 
     * This is for multi part concepts like 
     * date, or person, location, or company... (and hundreds of others)
     * that require multiple parts to resolve properly.
     * 
     * We've confirmed what we have matches a certain pattern..
     * Let's look it up or make a new one.
     */
     lookupOrMake = async($role)=> {
        var partials = [];
        var rzpp1 = null;
        var rzpp2 = null;
        
        //resolve pool 1 priority (ie, YEAR for date)
        if (this.patResultingClass.rzpp1!=undefined) rzpp1 = this.patResultingClass.rzpp1;

        //resolve pool 2 priority (ie, DAY for date)
        if (this.patResultingClass.rzpp2!=undefined) rzpp2 = this.patResultingClass.rzpp2;
        var sortIt=false;

        for (var p=0; p<this.patPieces.length; p++) {
            var priority = 100000;
            var piece = this.patPieces[p];
            
            //this should be a be requirement on the neuron set.
            if (piece.relNOBJ==null) {
                console.log("%c to do, be requirement on matching neurons in ez pat", TALK.bgtitle);
            }
            else {
                if (rzpp1!=null && piece.relNOBJ!=null && piece.relNOBJ.nid==rzpp1) {
                    sortIt=true;
                    priority = 1;
                }
                else if (rzpp2!=null && piece.relNOBJ!=null && piece.relNOBJ.nid==rzpp2) {
                    sortIt=true;
                    priority = 2;
                }  

                partials.push({
                    r:piece.relNOBJ.temp_nid, 
                    reln_debug:piece.relNOBJ.text,
                    t:piece.toNOBJ.temp_nid, 
                    to_debug:piece.toNOBJ.text,
                    toNOBJ:piece.toNOBJ, 
                    priority:priority
                });
            }
        }
        
        if (sortIt==true) {
            partials.sort(function($a, $b) {
                if ($a.priority > $b.priority) return 1;
                else if ($a.priority < $b.priority) return -1;
                return 0;
            });
        }

        //if (this.patResultingClass!=null && this.patResultingClass.formatFromResolve!=undefined) {
            //this.matchObj.lookup_str = this.patResultingClass.formatFromResolve(this.matchObj);
            //console.log("%c**lookup** %c"+this.patResultingClass.text+" %c"+this.matchObj.lookup_str, TALK.bgblack, TALK.green, TALK.bggreen);
        //}
        console.log(this.matchObj);
        console.log(partials);

        //NOTE: if this.new==true don't bother looking up...
        var nObjs = [];
        console.log("%c lookup, make new if not found "+this.patResultingClass.text, TALK.bgred);
        var rezObj = MEMZ.ez_multiRelResolver(this.patResultingClass.temp_nid, partials);
        nObjs = rezObj.nObjs; 
        

        if (nObjs.length==0) {
            console.log("%MAKE NEW "+this.patResultingClass.text, TALK.poolTitle);

            //why is this applyNamePatFromPartials????????
            var newname = anne_x_neural_pat.applyNamePatFromPartials(this.patResultingClass, partials, $role);
            console.log(newname);

            this.finalNOBJ = await ANNE.BRAIN.ctp.makeLocalNeuron(newname, NTYPE_INST, this.patResultingClass);
            
            this.finalNOBJ.bNEW=true;
            this.finalNOBJ.__functInherit(this.patResultingClass, "anne_x_neuron_pat.lookupOrMake");
            ANNE.BRAIN.ctp.temp_new_neurons.push(this.finalNOBJ);

            //make relons from partials.
            //NOTE: we are only doing TYPE_BE here as they are concept-defining.
            //this means if you want a person to be unique using bday, it better
            //be a TYPE_BE relon.
            console.log("%c TODO: make relons in neuron from partials", TALK.bgred);
            $role.updateResolved(this.finalNOBJ);
        }

        //ie, "The movie King Kong was awesome!"
        //(which one?  what year?  There are multiple movies named King Kong)
        else {
            console.log("%RESOLVED "+this.patResultingClass.text, TALK.bgyellow);
            this.finalNOBJ = nObjs[0];
            $role.updateResolved(this.finalNOBJ);
            if (nObjs.length>1) {
                console.log("%UT OH, multiple resolutions", TALK.bgyellow);
                console.log(nObjs);
            }
        }
    }



    /**
     * sort by type TYPE_BRAIN0, TYPE_BRAIN1, TYPE_BRAIN2, etc.
     */
    brainSort=function($a, $b) {
        if ($a.type > $b.type) return 1;
        else if ($a.type < $b.type) return -1;
        return 0;
    }
}
















//has to be redone without js injection ...  :(




var ANNE_X = {};

/*** makeNewDSFromNeuron ************************************************
******************************************************** MR.Scatman ****/
ANNE_X.makeNewDSFromNeuron=async($srcNObj, $props=[])=> {
    var owner = USER._userNObj;

    var DDS_nObj = await ANNE.MEM.makeGhostNObj(owner, NTYPE_INST, "𝄐datasrc from "+$srcNObj.text, ANNE.MEM.ANNE_DDS, true);
    ANNE.MEM.makeTempRelon(DDS_nObj, TYPE_BE, ANNE.MEM.ANNE_DDS_SRC.nid, TO_TYPE_ANNE, $srcNObj.nid, true);
    
    //we're doing the default THIS CLASS params (not inherited) as the props.
    //we will also do MEM.name.
    if ($props.length==0) {

        //name prop for primary values in the data source..
        ANNE.MEM.makeTempRelon(DDS_nObj, TYPE_HAS, ANNE.MEM.DS_PROP.nid, TO_TYPE_ANNE, ANNE.MEM.name.nid, true);
    }

    return DDS_nObj;
}


/*** newFilterFromString ************************************************
******************************************************** MR.Scatman ****/
ANNE_X.newFilterFromString = async($ANNEDDS, $str, $relnPropNOBJ)=> {
    var filterNObj = await anne_x_filter.ads_filter_from_string($str, $relnPropNOBJ, true);
    if (filterNObj==null) return;
    if ($ANNEDDS!=null) {
        MEM.makeTempRelon($ANNEDDS, TYPE_BE, MEM.ANNE_DDS_FILTER.nid, TO_TYPE_ANNE, filterNObj.nid, true);
        if ($relnPropNOBJ!=null) {

            //this tells the filter what property the expression sequence applies to.
            MEM.makeTempRelon(filterNObj, TYPE_HAS, MEM.FILTER_PROP.nid, TO_TYPE_ANNE, $relnPropNOBJ.nid, true);
        }
    }
    return filterNObj;
}


/*** init ***************************************************************
******************************************************** MR.Scatman ****/
ANNE_X.init = function() {

    

    return;     //have to redo all of these neurons.. 

    //a DDS SRC can have numerous DDS filters.
    //these tell the DDS src how to narrow down the DDS SRC.
    MEM.ANNE_DDS_FILTER = MEMG.getNeuronByNid('7775569045052567774');
    MEM.FILTER_PROP = MEMG.getNeuronByNid('7775571650213519241');

    //the column property name (PARAM_NAME) you want to include in the DS.
    MEM.DS_PROP = MEMG.getNeuronByNid('7775586770145186367');

    //anne expression neurons are sequence neurons.
    //the DDS FILTER will supply some target info (target prop we are looking to apply the filter on..)
    //the expression neuron will be a SEQ neuoron that encodes the simple or complex expression.

    MEM.X_OPS = MEMG.getNeuronByNid('777559007866194998');
    MEM.XOP_AND = MEMG.getNeuronByNid('7775607609061661515');
    MEM.XOP_OR = MEMG.getNeuronByNid('7775609713660717343');
    MEM.XOP_equal = MEMG.getNeuronByNid('7775618696518156032');
    MEM.XOP_notequal = MEMG.getNeuronByNid('7775628753205217910');
    MEM.XOP_lessthan= MEMG.getNeuronByNid('7775667956029190870');
    MEM.XOP_lessthanequal= MEMG.getNeuronByNid('7775675477949841201');
    MEM.XOP_greaterthan= MEMG.getNeuronByNid('7775682071577396678');
    MEM.XOP_greaterthanequal = MEMG.getNeuronByNid('7775684994215435705');
    
    MEM.XOP_IF = MEMG.getNeuronByNid('7775702835389488107');
    MEM.XOP_ELSE = MEMG.getNeuronByNid('7775716167985707765');
    MEM.XOP_THEN = MEMG.getNeuronByNid('7775752784554929130');
    
    MEM.XOP_MULTIPLY = MEMG.getNeuronByNid('7775775934920269173');
    MEM.XOP_DIVIDE = MEMG.getNeuronByNid('7775780147339066069');
    MEM.XOP_MINUS = MEMG.getNeuronByNid('7775782301047634702');
    MEM.XOP_ADD = MEMG.getNeuronByNid('7775806504134889585');
    MEM.XOP_NOT = MEMG.getNeuronByNid('7775810666689980978');
    MEM.XOP_contains = MEMG.getNeuronByNid('7775870571475075997');
    
    MEM.XOP_EXISTS = MEMG.getNeuronByNid('7775881894819729095');

    MEM.XOP_PERCENT = MEMG.getNeuronByNid('7775899954353393269');
    MEM.XOP_MOD = MEMG.getNeuronByNid('7775907469918129159');
    
    ANNE_X.anne_map = [MEM.XOP_AND, MEM.XOP_OR, MEM.XOP_MULTIPLY, MEM.XOP_ADD, MEM.XOP_MINUS, 
        MEM.XOP_notequal, MEM.XOP_AND, MEM.XOP_OR, MEM.XOP_NOT, 
        MEM.XOP_equal, MEM.XOP_equal, MEM.XOP_ADD, MEM.XOP_MINUS, MEM.XOP_MULTIPLY, MEM.XOP_DIVIDE,
        MEM.XOP_greaterthan, MEM.XOP_lessthan, MEM.XOP_greaterthanequal, 
        MEM.XOP_lessthanequal, MEM.XOP_equal, MEM.XOP_notequal, MEM.XOP_percent, MEM.XOP_EXISTS];

}

class anne_x_tokenize {
    constructor($str) {
        this.operators = ['AND', 'OR', 'TIMES', 'PLUS', 'MINUS', 'NOT EQUAL', '&', '|', '!', '=', '==','+', '-', '*', '/', '>', '<', '>=', '<=', '==', '!=', '%', "EXISTS"];
        this.jso_map = ["&&", "||", "*", "+", "-", "!=", "&&", "||", "!", "==", "==", "+", "-", "*", "/", ">", "<", ">=", "<=", "==", "!=", "%", "!=undefined"];

        this.tokens = [];
        return this.tokenize_x($str);
    }

    opIndex = function ($str) {
        return this.operators.indexOf($str);
    }
    
    isNum=function($n) {
        return !isNaN(parseFloat($n)) && isFinite($n);
    }

    isAlpha=function($v) {
        const regExp = new RegExp(/^[a-z0-9]+$/i)
        return regExp.test($v);
    }

    isAlphaNum=function($v) {
        if (this.isNum($v)==true) return true;
        if (this.isAlpha($v)==true) return true;
        return false;        
    }

    tokenize_x($str) {
        var str = $str.trim()
        var s = "";
        var s2 = "";
        var snext = "";
        var current_token = null;

        for (var i = 0; i < str.length; i++) {
            s = str[i];
            s2 = s2+s;
            if (i<str.length-1) snext = str[i+1];
            else snext = "";

            //INSTANT FIND
            if (s.trim() == '(' || s.trim() == ')') {
                s.trim() == '(' ? this.tokens.push({ type: 'LPAREN' }) : this.tokens.push({ type: 'RPAREN' })
                current_token = null;
                s2 = "";
                continue;
            }
            else if (s.trim() == '[' || s.trim() == ']') {  //handle brackets the same
                s.trim() == '[' ? this.tokens.push({ type: 'LBRACKET' }) : this.tokens.push({ type: 'RBRACKET' })
                current_token = null;
                s2 = "";
                continue;
            }
            else if (s.trim() == '{' || s.trim() == '}') {  //handle brackets the same
                s.trim() == '{' ? this.tokens.push({ type: 'LCURLY' }) : this.tokens.push({ type: 'RCURLY' })
                current_token = null;
                s2 = "";
                continue;
            }
            var opI = this.opIndex(s.trim());
            if (opI!==-1) {
                this.tokens.push({ type: "OP", value: s.trim(), opI: opI, op_map:this.jso_map[opI]});
                current_token = null;
                s2 = "";

                //peek ahead for multi char op
                if (i<str.length-1) {
                    var opI2 = this.opIndex(str[i+1]);
                    if (opI2!==-1) {
                        this.tokens[this.tokens.length-1].value+=str[i+1];
                        i+=1;
                    }
                }
                continue;
            }
            
            if (s == ';' || s == '\n') {
                //this.tokens.push({ type: "EOL" });
                current_token = null;
                s2 = "";
                continue;
            }

            if (s==" ") {   
                this.tokens.push({ type: "SPACE" });
                current_token = null;
                s2 = "";
                continue;
            }
             
            //we are still parsing a token.
            if (current_token!=null) {
                if (current_token.type=="NUM") {
                    if (this.isNum(s)==true) {
                        current_token.value = s2;  //update the value
                        continue;   //it is more num
                    }

                    //switch it to alpha if there is ANY alpha in the string.
                    else if (this.isAlpha(s)==true) {
                        current_token.value = s2;  //update the value
                        current_token.type="ALPHA";
                        continue;   //it is more num
                    }
                }
                else if (current_token.type=="ALPHA") {
                    if (this.isAlphaNum(s)==true) {
                        current_token.value = s2;  //update the value

                        //this is actually an alpha operator! (AND.. etc)
                        var opiI = this.opIndex(s2);
                        if (this.isAlpha(s) && opiI!==-1) {
                            current_token.type="OP";
                            current_token.op_map=this.jso_map[opiI];
                            current_token.opI=opiI, 
                            current_token=null;
                        }
                        continue;   //it is more num
                    }
                    else {
                        if (s=="." && this.isAlphaNum(snext)==true) {
                            current_token.value = s2;
                            current_token.sub_type = "CODEPROP";
                            continue;
                        }

                        this.tokens.push({ type: "UNHANDLED", value: s.trim()});
                        current_token = null;
                    }
                }
            }

            //we only have s so far.
            else {
                if (this.isNum(s.trim())==true) {
                    this.tokens.push({ type: "NUM", value: s.trim()});
                    current_token = this.tokens [ this.tokens.length-1];
                }
                else if (this.isAlpha(s.trim())==true) {
                    this.tokens.push({ type: "ALPHA", value: s.trim()});
                    current_token = this.tokens [ this.tokens.length-1];
                }
            }
        }

        //this.tokens.push({ type: "EOF" });
        return this.tokens;
    }
}

class anne_x_parse {
    constructor() {}
    static parse = function ($str) {
        var tokens = new anne_x_tokenize($str);
        for (var t=0; t < tokens.length; t++) {
            var token = tokens[t];
            if (t>0 && t<tokens.length-1 && tokens[t-1].type=="ALPHA" && token.type=="SPACE" && tokens[t+1].type=="ALPHA") {
                tokens[t-1].value += " "+ tokens[t+1].value;
                tokens.splice(t, 2);
                t=t-1;
            }
            if (t>0 && (tokens[t-1].type=="SPACE")) {
                tokens.splice(t-1, 1);
                t=t-1;
                continue;
            }   
        }
        return tokens;
    }
}



class anne_x_filter {
    constructor(){}

    //we need to get this string into a legit format for
    //us to apply javascript (and test it)..

    //for neuron-based filters, we store the short-form text as the neuron name..
    //ie,  text=='bob' OR text=='suzy'   ==>  shortForm:  bob or suzy
    //all filtered datasources must be saved as relons...

    static ads_filter_from_string = async($filterStr, $relnPropNOBJ, $bMAKENEURON=false)=> {
        var tokens = anne_x_parse.parse($filterStr);
        console.log(tokens);
        
        if ($bMAKENEURON==true && $relnPropNOBJ!=null) {
            
            //this is a presumed "contains"..
            if (tokens.length==1 && tokens[0].type=="ALPHA") {
                var owner = null;
                if (typeof ANNE.USER !='undefined' && ANNE.USER!=null && ANNE.USER._userNObj!=null && ANNE.USER._userNObj.nid) owner = USER._userNObj;
                else owner = ANNE.MEM.ANNE;

                //this is async now. we will need a non-seed requesting one. 
                var filterNOBJ = await ANNE.MEM.makeGhostNObj(owner, NTYPE_INST, "contains "+$filterStr, ANNE.MEM.ANNE_DDS_FILTER, true);
                ANNE.MEM.makeTempRelon(filterNOBJ, TYPE_HAS, MEM.FILTER_PROP.nid, TO_TYPE_ANNE, $relnPropNOBJ.nid, true);
                ANNE.MEM.makeTempRelon(filterNOBJ, TYPE_HAS, ANNE.MEM._1.nid, TO_TYPE_ANNE, ANNE.MEM.XOP_contains.nid, true);
                ANNE.MEM.makeTempRelon(filterNOBJ, TYPE_HAS, ANNE.MEM._2.nid, TO_TYPE_STRING, tokens[0].value, true);
                
                //evals return 0 (hide) or non-zero (pass, don't hide record)
                filterNOBJ._jsEval = "(('{$VAL}'.indexOf_lc('"+$filterStr+"'))+1)";
            }

            //we're going to have to put in a litte more work.. lolz.
            else {
                //alert("unhandled expression parse to relons");
                //anne_x_filter.orderOfOps(tokens);
                var filterNOBJ = await ANNE.MEM.makeGhostNObj(owner, NTYPE_INST, "contains "+$filterStr, ANNE.MEM.ANNE_DDS_FILTER, true);
                MEM.makeTempRelon(filterNOBJ, TYPE_HAS, ANNE.MEM.FILTER_PROP.nid, TO_TYPE_ANNE, $relnPropNOBJ.nid, true);
                
                var jsevalstr = "";
                var rel_index = 0;
                for (var t=0; t < tokens.length; t++) {
                    var token = tokens[t];

                    if (token.type=="OP") {
                        var opI = token.opI;
                        var opNObj = ANNE_X.anne_map[opI];
                        var jsop = token.op_map;
                        MEM.makeTempRelon(filterNOBJ, TYPE_HAS, MEM['_'+rel_index++].nid, TO_TYPE_ANNE, opNObj.nid, true);
                        jsevalstr += " "+jsop+" ";
                    }

                    else if (token.type=="ALPHA") {
                        MEM.makeTempRelon(filterNOBJ, TYPE_HAS, MEM['_'+rel_index++].nid, TO_TYPE_ANNE, MEM.XOP_contains.nid, true);
                        MEM.makeTempRelon(filterNOBJ, TYPE_HAS, MEM['_'+rel_index++].nid, TO_TYPE_STRING, tokens.value, true);
                        jsevalstr += "(('{$VAL}'.indexOf_lc('"+token.value+"'))+1)";
                    }

                    else if (token.type=="CODEPROP") {
                        continue; //skip for now;
                    }

                    else if (token.type=="LPAREN") {
                        jsevalstr+="(";
                    }

                    else if (token.type=="RPAREN") {
                        jsevalstr+=")";
                    }
                    else if (token.type=="LBRACKET") {
                        jsevalstr+="[";
                    }

                    else if (token.type=="RBRACKET") {
                        jsevalstr+="]";
                    }
                    else if (token.type=="LCURLY") {
                        jsevalstr+="{";
                    }

                    else if (token.type=="RCURLY") {
                        jsevalstr+="}";
                    }
                }

                filterNOBJ._jsEval = jsevalstr;
                console.log(jsevalstr);
            }

            
            //we need store the expression tokens in seq so they are handled right.
            //MEM.makeTempRelon(filterNObj, TYPE_HAS, MEM.FILTER_PROP.nid, TO_TYPE_ANNE, $relnPropNOBJ.nid, true);
            return filterNOBJ;
        }

        //bob OR suzy 
        // =  text=="bob" OR text=="suzy"
        // =  (gender = male && country of origin == USA) OR (dogowner=true)
        //we need to apply the order of operations and test the filter against the dataObj.
        //the result should be a TRUE or FALSE
    }


    static orderOfOps = function($tokens) {
        
    }

    //this is a filter sequence neuron.. it should be sorted..
    //and the sequence relons printed out (only)
    static printRelonFilters=function($filters) {
        var output = [];
        console.log("anne_x.printRelonFilters");
        for (var f=0; f < $filters.length; f++) {
            MEM.sequence.set_value ($filters[f]);

            //for now we cheat and just take the filter.text.
            output.push($filters[f].text);
        }
        if (output.length) return output.join("<BR>");
        return "";
    }


    _makeRelonsFiltersFromTokens = function() {
        var paren_gen = 0;

        for (var t=0; t < this.tokens.length; t++) {
            var token = this.tokens[t];
            token.paren_gen = paren_gen;
            if (token.type=="LPAREN") {
                paren_gen+=1;
                continue;
            }
            else if (token.type=="RPAREN") {
                paren_gen-=1;
                continue;
            }
        }

        //we need to process the tokenized list into relon filters.
        //priority of order of operations must be applied.

    }

    //using the filter.. test the dataObj's dataProp against our filterStr.
    static testFiltersRuntime = function ($VAL, $filters) {
        for (var f=0; f < $filters.length; f++) {
            var filterNOBJ = $filters[f];
            var jsEval = filterNOBJ._jsEval;
            if (jsEval==undefined) continue;
            $VAL = $VAL.replace("'","");
            var jsEval2 = jsEval.replaceAll("{$VAL}", $VAL);
            var evalResult = eval(jsEval2);
            if (evalResult!=false && evalResult!=0) continue;
            return false;
        }
        return true;
    }
}