
const LSM_DIRECT = 1;
const LSM_RELATIVE = 2;

//location searches are not normal searches.
class loc_search_pool {
    constructor($str, $constituent_of_nOBJ, $altBeNid=null, $fireThresh=1) {
        this.str = $str;
        this.constituent_of_nOBJ = $constituent_of_nOBJ;
        this.altBeNid = $altBeNid;
        this.fireThresh = $fireThresh;
        this.search_matches = [];
        this.direct_full_matches = 0;
        this.direct_partial_matches = 0;
        this.relative_full_matches = 0;
        this.relative_partial_matches = 0;
    }


    /**
     * 
     * Add a match to the search match pool
     * @param {anne_neuron} $nObj 
     * @param {Number} $fire 
     * @param {int} $co - constiuent of match type (LSM_DIRECT, as requested  or LSM_RELATIVE, a near neighbor)
     * @param {int} $co_tnid - if LSM_RELATIVE, the tnid of the relative constiuent_of we found.
     * @returns {loc_search_match}
     */
    addMatch = function($nObj, $fire=1, $co=0, $co_tnid=null) {
        if ($co==LSM_DIRECT) {
            if ($fire==1) this.direct_full_matches+=1;
            else this.direct_partial_matches+=1;
        }
        else {
            $fire = $fire*.9;
            if ($fire==1) this.relative_full_matches+=1;
            else this.relative_partial_matches+=1;
        }

        var lsm  = new loc_search_match($nObj, $fire, $co, $co_tnid);
        this.search_matches.push(lsm);
        return lsm;
    }



    /**
     * 
     * @param {*} $bSORT 
     * @returns 
     */
    getMatches = function($bSORT=true) {
        if (this.search_matches.length==0) return this.search_matches;
        if ($bSORT==true) {
            this.search_matches.sort(function($a, $b) {
                if ($a.fire>$b.fire) return -1;
                else if ($a.fire<$b.fire) return 1;
                return 0;
            });
        }
        return this.search_matches; 
    }
}


/**
 * 
 */
class loc_search_match {
    constructor($nObj, $fire=1, $co=0, $co_tnid=null) {
        this.nObj = $nObj;
        this.fire = $fire;
        this.co = $co;
        this.co_tnid = $co_tnid;
    }
}


class anne_chase extends anne_singleton {
    constructor(){
        super(false);
	}

    /**
     */ 
     getAllPreloads = async()=> {
        await ANNE.API.getAllPreloads();
        return ANNE.MEM.getAllOfOrigin(ORIGIN_NODE_PRELOAD);
    } 

    
    /**
     * we want be class and text info for anything this nObj is connecting to.
     * @param {*} $nObj 
     * @returns 
     */
    oneGenQueryable=async($nObj, $addTypes=[])=>{
        var queryable_nids = [];
        
        if ($nObj.bes.length==0 || $nObj.text=="") queryable_nids.push($nObj.nid);

        for (var r=0; r < $nObj.outs.length; r++) {
            if ($nObj.outs[r].to_type!=TO_TYPE_ANNE) continue;
            var toNOBJ = MEMG.getNeuronByNid($nObj.outs[r].to_s);
            if (toNOBJ==null) continue;
            if (toNOBJ.ghost==true) continue;
            if (toNOBJ.text=="" || toNOBJ.bes.length==0) {
                if (queryable_nids.indexOf(toNOBJ.nid)==-1) queryable_nids.push(toNOBJ.nid);
            }
        }

        var types = [0,1];
        if ($addTypes.length>0) {
            for (var t=0; t < $addTypes.length; t++) {
                if (types.indexOf($addTypes[t]==-1)) types.push($addTypes[t]);
            }
        }

        //we don't want to pull all in names -- here..
        //for classes this could be crazy huge.. like..
        //millions, for "person" class.
        if ($nObj.ntype==NTYPE_INST) {
            for (var r=0; r < $nObj.ins.length; r++) {
                if ($nObj.ins[r].to_type!=TO_TYPE_ANNE) continue;
                var fromNOBJ = MEMG.getNeuronByNid($nObj.ins[r].from_nid);
                if (fromNOBJ==null) continue;
                if (fromNOBJ.ghost==true) continue;
                if (fromNOBJ.text=="" || fromNOBJ.bes.length==0) {
					if (queryable_nids.indexOf(fromNOBJ.nid)==-1) queryable_nids.push(fromNOBJ.nid);
                }
            }
        }

        if (queryable_nids.length>0) {
            await ANNE.CHASE.lookupMinList(queryable_nids, types, "onegenQ", true);
        }
    }


    /**
     * Get all outs not of types
     * @param {*} $nObj 
     * @param {*} $bMUSTPULL 
     * @param {*} $type 
     * @param {*} $debugJsonEl 
     * @returns 
     */
    queryForOuts_not = async($nObj, $bMUSTPULL=false, $types=[], $debugJsonEl="", $callback=null)=>{
        if (($nObj instanceof anne_neuron) == false) {
            console.warn("FIX UR CALL --> needs an object for nObj");
            return;
        }
        if ($nObj.ghost==true || $nObj.temp==true) {
            console.warn("no ghosts or temps found in the ANNODE layer");
            return $nObj.outs;
        }

        var nid = $nObj.nid;
        
        if ($types.length) {
            var aRelons = await ANNE.API.queryForOuts_not(nid, $bMUSTPULL, $types);
        }
        else {
            //alert("optimizer skip query");
            console.log("%c queryForOuts %c optimizer skip query", ANNE.TALK.bgredbig, ANNE.TALK.bgpink);
        }
        
        if ($debugJsonEl!="") ANNE.A.setJSON($debugJsonEl, aRelons);

        if($callback!=null) $callback($nObj.outs);
        return $nObj.outs;
    }


    /**
     * 
     * @param {*} $nObj 
     * @returns 
     */
    queryMinBem = async($nObj) =>{
        if ($nObj.ghost==true || $nObj.temp==true) {
            console.warn("no ghosts or temps found in the ANNODE layer");
            return $nObj.outs;
        }
        return await ANNE.API.queryForOuts_BEM($nObj.nid, true, [0,1]);
        
    }

    /**
     * note:  this does no filtering for garden/worldview. 
     * @param {anne_neuron}  - Neuron Object to query
     * @param {Boolean} - Require pull despite flags
     * @param {Array}   - Types (dimensions) of relons desired
     * @param {string}  - Optional debug el for json printing
     * @author Mr.Scatman                                 
    */
    queryForOuts = async($nObj, $bMUSTPULL=false, $type=[], $debugJsonEl="")=>{
        if (($nObj instanceof anne_neuron) == false) {
            console.warn("FIX UR CALL --> needs an object for nObj");
            return;
        }
        if ($nObj.ghost==true || $nObj.temp==true) {
            console.warn("no ghosts or temps found in the ANNODE layer");
            return $nObj.outs;
        }

        var nid = $nObj.nid;
        if ($type.length) {
            var aRelons = await ANNE.API.queryForOuts(nid, $bMUSTPULL, $type);
        }
        else {
            //alert("optimizer skip query");
            console.log("%c queryForOuts %c optimizer skip query", ANNE.TALK.bgredbig, ANNE.TALK.bgpink);
        }
        
        if ($debugJsonEl!="") ANNE.A.setJSON($debugJsonEl, aRelons);
        return $nObj.outs;
    }


    /**
     * 
     * @param {*} $nObj 
     * @param {*} $bMUSTPULL 
     * @param {*} $type 
     * @param {*} $debugJsonEl 
     * @returns 
     */
    queryForOutsAndIns = async($nObj, $bMUSTPULL=false, $type=[], $debugJsonEl="")=>{
        if (($nObj instanceof anne_neuron) == false) {
            console.warn("FIX UR CALL --> needs an object for nObj");
            return;
        }
        if ($nObj.ghost==true || $nObj.temp==true) {
            console.warn("no ghosts or temps found in the ANNODE layer");
            return $nObj.outs;
            return;
        }
        var nid = $nObj.nid;
        
        //PROCEED ONLY IF WE HAVE SOMETHING NEW TO CHASE.
        var aRelons = await ANNE.API.queryForOutsAndIns(nid, $bMUSTPULL, $type);
        
        if ($debugJsonEl!="") ANNE.A.setJSON($debugJsonEl, aRelons);
        return $nObj.outs;
    }


    /**
     * 
     * @param {*} $nids 
     * @param {*} $bMUSTPULL 
     * @param {*} $types 
     * @returns 
     */
    queryForOutsAndInsFromNids = async($nids, $bMUSTPULL=false, $types=[])=>{
        var nids = [];
        for (var n=0; n < $nids.length; n++) {
            if (($nObjs[n] instanceof anne_neuron) == false) {
                console.warn("FIX UR CALL --> needs an object for nObj");
                continue;
            }
            if ($nObjs[n].ghost==true || $nObjs[n].temp==true) {
                console.warn("no ghosts or temps found in the ANNODE layer");
                continue;
            }
        }
        
        //PROCEED ONLY IF WE HAVE SOMETHING NEW TO CHASE.
        return await ANNE.API.queryForOutsAndInsFroms(nids, $bMUSTPULL, $types);
    }


    /**
     * 
     * @param {*} $nObjs 
     * @param {*} $bMUSTPULL 
     * @param {*} $types 
     * @returns 
     */
    queryForOutsAndInsFroms = async($nObjs, $bMUSTPULL=false, $types=[])=>{
        var nids = [];
        for (var n=0; n < $nObjs.length; n++) {
            if (($nObjs[n] instanceof anne_neuron) == false) {
                console.warn("FIX UR CALL --> needs an object for nObj");
                continue;
            }
            if ($nObjs[n].ghost==true || $nObjs[n].temp==true) {
                console.warn("no ghosts or temps found in the ANNODE layer");
                continue;
            }
            nids.push($nObjs[n].nid);
        }
        
        //PROCEED ONLY IF WE HAVE SOMETHING NEW TO CHASE.
        return await ANNE.API.queryForOutsAndInsFroms(nids, $bMUSTPULL, $types);
    }



    /**
     * Chase outs and ins including the BEM line of the nObj.nid  (its parent line outs)
     * @param {*} $nObj 
     * @param {*} $bMUSTPULL 
     * @param {*} $type 
     * @param {*} $debugJsonEl 
     * @returns 
     */
    queryForOutsAndIns_BEM = async($nObj, $bMUSTPULL=false, $type=[], $debugJsonEl="")=>{
        if (($nObj instanceof anne_neuron) == false) {
            console.warn("FIX UR CALL --> needs an object for nObj");
            return;
        }
        if ($nObj.ghost==true || $nObj.temp==true) {
            console.warn("no ghosts or temps found in the ANNODE layer");
            return $nObj.outs;
            return;
        }
        var nid = $nObj.nid;
        if ($type.length) {
            var aRelons = await ANNE.API.queryForOutsAndIns_BEM(nid, $bMUSTPULL, $type);
        }
        else {
            console.log("%c queryForOutsAndIns_BEM %c optimizer skip query", ANNE.TALK.bgredbig, ANNE.TALK.bgpink);
        }
        if ($debugJsonEl!="") ANNE.A.setJSON($debugJsonEl, aRelons);
        return $nObj.outs;
    }


    /**
     * 
     * @param {*} $nObj 
     * @param {*} $bMUSTPULL 
     * @param {*} $type 
     * @param {*} $debugJsonEl 
     * @returns 
     */
    queryForOuts_BEM = async($nObj, $bMUSTPULL=false, $type=[], $debugJsonEl="")=>{
        if (($nObj instanceof anne_neuron) == false) {
            alert("FIX UR CALL --> needs an object for nObj");
            return;
        }
        if ($nObj.ghost==true || $nObj.temp==true) {
            console.warn("no ghosts or temps found in the ANNODE layer");
            return $nObj.outs;
            return;
        }
        var nid = $nObj.nid;
        if ($type.length) {
            var aRelons = await ANNE.API.queryForOuts_BEM(nid, $bMUSTPULL, $type);
        }
        else {
            console.log("%c queryForOuts_BEM %c optimizer skip query", ANNE.TALK.bgredbig, ANNE.TALK.bgpink);
        }
        if ($debugJsonEl!="") ANNE.A.setJSON($debugJsonEl, aRelons);
        return $nObj.outs;
    }


    /**
     * Get a global value.
     * Numbers, Dates, and Keywords are globals and created by the nodes,
     * sponsored by identities.  Get a global from local memory first.  If it exists
     * that is it.  If not, go check the node for the global.
     * If it exists, that's it.
     * If not, the user can trigger the creation of the global by using a 
     * globalNeuron TX type.  This will pay the node operator to create a global.
     * 
     * Locations are too but they are handled a bit differently.
     * 
     * @author Mr.Scatman             
    */
    getGlobal=async($classNOBJ, $val)=>{
        var nObj = ANNE.MEMG.getGlobalNeuron($classNOBJ, $val);
        if (nObj!=null) {
			
			if (nObj.ghost==false) {
	            console.warn("*** WE ALREADY HAVE THIS GLOBAL IN LOCAL MEM! **");
	            return {status:"good-but", result:nObj};
	        }
			else {
				console.warn("*** WE ALREADY HAVE THIS GLOBAL GHOST IN LOCAL MEM! **");
	            return {status:"good-but", result:nObj};
			}
		}
        
        var response = await ANNE.API.queryForGlobal($classNOBJ.nid, $val);
        
        if (response.status=="good" && response.message=="available") {
            //console.log(response);
            return response;
        }

        //sb.b good but.
        else if (response.status=="good-but" && response.result) {
            var rObj = response.result;
            if (rObj!=null) {
                ANNE.TALK.printRelon(rObj);
                await ANNE.MEMP.processNeuronsFromRelons([rObj], true, true, "relons", "CHASE.getGlobal");
                var nObj = MEMG.getNeuronByNid(rObj.from_nid);
                response.result = nObj;
                return response;
            } 
            else {
                console.log(response);
                return [];
            }
        }
        else return [];
    }


    /**
     * 
     * @param {*} $nid 
     */
    loadChildrenFromIDB=async($nids)=> {
		console.time("loadChildrenFromIDB");
		var aRelonsToProcess = await ANNE.DB.loadAll("relons");
		var aRelons = [];
		for (var r=0; r < aRelonsToProcess.length; r++) {
			if ($nids.indexOf(aRelonsToProcess[r].be)==-1) {
				continue;
			}
			aRelons.push(aRelonsToProcess[r]);
		}
        if (aRelons.length) await ANNE.MEMP.processNeuronsFromRelons(aRelons, true, true, "relons", "CHASE.loadChildrenFromIDB");
		console.timeEnd("loadChildrenFromIDB");
        return aRelons;
    }

    
    /**
     * With a list of nids...
     * load their BEM outs from local IDB into RAM.
     * @param {*} $nids 
     * @returns 
     */
    loadLocalBEMs=async($nids)=>{
        var nids = [];
        for (var n=0; n < $nids.length; n++) {
            var nid = $nids[n];
            if (nids.indexOf(nid)!==-1) continue;
            nids.push(nid);
            var nObj = MEMG.getNeuronByNid(nid);
            var parent = nObj.getParent();
            while(parent!=null) {
                if (nids.indexOf(parent.nid)!=-1) break;
                nids.push(parent.nid);
                parent = parent.getParent();
            }
        }
		var aRelons = await ANNE.IDB.loadIDBNidOuts(nids);
	    console.log("%c loaded "+aRelons.length+" from localBEMs", ANNE.TALK.bggreen);
	}

    /**
     * 
     * @param {*} $nObj 
     * @returns 
     */
    loadLocalBEM=async($nObj)=>{
        var nids = [$nObj.nid];
        var parent = $nObj.getParent();
        while(parent!=null) {
            nids.push(parent.nid);
            parent = parent.getParent();
        }
        return await ANNE.IDB.loadIDBNidOuts(nids);
    }



    /**
     * 
     * THIS DOES NOT SEARCH LOCAL MEMORY FIRST.
     * USE A SPECIALIZED FUNCTION FIRST like MEMG.getOrMakeIfNewPerson.
     * 
     * This returns multiple results in nObjs.
     * no array value means we do not know of this neuron.
     * 
     * Note: we could do a multiple relon version of this too (if needed)
     * @param {*} $classNObj 
     * @param {*} $val 
     * @returns 
     */
    queryFoNeuronByClassAndRelon=async($classNObj, $val, $reln_nid, $to_type, $to_s)=>{
        var response = await ANNE.API.queryFoNeuronByClassAndRelon($classNObj.nid, $val, $reln_nid, $to_type, $to_s);
        if (response.status=="good" && response.message=="available") {
            return [];
        }
        else if (response.status=="good-but" && response.results!=undefined) {
			var relons = response.result;
			if (Array.isArray(relons)==false || relons.length==0) {
				relons = response.results;
			}
            if (relons!=undefined && relons!=null && relons.length) {
                relons = await ANNE.MEMP.processNeuronsFromRelons(relons, true, true, "relons", "CHASE.queryFoNeuronByClassAndRelon");
                if (ANNE.SETTINGS.debug()==true) ANNE.TALK.printRelons(relons);
                var matches = [];
                var unique_tnids = [];
                for (var r=0; r < relons.length; r++) {
                    var nObj = MEMG.getOrMakeIfNew(relons[r].from_nid);
                    if (unique_tnids.indexOf(nObj.temp_nid)==-1) {
                        matches.push(nObj);
                        unique_tnids.push(nObj.temp_nid);
                    }
                }
                return matches;
            } 
            else {
                //"helper"
                //"matching_text_nids" -- nids that matched w/ no matching reln/to_s
                //"contradictors" -- nids that contradicted
                if (response.matching_text_nids!=undefined) {
                    if (response.contradictions!=undefined) {
                        var matches = [];
                        for (var m=0; response.matching_text_nids.length; m++) {
                            var match_nid = response.matching_text_nids[m];
                            if (response.contradictions.indexOf(match_nid)!=-1) continue;
                            var nOBJ = MEMG.getOrMakeIfNew(match_nid);
                            matches.push(nOBJ);
                        }
                        return matches;
                    }
                }

				console.log(response);
                return [];              //matches exist but contradict, we have to create a new thing.
            }
        }
    }


    /**
     * Used as a psuedo optimization for searching by ALT BE CLASS first, then text.
     * Classes like "occupation" are not instantiated.  This would be a large classication error.
     * We simply create a new ec_living "thing" and laterally tag it with an alt TYPE_BE of "occupation".
     * To find an occupation we can snag lateral TYPE_BE occupation relons that are ec_things...
     * THEN we can validate the TEXT string (more expensive) instead of forcing the annode to check
     * against all text strings for all things.
     * @param {*} $classNObj 
     * @param {*} $text 
     * @param {*} $alt_be_reln 
     * @param {*} $alt_be_to 
     * @returns 
     */
    queryFoNeuronByAltBeTexAndBe=async($classNObj, $text, $alt_be_reln, $alt_be_to)=>{
        var response = await ANNE.API.queryFoNeuronByAltBeTexAndBe($classNObj.nid, $text, $alt_be_reln, $alt_be_to);
        if (response.status=="good" && response.message=="available") {
            console.log(response);
            return [];
        }
        
        else if (response.status=="good-but" && response.results!=undefined) {
            
            var relons = response.results;
            if (relons!=null && relons.length) {
                relons = await ANNE.MEMP.processNeuronsFromRelons(relons, true, true, "relons", "queryFoNeuronByAltBeTexAndBe");
                ANNE.TALK.printRelons(relons);
                matches = [];
                for (var r=0; r < relons.length; r++) {
                    var nObj = MEMG.getNeuronByTempNid(relons[r].f);
                    matches.push(nObj);
                }
                return matches;
            } 

            else {
                console.log(response);
                return [];
            }
        }

        return [];
    }


    
    /**
     * Look for a subanndle under a givne parent anndle/subanndle.
     * @param {anne_neuron} $parentNOBJ 
     * @param {String} $val 
     * @returns 
     */
    getSubAnndle=async($parentNOBJ, $val)=>{
        console.log("(CHASE.getSubAnndle) for class:"+$parentNOBJ.text+" value:"+$val);
        var nObj = MEMG.getSubAnndle($parentNOBJ, $val);
        if (nObj!=null) {
            console.warn("*** WE ALREADY HAVE THIS GLOBAL IN LOCAL MEM! **");
            return nObj;
        }
        var response = await ANNE.API.getSubAnndle($classNObj.nid, $val);
        if (response.status=="good" && response.message=="available") {
            console.log(response);
            return response;
        }
        else {
            var rObj = response.result;
            if (rObj!=null) {
                ANNE.TALK.printRelon(rObj);
                await ANNE.MEMP.processNeuronsFromRelons([rObj], true, true, "relons", "getSubAnndle");
                var nObj = MEMG.getNeuronByNid(rObj.from_nid);
                response.result = nObj;
                return response;
            } 
            else {
                console.log(response);
                return response;
            }
        }
    }

    
    /**
     * Get open hanais that pertain to me.
     * note: LNOTS (local notifications) will be available from the ANNODES that have handshoot
     * w/ this userID within the last 7 days.
     * 
     * This function can be used to get hanais that are either outside of this range or were
     * not captured (at the time they came in) by an the present ANNODE that the user is handshook with (using).
     * 
     * @param {String} - Nid long string
     * @param {int} - block height min
     * @author Mr.Scatman             
    */
    queryForHanai = async($nid, $height=0)=> {
        return await ANNE.API.queryForHanai($nid, $height);
    }

    //see use in dossier
    getPendingHanaiResponse=async($txid)=> {
        var rez =  await ANNE.API.queryForHanaiResponseTo($txid);
        console.log("(getPendingHanaiResponse)",rez);
        return rez;
    }


    /**
     * Get granparent (one), uncles/aunties, brother/sister and cousin locations relative to a location neuron.
     * Any of these can match when we are trying to find a location in memory.
     * 
     * We want to avoid creating a Statue of Liberty site in NYC, then another one in the Bourough of Manhattan.
     * 
     * If we already store it in NYC then we need not create another Statue of Liberty.  We can always alter
     * the parent location (Statue of Liberty->[TYPE_BE]->constituent_of->NYC) to be more refined.
     * 
     * NOTE: This is intended to be processed by the getGlobalLocation_localonly which looks at the ins
     * irrespective of the FROM and takes the FROM as a legit potential match to string match against.
     * 
     * @param {*} $parentLocation 
     * @returns 
     */
    getSpotlightConstituentIns = function($parentLocation) {
        var ins_spotlight = [];

        for (var i=0; i <  $parentLocation.ins.length; i++) {
            var rel =  $parentLocation.ins[i];
            if (rel.r == ANNE.MEM.constituent_of.temp_nid) {

                //sibblings to the level we are testing a new alt location (brothers/sisters)
                ins_spotlight.push(rel);

                //children locations under the sibblings of our parent location.
                //ie, newphews/nieces.
                var fromNOBJ = MEMG.getNeuronByTempNid(rel.f);
                ins_spotlight = ins_spotlight.concat(fromNOBJ.ins);
            }
        }

        //grandparent location
        var co = $parentLocation["be_"+ANNE.MEM.constituent_of.temp_nid];
        if (co!=undefined) {
            var grandParentLocation = MEMG.getNeuronByTempNid(co);
            var gpco = MEMG.getBeRelon(grandParentLocation.outs, MEM.constituent_of.temp_nid);
            if (gpco!=null) ins_spotlight.push(gpco); //this gets gramps in the array.
            for (var i=0; i <  grandParentLocation.ins.length; i++) {
                var rel =  grandParentLocation.ins[i];
                if (rel.r == ANNE.MEM.constituent_of.temp_nid) {

                    //sibblings to our parent location. (uncles/aunties)
                    ins_spotlight.push(rel);
                    
                    //children of our uncles/aunties  (cousins)
                    var fromNOBJ = MEMG.getNeuronByTempNid(rel.f);
                    var co2 = MEMG.getBeRelons(fromNOBJ.ins, ANNE.MEM.constituent_of.temp_nid);
                    ins_spotlight = ins_spotlight.concat(co2);
                }                
            }
        }

        return ins_spotlight;
    }


    /**
     * Search local mem for a location within a consitiuent of a specific altBe.
     * 
     * @param {loc_search_pool} $lsp 
     */
    getGlobalLocation_localonly=function($lsp) {
        var altBeNOBJ = MEMG.getNeuronByNid($lsp.altBeNid);
        var ins = [];
        var lower = $lsp.str.toLowerCase();

        //we should be a bit more flexible on constituent location and look up and down their constituent tree.
        //ie,    NYC -> up to New York..  NYC -> down to the Bronx, Brooklyn, Manhattan, Queens and Staten Island.
        ins = this.getSpotlightConstituentIns($lsp.constituent_of_nOBJ);
        for (var r=0; r < ins.length; r++) {
            var rel = ins[r];
        
            var fromLocNOBJ = MEMG.getNeuronByTempNid(rel.f);   //this nObj is a constituent of our location.
            if (ANNE.isANNE(fromLocNOBJ.owned_by)==false && fromLocNOBJ.owned_by!=USER._userNObj.nid) continue;

            //this is the proper alt-be  (ie, a site in NEW YORK CITY)
            if (altBeNOBJ!=null && MEMG.isAltBeOf(fromLocNOBJ.outs, altBeNOBJ.temp_nid)==false) continue;

            //see if the text matches or is close!
            var fire = ANNE.UTILS.similarity(fromLocNOBJ.text.toLowerCase(), lower, 1000);

            //look at what we want to do here with near hits..
            //if we reutrn a match object we have to handle that in all cases down stream from
            //our search.
            if (fire >= $lsp.fire) {
                var co = 0;
                var co_tnid = null;
                if (rel.t==$lsp.constituent_of_nOBJ.temp_nid) co=$lsp.LSM_DIRECT;
                else {
                    co=$lsp.LSM_RELATIVE;
                    co_tnid = rel.t;
                }
                $lsp.addMatch(fromLocNOBJ, fire, co, co_tnid);
            }
            else if (fire>.8) console.log("(close) partial "+fromLocNOBJ.text+" fire:"+fire);
        }
    }


    /**
     * Don't use this if you can use the location search with a known constituent parent.
     * This search can return more than one result because there is often locations
     * that are named the same.  (ie, multiple cities in the US are named the same thing)
     * 
     * We use this function when we don't have anything else to go on.
     * @param {*} $location 
     * @returns 
     */
    getNamedLocation=async($location) => {
        var matches = [];
        var matches_tnids = [];
        var location = $location.toLowerCase();
        for (var c = 0; c < MEM.location.children.length; c++) {
            var nObjLoc = MEMG.getNeuronByTempNid(MEM.location.children[c]);
            if (nObjLoc.text.toLowerCase()==location) {
                if (matches_tnids.indexOf(nObjLoc.temp_nid)==-1) {
                    matches.push(nObjLoc);
                    matches_tnids.push(nObjLoc.temp_nid);
                }
            }
            else if (nObjLoc.aka_lower==undefined) continue;
            else if (nObjLoc.aka_lower.indexOf(location)!=-1) {
                if (matches_tnids.indexOf(nObjLoc.temp_nid)==-1) {
                    matches.push(nObjLoc);
                    matches_tnids.push(nObjLoc.temp_nid);
                }
            }
        }
        if (matches.length>0) return matches;

        //search for it in the hypergraph.
        var rels = await ANNE.API.queryForNamedLocation($location);

        if (rels.length==0) return [];

        //search the returned relons.
        for (var r = 0; r < rels.length; r++) {
            var nObjLoc = MEMG.getNeuronByTempNid(rels[r].from_nid);
            if (nObjLoc.text.toLowerCase()==location) {
                if (matches_tnids.indexOf(nObjLoc.temp_nid)==-1) {
                    matches.push(nObjLoc);
                    matches_tnids.push(nObjLoc.temp_nid);
                }
            }
            else if (nObjLoc.aka_lower==undefined) continue;
            else if (nObjLoc.aka_lower.indexOf(location)!=-1) {
                if (matches_tnids.indexOf(nObjLoc.temp_nid)==-1) {
                    matches.push(nObjLoc);
                    matches_tnids.push(nObjLoc.temp_nid);
                }
            }
        }
        if (matches.length>0) return matches;
    }


    /**
     * Get a location in a constituent location.
     * 
     * If altbe is a site or geo site we want to be a bit more flexible on the constituent location.
     * We should, technically, allow for finding the Statue of Liberty in New York... even if it is in New York City.
     * 
     * @param {Array}
     * @returns {location_search_pool}
     * @author Mr.Scatman             
    */
    getGlobalLocation=async($location, $constituent_locationNOBJ, $altBeNid, $exactThresh=1)=>{
        var lsp = new loc_search_pool($location, $constituent_locationNOBJ, $altBeNid, $exactThresh);
        this.getGlobalLocation_localonly(lsp);
        var matches = lsp.getMatches();
        var nObjs = [];
        if (matches.length>0) {
            for (var m=0; m < matches.length; m++) {
                var match = matches[n];
                var nObj = match.nObj;
                if (nObj.owned_by==ANNE.USER._userNObj.nid || ANNE.isANNE(nObj.owned_by)==true) {
                    if (match.co==LSM_DIRECT) nObjs.push(nObj);
                }
            }
        }
        
        if (nObjs.length>0) return lsp; //bail without a query of ANNE.

        var response = await ANNE.API.queryForLocation($location, $constituent_locationNOBJ.nid, $altBeNid);
        console.log("response:"+response);
        if (response.status=="good" && response.message=="available") return lsp;
        else {
            var rObj = response.result;
            if (rObj!=null) {
                ANNE.TALK.printRelon(rObj);
                await ANNE.MEMP.processNeuronsFromRelons([rObj], true, true, "relons", "getGlobalLocation");
                var nObj = MEMG.getNeuronByNid(rObj.from_nid);
                lsp.addMatch(nObj, 1, LSM_DIRECT);
                return lsp;
            }
            else {
                alert("fix this.. it should have returned back the getGlobalLocation exiting value");
            }
        }
        return lsp;
    }
    


    /**
     * Make a global location
     * 
     * @param {string} - the name of the location
     * @param {string} - The nid longstring of a KNOWN location that
     * this new location will be a constituent of.
     * 
     * @author Mr.Scatman             
     */
    makeGlobalLocation=async($location, $constituent_locationNOBJ, $bSINGLESEND=true, $gui_extras=null)=>{
        if ($location==undefined || $location==null || $location=="") return;
        var msg = "<span class='q-part rel-reln'>make global location</span> <span class='q-part rel-to'>"+$location+"</span>";
        var rObj = ANNE.MEM.makeTempRelon($constituent_locationNOBJ, TYPE_SYS_MSG, MEM.constituent.nid, TO_TYPE_STRING, $location, true);
        rObj.msg = msg;
        if ($gui_extras) rObj.gui_extras=$gui_extras;
        var cmdObj = await ANNE.SENDQ.addRelonToSendQ(rObj, $bSINGLESEND, ANNE.TXB.TXTYPE_MAKE_GLOBAL_LOCATION);
        if (cmdObj!=null && $gui_extras!=null && $gui_extras.znid!=undefined) cmdObj.znid=$gui_extras.znid;
        return cmdObj;
    }


    //we have a pending globalLocation that requires another location be made first.
    makePendingGlobalLocation=async($location, $constituent_locationNOBJ, $bSINGLESEND=true, $gui_extras=null)=>{
        if ($location==undefined || $location==null || $location=="") return;
        //var rObj = MEM.makeTempRelon($constituent_locationNOBJ, TYPE_SYS_MSG, MEM.constituent.nid, TO_TYPE_STRING, $location, true);
        var msg = "<span class='q-part rel-reln'>make global location</span> <span class='q-part rel-to'>"+$location+"</span>";
        var rObj = {type:TYPE_SYS_MSG, from_nid:null, pending_from:"new_"+$constituent_locationNOBJ.nid, reln:MEM.constituent.nid, to_type:TO_TYPE_STRING, to_s:$location}
        rObj.msg = msg;
        if ($gui_extras) rObj.gui_extras=$gui_extras;
        await ANNE.SENDQ.pendingDynamicVarRelon(rObj, true, false, false, ANNE.TXB.TXTYPE_MAKE_GLOBAL_LOCATION, $gui_extras, false);
        return rObj;
    }

    
    /**
     * Make a global neuron via ANNODE request.
     * NOTE: globals are just classes of neurons.  They could be 
     * created with keyless request (no special handling needed)
     * HOWEVER, we want to handle certain classes with extra care to 
     * as we intend to get them into the ANNE worldview. (some of them)
     * 
     * There DOES NOT NEED TO BE EXPANSION to annode handling to handle 
     * new classes.  They can be created and simply made.
     * 
     * @param {anne_neuron} - the class of global neuron we are making
     * @param {string} - value of the global
     * @param {boolean} - send now?
     * @param {Object} - the guiextras return object we use to do stuffs w/.
     * @author Mr.Scatman             
     */
    makeGlobal=async($classNObj, $val, $bSINGLESEND=true, $gui_extras=null)=>{
        var cmd_nid=0;
        if ($val==undefined || $val==null || $val=="") return;
        var to_type = TO_TYPE_STRING;
        var msg = "";
        //we actually are making integers (not just numbers)...
        if ($classNObj.temp_nid==MEM.integer.temp_nid) {
            cmd_nid = ANNE.MEM.MAKE_INTEGER.nid;
            to_type = TO_TYPE_NUMBER;
            msg = "<span class='q-part rel-reln'>make global number</span>";
        }
        else if ($classNObj.temp_nid==MEM.anndle.temp_nid) {
            cmd_nid = ANNE.MEM.MAKE_ANNDLE.nid;
            to_type = TO_TYPE_STRING;
            msg = "<span class='q-part rel-reln'>make anndle</span>";
        }
        //this will create in THEIR worldview.. (not ANNE's)
        else if ($classNObj.temp_nid==MEM.ec_action.temp_nid && $classNObj.ntype==NTYPE_CLASS) {
            cmd_nid = ANNE.MEM.MAKE_ACTION.nid;
            to_type = TO_TYPE_STRING;
            msg = "<span class='q-part rel-reln'>make global action (verb)</span>";
        }
        //YYYY/MM/DD
        else if ($classNObj.temp_nid==MEM.date_gregorian.temp_nid) {
            cmd_nid = ANNE.MEM.MAKE_DATE.nid;
            to_type = TO_TYPE_DATE;
            msg = "<span class='q-part rel-reln'>make global date</span>";
        }
        else if ($classNObj.temp_nid==MEM.anorg.temp_nid) {
            cmd_nid = MEM.MAKE_ANORG.nid;
            to_type = TO_TYPE_STRING;
            msg = "<span class='q-part rel-reln'>make an anorg</span>";
        }
        else if ($classNObj.temp_nid==MEM.company.temp_nid) {
            cmd_nid = ANNE.MEM.MAKE_COMPANY.nid;
            to_type = TO_TYPE_STRING;
            msg = "<span class='q-part rel-reln'>make global company</span>";
        }
        //this is the url to get the img data.
        else if ($classNObj.temp_nid==MEM.neuronImg.temp_nid) {
            cmd_nid = ANNE.MEM.MAKE_IMG.nid;
            to_type = TO_TYPE_URL;
            msg = "<span class='q-part rel-reln'>make global IMG reference</span>";
        }
        else if ($classNObj.temp_nid==MEM.person.temp_nid) {
            cmd_nid = ANNE.MEM.MAKE_PERSON.nid;
            to_type = TO_TYPE_STRING;
            msg = "<span class='q-part rel-reln'>make person</span>";
        }
        else if ($classNObj.temp_nid==MEM.bounty.temp_nid) {
            cmd_nid = ANNE.MEM.MAKE_BOUNTY.nid;
            to_type = TO_TYPE_STRING;
            msg = "<span class='q-part rel-reln'>make bounty</span>";
        }
        else {
            alert("unhandled global create attempt");
            console.warn("unhandled global create attempt");
            return;
        }

        msg = msg + " <span class='q-part rel-to'>"+$val+"</span>";
        var rObj = ANNE.MEM.makeTempRelon(ANNE.USER._userNObj, TYPE_SYS_MSG, cmd_nid, to_type, $val, true);
        if (rObj==null) {
			console.warn("rObj failed to make temp CMD_NID relon!!! BIG TROUBLE WILL HAPPEN!!!");
		}
		rObj.msg = msg;
        
        if ($gui_extras) rObj.gui_extras=$gui_extras;   //single send is safer to use rObj.gui_extras not cmdObj.gui_extras

        var tempCmdObj = await ANNE.SENDQ.addRelonToSendQ(rObj, $bSINGLESEND, TXB.TXTYPE_MAKE_GLOBAL);
		
        if (tempCmdObj!=null) {
			tempCmdObj.delay = 500;
		}
		else {
			console.warn("big trouble will happen here!!!  Investigate why no tempCmdObj was created!!", rObj);
		}
		
        if (tempCmdObj && $gui_extras!=null && $gui_extras.znid!=undefined) tempCmdObj.znid = $gui_extras.znid;

        return tempCmdObj;
    }


   /**
     * Make a keylesss neuron via NODE request.
     * Keylesss neurons are made via standard sendPayData relons.
     * The payTo must be set to the node that the request it broadcast to.
     * 
     * NOTE: The ANNODES will reject any request to make a keyless class with a parent class
     * that does not yet exist.  In the case of a new top-class use the parent class of 
     * MEM.class (7777942827590344670) 
     * 
     * or 
     * 
     * MEM.gen_class (7770801601458804038) if you intend
     * this new class to be used in a property-like capacity (if it will be used mostly 
     * as the reln: in a relon, as opposed to being a class
     * that ultimately has children that are instances).
     * 
     * @param {anne_neuron} - the parent/be.
     * @param {anne_neuron} - MEM.inst or MEM.class
     * @param {function} - callback
     * @param {boolean} - trigger sending now (don't queue up further requests)
     * @author Mr.Scatman             
     */
    makeKeylessNeuron=async($classNObj, $instOrClassNOBJ, $callbackSingleSend=null, $bINSTANTSEND=true, $gui_extras=null)=>{
        var msg = "new keyless "+$instOrClassNOBJ.text+" of '"+$classNObj.text+"'";
        //console.log("%cKEYLESS %c"+msg, ANNE.TALK.bgred, ANNE.TALK.bggreen);
        var rObj = ANNE.MEM.makeTempRelon(ANNE.USER._userNObj, TYPE_SYS_MSG, $instOrClassNOBJ.nid, TO_TYPE_ANNE, $classNObj.nid, true);
        rObj.msg = msg;
        if ($gui_extras!=null) rObj.gui_extras = $gui_extras;
        var cmdObj = await ANNE.SENDQ.addRelonToSendQ(rObj, $bINSTANTSEND, TXB.TXTYPE_KEYLESS_NEURON, $callbackSingleSend);
        if (cmdObj!=null) {
            cmdObj.delay = 500; //we need to delay this to ensure a different epoch sec timestamp lest big troubles.
        }
        if ($gui_extras!=null && $gui_extras.znid!=undefined) cmdObj.znid=$gui_extras.znid;
        //console.log(cmdObj);
        return rObj;
    }

    /**
     * 
     * @param {anne_neuron} $classNObj 
     * @param {anne_neuron} $instOrClassNOBJ 
     * @param {function} $callbackSingleSend 
     * @param {Boolean} $bINSTANTSEND 
     * @param {Object} $gui_extras 
     */
    makePDNNeuron=async($classNObj, $instOrClassNOBJ, $callbackSingleSend=null, $bINSTANTSEND=true, $gui_extras=null)=>{
        var msg = "new PDN "+$instOrClassNOBJ.text+" of '"+$classNObj.text+"'";
        var rObj = ANNE.MEM.makeTempRelon(USER._userNObj, TYPE_SYS_MSG, $instOrClassNOBJ.nid, TO_TYPE_ANNE, $classNObj.nid, true);
        rObj.msg = msg;
        if ($gui_extras!=null) rObj.gui_extras=$gui_extras;
        await ANNE.SENDQ.addRelonToSendQ(rObj, $bINSTANTSEND, TXB.TXTYPE_PDN, $callbackSingleSend);
    }

    /**
     * Send a custodial neuron.
     * This will receive back critical data required to finish the custodial.
     * The from is a dynamic neuron (the custodial nid) but the to of the ANNOUNCE is the txid of the CUSTODIAL request.
     * The pubKey from the annode is passed back through and is must be set 
     * @param {anne_neuron} $classNObj 
     * @param {anne_neuron} $instOrClassNOBJ 
     * @param {function} $callbackSingleSend 
     * @param {Boolean} $bINSTANTSEND 
     * @param {Object} $gui_extras 
     */
    makeCustodialNeuron=async($classNObj, $instOrClassNOBJ, $callbackSingleSend=null, $bINSTANTSEND=true, $gui_extras=null)=>{
        var msg = "new custodial inst of '"+$classNObj.text+"'";
        var rObj = ANNE.MEM.makeTempRelon(ANNE.USER._userNObj, TYPE_SYS_MSG, $instOrClassNOBJ.nid, TO_TYPE_ANNE, $classNObj.nid, true);
        rObj.msg = msg;
        if ($gui_extras!=null) rObj.gui_extras=$gui_extras;
        await ANNE.SENDQ.addRelonToSendQ(rObj, $bINSTANTSEND, TXB.TXTYPE_CUSTODIAL, $callbackSingleSend);

        //we need to snag the last txid (among a few other vars in the response handler).
        //pop in 120 coins (12,000 annies) so start it off with.
        //NOTE -- ALL USAGE OF THE FAUCET OR CUSTODIAL SHOULD USE A referencedTransactionFullHash
        //to THIS ANNOUNCEMENT --- to ensure is mined in the same block or PRIOR to the usage of this.
        //We have to do this because this TX will be itself using a sys_ref_txid to the custodial creation transaction!
        //This means the ANNOUNCE may get delayed till the next block and then the fromception would fail for the TXs.

        //NOTE:  let's tell the node to wait on MEMPOOL TXS if they fail fromception.  Can we do that?  Hmm.
        //200 annies, 100 for fee, 100 init fund.
        var announce_relon = {annies:200, type:TYPE_SYS_MSG, from_nid:USER._userNObj.nid, reln:MEM.MAKE_CUSTODIAL.nid, to_type:TO_TYPE_TXID, to_s:null, pending_to_s:"new_"+rObj.txid};
        await ANNE.SENDQ.pendingDynamicVarRelon(announce_relon, false, false, true, TXB.TXTYPE_ANNOUNCE_CUSTODIAL_NEURON, null, false);
        return rObj;
    }


    
    /**
     * WIth a list of nids, check for min info chased.  If not, chase it.    
     * This function should be called over the chaseMin and chaseGhostMin as 
     * it smartly ascertains how to best chase the neuron.                   
     * bCHASED_min indicates we've looked up the 0/1/2 relons.  
     * @param {Array}
     * @author Mr.Scatman             
     */
     lookupMinTNids=async($temp_nids)=> {
        var nid_list = [];
        var nObjs = [];
        for (var n=0; n < $temp_nids.length; n++) {
            var nObj = MEMG.getNeuronByTempNid($temp_nids[n]);
            if (nObj.nid==USER.public_nid) continue;    //don't look for our own name.
            if (nObj.ghost==true) continue;
            if (nObj.bCHASED_min==true) continue;
            if (nObj.preloaded==true) continue;      
            nObjs.push(nObj);
            nid_list.push(nObj.nid);
            nObj.bCHASED_min=true;  //don't do this again :)
        }
        if (nid_list.length!=0) await this.lookupMinList(nid_list, [0,1,2]);
        return;
    }


    /**
    * Lookup a list of nids.
    * Typically used on classess that show up as relns.  We do not need
    * to get everything for these, until we do.  Then we do.  Otherwise,    
    * just get the basic min info.                                          
    * This will get any type:1 relons (including the be defining parent   
    * class/inst and the BE mem.name (for .text property)                   
    * bCHASED_min was set coming in to this function so neurons will not be 
    * chased again for mins.
    * @param {Array}
    * @author Mr.Scatman                                                
    */
    lookupMinList=async($nidslist, $types=[0,1,2], $from="", $bFIXAFTERIDB=false)=>{
        var aRelons = await ANNE.API.lookupNids($nidslist, $types, true, true, $from, $bFIXAFTERIDB);
        return aRelons;
    }
    lookupNonAwareList=async($nidslist)=>{
        var aRelons = await ANNE.API.lookupNidsNonAwareness($nidslist);
        return aRelons;
    }
    lookupNidsGetAllOuts=async($nidslist, $offset=0)=>{
        var aRelons = await ANNE.API.lookupNidsGetAllOuts($nidslist);
        if (aRelons.length==ANNE.API.getRelonLimit()) {
            console.warn("IMPLEMENT OFFSET RETURN LISTS");
            console.warn("THERE ARE MORE RELONS TO GET");
        }
        return aRelons;
    }

    /**
     */ 
    sendACHATMsgForSeshId=async($msg, $sesh_id, $el_error=null)=>{
        if ($msg=="") return;
        var local_db_msgs = await DB_local.__getACHATMsgsOnly();
        var max_msgid_for_this_sesh = 0;
        var msg_ids = [];
        for (var m=0; m < local_db_msgs.length; m++) {
            var msg = local_db_msgs[m];
            if (msg_ids.indexOf(msg.msg_id)==-1) msg_ids.push(msg.msg_id);
            if (parseInt(msg.msg_id) > max_msgid_for_this_sesh) max_msgid_for_this_sesh = parseInt(msg.msg_id);
        }
        var msgs = await ANNE.API.sendACHATMsgForSeshId($msg, $sesh_id, max_msgid_for_this_sesh, $el_error);
        return this.processAChatMsgs(msgs, msg_ids);
    }


    /**
     */
    getACHATMsgsForSeshId=async($sesh_id, $el_error=null)=>{
        var local_db_msgs = await DB_local.__getACHATMsgsOnly();
        //critical optimizer to help relieve burden from annode re-encrypting
        //messages back to us that we already have.
        var max_msgid_for_this_sesh = 0;
        var msg_ids = [];
        for (var m=0; m < local_db_msgs.length; m++) {
            var msg = local_db_msgs[m];
            if (msg_ids.indexOf(msg.msg_id)==-1) msg_ids.push(msg.msg_id);
            if (parseInt(msg.msg_id) > max_msgid_for_this_sesh) max_msgid_for_this_sesh = parseInt(msg.msg_id);
        }
        var msgs = await ANNE.API.getACHATMsgsForSeshId($sesh_id, max_msgid_for_this_sesh, $el_error);
        return ANNE.CHASE.processAChatMsgs(msgs, msg_ids);
    }


    /**
     */
    processAChatMsgs=async($msgs, $msg_ids)=>{
        var ssObj = await ANNE.USER.getSharedSecret();
        if (ssObj==null) return $msgs;
        var npid = ANNE.API.getNextProofId();
        var new_data = [];
        for (var m=0; m < $msgs.length; m++) {
            var msgObj = $msgs[m];
            if ($msg_ids.indexOf(msgObj.msg_id)==-1) {
                $msg_ids.push(msgObj.msg_id);
                new_data.push(msgObj);
                msgObj.anne_app_user = ANNE.USER._userNObj.nid;
            }

            if (msgObj.block_height > ANNE.HISTORY._GREATEST_LAST_KNOWN_BLOCK) {
                ANNE.HISTORY.updateGLKB_SOLID(msgObj.block_height);
            }

            //these msgs have an encrypted msgs field. 
            //decrypt the msgs (from the annode using the ephemeral key for achat msgs)
            msgObj.msg_decrypted = await ANNE.USER.decryptACHATmsg(msgObj.msg, ssObj.seed, npid);
        }
        //store msgs in local db.
        var done = await ANNE.DB_local.___insert(new_data, "achat_msgs");
        return $msgs;   //return new msgs.
    }


    /**
     * This gets new achat sessions from handshook annode
     * and then returns ALL in memory (from any annode)
     * These are displayed in the ACHAT tool.
     */
    getAllACHAT_sessions = async($bREFRESHONFAIL=true, $el_error=null)=> {
        var sessions = await ANNE.API.getACHATSessions($bREFRESHONFAIL, $el_error);     
        
        //note: should filter for user so other users don't get
        //messages that are not for them (anne_app_user)
        var overview = await DB_local.__getOverviewOnly();
        var new_data = [];
        function isNew($s) {
            for (var o=0; o < overview.length; o++) {
                if (overview[o].sesh_id==$s.sesh_id 
                    && overview[o].sesh_participant_id==$s.sesh_participant_id) return false;
            }
            return true;
        }
        if (sessions.length) {
            for (var s=0; s < sessions.length; s++) {
                var seshObj = sessions[s];
                if (isNew(seshObj)==true) {
                    new_data.push(seshObj);
                    overview.push(seshObj);
                }
                seshObj.annode_uri = ANNE._annodesLS.last_uri;
                seshObj.anne_app_user = ANNE.USER._userNObj.nid;
            }
            if (new_data.length) {
                var done = await ANNE.DB_local.___insert(new_data, "achat_overview");
            }
        }
        console.log(overview);
        return overview;
    }



    /**
     * Get all the TXs that touch the USER.
     * outs/ins and relon_sponsored.
     * Note: all childRelPat relons broadcast will have the user as the relon_sponsor.
     * Note2: we have to make sure to optimize that query to efficiently grab all that data.
     * @param {*} $nid 
     */
    getAllUserPayments = async($nid, $bDOSPONSORED=false, $bDOIDB=true)=>{
        if (ANNE.SETTINGS.bCONSOLE_DEBUG==true) console.log("(getAllUserPayments)");
        
        //get all annies in/out from broadcast/internal payments.
        //also snag any in/out relons with .annies on it.
        await ANNE.API.getAllPayments($nid, "getAllUserPayments", $bDOIDB);
        
        var nOBJ = MEMG.getNeuronByNid($nid);
        var rels = nOBJ.outs.concat();

        //debug --- WE WILL REMOVE THIS LATER.
        for (var r=0; r < rels.length; r++) { rels[r].debug="from outs";}
        for (var r=0; r < nOBJ.ins.length; r++) { nOBJ.ins[r].debug="from ins";}
        
        if (ANNE.isANNE(ANNE.USER._userNObj.nid)==false) {       
            if (nOBJ.relon_sponsored!=undefined) {
                for (var r=0; r < nOBJ.relon_sponsored.length; r++) { nOBJ.relon_sponsored[r].debug="from relon sponsor";}
            }
        }
        
        //so... outs and ins..
        rels = rels.concat(nOBJ.ins);
        
        if ($bDOSPONSORED==true && ANNE.UTILS.isset(nOBJ.relon_sponsored)==true) {
            return ANNE.UTILS.uniqueConcatRels(rels, nOBJ.relon_sponsored);
        }
        return rels;
    }


    /**
     * This does not get all TXs.. (as of our new RAM optimizer IDB mashup)
     * @param {*} $nid 
     * @returns 
     */
    getAllUserTransactions3 = async($nid)=>{
        await ANNE.API.getAllPaymentsIn($nid);
        var nOBJ = MEMG.getNeuronByNid($nid);
        var rels = nOBJ.outs.concat();
        rels = rels.concat(nOBJ.ins);
        return rels;
    }


    /**
     */
    search_recents_by_class=async($nids=[], $bCHASEIT=true, $maxFireWeight = 0.01, $maxResults=200)=>{
        if ($bCHASEIT==true) {
            var relons = await ANNE.API.queryForRecents($nids, $maxResults);
            if (ANNE.SETTINGS.debug()==true) console.log(relons);
        }
        return ANNE.SEARCH.getRecents($nids, $maxFireWeight, $maxResults);
    }

    
    /**
     * 
     * @param {*} $searchVal 
     * @param {*} $class_tnid 
     * @param {*} $wvs 
     */
    anne_bem_search=async($searchVal, $class_tnid, $wvs=[])=> {
        var done = await ANNE.API.queryForText_BEM ($searchVal, $class_tnid, $wvs);
        console.log(done);

        //we need to make sure these are unique.. (TODO);
        var matches = await ANNE.SEARCH.searchNeuronText ($searchVal, $class_tnid, $wvs, false);
        if (matches.length) return matches;
    }


    /**
     * 
     * @param {String} $searchVal 
     * @param {Boolean} $bLOCALONLY 
     * @param {Array} $ntypes 
     * @param {Boolean} $bDOAKAS 
     * @returns 
     */
    anne_naked_neuron_search=async($searchVal, $bLOCALONLY=false, $ntypes=[NTYPE_INST, NTYPE_CLASS], $bDOAKAS=false)=> {
        
        //search all spaces
        if ($bLOCALONLY==false) await ANNE.API.queryForText ($searchVal, [], [], true);

        //search only local.
        return ANNE.SEARCH.searchTextStart($searchVal, $ntypes, false, $bDOAKAS);
    }


    /**
     * for app boot up to clean up some messes that may have bene caused.
     * for large TX sends where the user closes the app.. this could be no so frenly..
     * 6k unconfirmed TXs would not exactly be fun to chase out.  Maybe we max it out 
     * and chip away?
     * 
     * ANY TXIDS returned from this function will be deleted from IDB mem.
     * @param {*} $txids_to_check 
     * @returns 
     */
    getAllDEDTXIDS=async($txids_to_check)=> {
        var all_ded = await this.getTXIDSThatDontExistInMEMPOOOL($txids_to_check);

        //these TXIDS do not exist in the MEMPOOL...  were they mined or do they not actually exist??
        if (all_ded.length) {
			return await this.getTXIDSThatDontExistInANNE(all_ded);
		}
        return [];
    }


	/**
	 * 
	 * @param {*} $memPoolBootups 
	 */
	getTXIDSThatDontExistInMEMPOOOL=async($memPoolBootups)=>{
		var SHARD = 30;
		var txids = [];
        var all_deds = [];
		for (var t=0; t < $memPoolBootups.length; t++) {
            if ($memPoolBootups[t][0]=="z") continue;
			if (txids.length==SHARD) {
                var ded_txids = await ANNE.API.getTXIDSThatDontExistInMEMPOOOL(txids);
                txids=[];
                if (ded_txids.length) all_deds = all_deds.concat(ded_txids);
			}
            txids.push($memPoolBootups[t]);
		}
        if (txids.length) {
            var ded_txids = await ANNE.API.getTXIDSThatDontExistInMEMPOOOL(txids);
            if (ded_txids.length) all_deds = all_deds.concat(ded_txids);
        }
        return all_deds;
	}

    
    /**
     * 
     * @param {*} $txids_to_check 
     * @returns 
     */
	getTXIDSThatDontExistInANNE=async($txids_to_check)=> {
        var all_deds = [];
		var aRels = await this.lookupTxids($txids_to_check);

		if (aRels.length > 0) {
			var exist_txids = [];
			
			for (var a=0; a < aRels.length; a++) {
				exist_txids.push(aRels[a].txid);
			}
			
			for (var r=0; r < $txids_to_check.length; r++) {
                if (exist_txids.indexOf($txids_to_check[r])!=-1) {
					continue;
				}

                //not in real space.
                all_deds.push($txids_to_check[r]);
            }
		}
        else return $txids_to_check;

        return all_deds;
	}


    /**
     */
    queryForOutsFroms_not=async($arr, $bMUSTPULL=false, $types=[4], $bDOIDB=true)=> {
        var SHARD = 30;
        var this_shard = [];

        //one big one instead of repeated pulls of the entire IDB space to loop.
        var hash = ANNE.other.sha256($arr.join("")+$types.join(""));
		var phObj = ANNE.HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForOuts_not", hash);
		if ($bDOIDB==true) await ANNE.IDB.queryForOutsFroms_not(phObj, $arr, $types);

        const bDOIDB = false;
        
        for (var s=0; s < $arr.length; s++) {
            if (this_shard.length==SHARD) {
                await ANNE.API.queryForOutsFroms_not(this_shard, $bMUSTPULL, $types, bDOIDB);
                this_shard = [];
            }
            this_shard.push($arr[s]);
        }

        //if any left over.
        if (this_shard.length) {
            await ANNE.API.queryForOutsFroms_not(this_shard, $bMUSTPULL, $types, bDOIDB);
        }
        
        return;
    }
	


    /**
     * simple and fast.
     * @param {*} $text 
     * @param {*} $bTAKEIT 
     * @param {*} $bDOIDB 
     * @param {*} $bEXACT - we turn this on if a string is quoted when typed in. 
     * @returns 
     */
    nakedSearch = async($text, $bTAKEIT=true, $bDOIDB=true, $bEXACT=false, $WVs=[])=>{

        //we're ok with a quick hit value without looking up any new info
        if ($bTAKEIT==true) {
            var nObjs = await ANNE.MEM.searchTextIndex($text, true, $bEXACT);
            if (nObjs.length>0) return nObjs;
        }

        //search ANNE -- note, we don't want other people's stuff here, unless we say so.
        //we can use a diff function or alter this one later.  For now, only return ANNE.
        var from_nids = await ANNE.API.queryForNakedText($text, $bEXACT);

        //search our RAM 
        var nObjs = await ANNE.MEM.searchTextIndex($text, false);
        
        //pull snag IDB mem and load in RAM, typically so we have img ref, parent class ref.. etc
        if ($bDOIDB==true && nObjs.length) {
            var nids = MEMG.getNidsListFromNObjs(nObjs);

            for (var f=0; f < from_nids.length; f++) {
                if (nids.indexOf(from_nids[f]==-1)) nids.push(from_nids[f]);
            }

            //comprehensive list of nids to snag from IDB
            await ANNE.IDB.lookupNids(null, nids, []);
        }

		else if(from_nids.length>0) {
            //no RAM text index matches.. snag the from_nids we had from our naked query.
			await ANNE.IDB.lookupNids(null, from_nids, []);
			nObjs = MEMG.getNObjsFromList(from_nids);
		}

        return nObjs;
    }


    /**
     * We are trying to figure out if we have something or not
     * @param {*} $text 
     */
    nakeSearchForResolution = async($text, $classTempNid)=>{
        var matches = [];
        
        var nObjs = await ANNE.MEM.searchTextIndex($text);
        
        var nids = MEMG.getNidsListFromNObjs(nObjs);
        await ANNE.IDB.lookupNids(null, nids, []);

        for (var n=0; n < nObjs.length; n++) {
            if (nObjs[n].bes.indexOf($classTempNid)!=-1) {
                matches.push(nObjs[n]);
            }
        }

        if (matches.length>0) return matches;

        //hit the annode.
        await ANNE.API.queryForNakedText($text);
        nObjs = await ANNE.MEM.searchTextIndex($text);
        if (nObjs.length) {
            for (var n=0; n < nObjs.length; n++) {
                if (nObjs[n].bes.indexOf($classTempNid)!=-1) {
                    matches.push(nObjs[n]);
                }
            }    
        }

        return matches;
    }


    /**
     * Do a global search for the text string.
     * 
     * After the global search, do a local search.
     * Return those results.
     * 
     * NOTE: use the $so.context field to resolve ambiguities (multiple matches)
     * if possible.
     * 
     * @param {Array}
     * @author Mr.Scatman 
     */
    anne_global_search = async ($so, $bDOPRE=false, $bACCEPTRAMONLY=true)=> {

        //text searches are not fun when **not** preloading all of IDB.
        //await ANNE.IDB.textSearch($so.text, $so.exact);
        //await ANNE.IDB.queryForText($so.text, null);
    
        //we're ok here if we find a match before searching the ANNODE space.
        if ($bDOPRE==true) {
            var matches = await ANNE.SEARCH.searchNeuronText ($so.text, $so.bes_tnids, $so.wvs, $so.exact, $so.applywvtobeline, $so.inbeline); 
            
            //not chasing it, then return matches or no matches here.  Sorry.
            if ($so.chaseIt==false) {
                if (matches.length) return matches;
                return [];
            }
            if ($bACCEPTRAMONLY==true && matches.length>0) {
                return matches;
            }
        }

        if ($so.chaseIt==true) {
            var done = await ANNE.API.queryForText ($so.text, $so.wv, $so.bes, $so.exact, $so.applywvtobeline, $so.inbeline);
            console.log(done);
        }

        //to do, fix for applywvtobeline and inbeline logics (???)
        return await ANNE.SEARCH.searchNeuronText ($so.text, $so.bes_tnids, $so.wvs, $so.exact, $so.applywvtobeline, $so.inbeline);   
    }


    /**
     * We reroute a text search to a nid search based on nid size and all numerics.
     * @param {Array}
     * @author Mr.Scatman 
     */
    anne_nid_search = async ($nid, $bMUSTPULL=false)=> {
        var done = await ANNE.API.queryForOuts ($nid, $bMUSTPULL);
        console.log(done);
        var matches = [];
        var nObj = MEMG.getNeuronByNid($nid);
        if (nObj!=null) matches.push(nObj);
        return matches;
    }


    /**
     */
    getAllMyACHATSessions = async()=> {
        var done = await ANNE.API.queryForOuts ($nid);
        console.log(done);
    }


    /**
     * We just need to ensure we have data from the ANNODE in the local memory
     * so we can process a string that we know to be a location.
     */
    locationParts=async($aParts)=>{
        var done = await ANNE.API.queryForLocationParts ($aParts);
    }


    /**
     * @param {String} - String we are looking for
     * @param {String} - Be nid string of the parent class
     * @author Mr.Scatman 
     */
    queryForTextAndBe=async($text, $be_nid, $WVs=[MEM.ANNE_ACCTID], $bMUSTPULL=false)=>{
        if ($text==undefined || $text==null || $text=="") return [];
        
        //note: for specific classes with known properties this is not the efficient way
        //to find them.  We should target their optimized lookups stored in the class
        //.rzpp1 and .rzpp2 props (set by the optimizing relons 7770223143918999134 and/or 7772071294239796669)

        var done = await ANNE.API.queryForTextAndBe ($text, $be_nid, $WVs, $bMUSTPULL);
        console.log(done);
        var beNOBJ =  MEMG.getNeuronByNid($be_nid);
        var aMatches = [];
        var textlower = $text.toLowerCase();
        for (var i=0; i < beNOBJ.ins.length; i++) {
            var childNOBJ=MEMG.getNeuronByTempNid(beNOBJ.ins[i].f);
            if ($WVs.indexOf(childNOBJ.owned_by)==-1) continue;
            if (childNOBJ.text.toLowerCase()==textlower) aMatches.push(childNOBJ);
        }
        return aMatches;
    }

    queryForCustom=async($beNOBJ)=>{
        var rzpp1 = $beNOBJ.rzpp1;
        var rzpp2 = $beNOBJ.rzpp2;
    }

 
    /**
     * 
     * @param {*} $reln 
     * @param {*} $to 
     * @param {*} $to_type 
     * @returns 
     */
    getRelonsWithTo= async($reln, $to, $to_type)=> {
        var aRelons = await ANNE.API.getRelonsWithTo($reln, $to, $to_type);
        await ANNE.MEMP.processNeuronsFromRelons(aRelons, true, true, "relons", "getRelonsWithTo");
        return aRelons;
    }


    /**
     * This is a bit willy-wonka.
     * It used to be tight with forced registratoin polity nid.
     * Now it can be used for just a basic company lookup (no required registration polity)
     * This can return multiple values back.
     */
     getCompany=async($namestr, $registrationPolityNid=null)=>{
        console.log("(CHASE.getCompany)");
        var response = await ANNE.API.getCompany($namestr, $registrationPolityNid);
        //console.log("response:", response);
        if (response.status=="good" && response.message=="available") {
            return response;
        }
        else {
            var rObj = response.result;
            if (Array.isArray(rObj)==true) {
                var aRelons = rObj;
                var arelons2 = await ANNE.MEMP.processNeuronsFromRelons(aRelons, true, true, "relons", "getCompany");
                ANNE.TALK.printRelons(aRelons);
                var nObj = MEMG.getNeuronByNid(aRelons[0].from_nid);
                response.results = [nObj];
                response.result = nObj;
                return response;
            }
            else if (rObj!=null) {
                ANNE.TALK.printRelon(rObj);
                await ANNE.MEMP.processNeuronsFromRelons([rObj], true, true, "relons", "getCompany");
                var nObj = MEMG.getNeuronByNid(rObj.from_nid);
                response.results = [nObj];
                response.result = nObj;
                return response;
            } 
            else {
                alert("fix this.. it should have returned back the getCompany exiting value");
            }
        }
        return null;
     }


    /**
     */
    getPersonWithBday=async($bday, $namestr, $alias="")=> {
        console.log("(CHASE.getPersonWithBday)");
        var aMatches = await ANNE.SEARCH.searchNeuronText($namestr, [ANNE.MEM.person.temp_nid], [ANNE.USER._userNObj.nid, ANNE.MEM.ANNE_ACCTID], true);
        if (aMatches.length) {
            console.log(aMatches);
            return {status:"good", message:"taken", results:aMatches, result:aMatches[0]};
        }
        var response = await ANNE.API.queryForPersonWithBday($bday, $namestr, $alias);
        console.log("response:", response);
        if (response.status=="good" && response.message=="available") {
            console.log("GO TIME");
            return response;
        }
        else {
            var results = response.result;
            if (results!=null) {
                ANNE.TALK.printRelons(results);
                await ANNE.MEMP.processNeuronsFromRelons(results, true, true, "relons", "getPersonWithBday");

                //relons could hvae just been date relons..
                var aMatches = await ANNE.SEARCH.searchNeuronText($namestr, [ANNE.MEM.person.temp_nid], [ANNE.USER._userNObj.nid, MEM.ANNE_ACCTID], true);
                if (aMatches.length) {
                    console.log(aMatches);
                    return {status:"good", message:"taken", results:aMatches, result:aMatches[0]};
                }
                else return {status:"good",message:"available"};
            } 
            else {
                alert("fix this.. it should have returned back the getPersonWithBday exiting value");
            }
        }        
        return null;
    }

   
    
    /**
     */
    getPerson=async($namestr)=> {
        console.log("(CHASE.getPerson)");
        var aMatches = await ANNE.SEARCH.searchNeuronText($namestr, [ANNE.MEM.person.temp_nid], [ANNE.USER._userNObj.nid, ANNE.MEM.ANNE_ACCTID], true);
        if (aMatches.length) {
            console.log(aMatches);
            return {status:"good", message:"taken", results:aMatches, result:aMatches[0]};
        }
        var response = await ANNE.API.getPerson($namestr);
        console.log("response:", response);
        if (response.status=="good" && response.message=="available") {
            console.log("GO TIME");
            return response;
        }
        else {
            var results = response.result;
            if (results!=null) {
                ANNE.TALK.printRelons(results);
                await ANNE.MEMP.processNeuronsFromRelons(results, true, true, "relons", "getPerson");

                //relons could hvae just been date relons..
                var aMatches = await ANNE.SEARCH.searchNeuronText($namestr, [ANNE.MEM.person.temp_nid], [ANNE.USER._userNObj.nid, ANNE.MEM.ANNE_ACCTID], true);
                if (aMatches.length) {
                    console.log(aMatches);
                    return {status:"good", message:"taken", results:aMatches, result:aMatches[0]};
                }
                else return {status:"good",message:"available"};
            } 
            else {
                alert("fix this.. it should have returned back the getPerson exiting value");
                return {status:"bad", message:"unknown problem"};
            }
        }        
        return null;
    }



    /**
     * Get aplex data (annies_in, pos/neu/neg feedback, sell status).
     * @param {string} - nid long string for the neuron you want to get aplex data for.
     * @author Mr.Scatman 
     */
     get_aplex_data = async ($nid, $bMUSTPULL=false, $from="")=> {
        if ($nid[0]=="z") return null;
        var nObj = MEMG.getOrMakeIfNew($nid);
        if (nObj.ghost==true) return null;
        if($bMUSTPULL==false && nObj._aplex!=undefined) return nObj._aplex;
        console.log("%c NEURON APLEX bug hunt! (CHASE.get_aplex_data) from"+$from, ANNE.TALK.bgbigred);  
        nObj._aplex = await ANNE.API.getAplex (nObj);
        return nObj._aplex;
    }


    lookupTxid = async($txid)=> {
        var rels = await this.lookupTxids([$txid]);
        if (rels.length>0) return rels[0];
        return null;
    }


    lookupTxids = async($txids)=> {
        return await ANNE.API.lookupTxids($txids);
    }
    

    /**
     * 
     * @param {*} $nid 
     * @param {*} $bDOPUBKEY 
     * @returns 
     */
    lookupMinNid=async($nid, $bDOPUBKEY=false)=>{
        if (typeof $nid==="object") {
            console.warn("you should pass a number into lookupMinNid, not an object");
            if ($nid.nid!=undefined) {
                $nid = $nid.nid;
                console.warn("in this instance we were able to fix it for you.  PLEASE ALTER YOUR CODE");
            }
            else return null;
        }
        return await ANNE.API.lookupNid($nid, [0,1,2], $bDOPUBKEY);
    }


    /**
     * ARE WE STILL USING THIS?? lolz
     * Returns an array of vote neurons.
     */
     getVoteData = async($nObj)=> { 
        var newVoteData = await ANNE.API.getVoteData($nObj.nid);
        console.log("new vote data", newVoteData);
        return MEMG.getVoteData($nObj.nid);
    }



    /*** chaseDDS_Src ***************************************** x.x.20xx ****
    * Chase out a data source until we reach a non-DDS neuron..  This is    *
    * the root query/neuron and we can get its children.                    *
    * The rest of the downstream DDS wrappers were filters off of that      *
    * original source.   We will gather all WV/garden/deny/lets and filters *
    * for the entire ANNE DDS path...  This will allow us to re-create the  *
    * datasource, live.                                                     *
    * THIS WILL ULTIMATELY RETURN OUT THE SOURCE NEURON.                    *
    ********************************************************* Mr.Scatman ***/
    chaseDDS_Src = async($nObj)=> {
        console.warn("TO DO");
        await ANNE.CHASE.lookupMinNid($nObj.nid);
        if ($nObj.bes[0]!=ANNE.MEM.ANNE_DDS.temp_nid) {
            await ANNE.CHASE.chaseIns_GWV($nObj);
            return $nObj;   
        }
        else {
            var DDS_SRC_NID = $nObj["be_"+ANNE.MEM.ANNE_DDS_SRC.temp_nid];
            if (DDS_SRC_NID==undefined) {
                var ddsRelon = MEMG.getRelonFromWithRToType($nObj.outs, TYPE_BE, ANNE.MEM.ANNE_DDS_SRC.nid, TO_TYPE_ANNE); 
                if (ddsRelon==null) return;
                DDS_SRC_NID = ddsRelon.t;
            }
            console.log(DDS_SRC_NID);
            var DDSNOBJ = MEMG.getNeuronByTempNid(DDS_SRC_NID); 
            return await ANNE.CHASE.chaseDDS_Src(DDSNOBJ);
        }
    }


    /*** chaseNewInfo ***************************************** x.x.2019 ****
    * User on tree/folder clicks and graph neuron clicks.                   *
    ********************************************************* Mr.Scatman ***/
     chaseNewInfo=async($nObj, $type=[0,1,2,3])=> {
        if ($nObj==null || $nObj==undefined) {
            console.warn("(chaseNewInfo) hunt this down");
            return [];
        }

        //get 0/1/2/3 outs..  and INS
        if ($nObj.ntype==NTYPE_CLASS) {
            var aRelons2 = await ANNE.CHASE.queryForOuts($nObj, false, $type);
            var aRelons = await ANNE.CHASE.chaseIns_GWV($nObj, false);
        }

        //get 0/1/2/3 outs only
        else {
            var aRelons = await ANNE.CHASE.queryForOuts($nObj, false, $type);
        }
        return;
    }


    //READRESS THIS NOW THAT WE USE THE ANNODE AS A PERSONAL WEB SERVER.
    //THESE OPTIMIZATIONS ARE A PAIN IN THE ASS AND LESS NEEDED/USEFUL
    //WHEN WE ARE NOT TRYING TO SERVICE MANY USERS WITH ONE ANNDOE.

    /*** chaseIns_GWV ***************************************** x.x.20xx ****
    * THIS FUNCTION MUST HONOR THE GARDEN TYPE AND WORLDVIEW RULES.  IT     *
    * CAN ONLY BE RUN IF THE nOBJ.bCHASED_min is set to true, indicating    *
    * we have gotten all the data we need to properly chase the ins.        *
    ********************************************************* Mr.Scatman ***/
    chaseIns_GWV = async($nObj, $bMUSTPULL=false, $intypes=[1,3], $chasetypes=[0,1,2], $bOVERRIDE_CHECK=false)=>{
        if ($bOVERRIDE_CHECK==false && $nObj.bCHASED_min==false) {
            await ANNE.CHASE.lookupMinNid($nObj.nid);
        }
        var nid = $nObj.nid;
        
        //rules object looks like:  
        //{let:[nid addesses], deny:[nid addesses]};
        //NONE indicates DO NOT apply that filter to the query.

        //only chase it (call the middle layer) if we haven't already done so this (bCHASE_ins)
        //note, it is complex to try and optimize by intypes and chasetypes thus we will not
        //try to limit queries based on tracking that.  The caller can determine how to limit
        //these types of queries.  (ie, like with bCHASED_Ins flag used for dates in the threejs viewer)
        if ($bMUSTPULL==true) {
            var aRelons = await ANNE.API.chaseInsGWV(nid, $nObj.trusted_adopt, $bMUSTPULL, $intypes, $chasetypes);
        }

        return $nObj.outs;
    }



    /**
     */
    getAllMyAwarnesses = async($identity_nid, $bCHASEREMEMBERZ=true)=> {
        var offset = 0;
        var max_results = 1000;     //this is set in the annode code.
        var final_relons = [];
        var maxloops = 5;
        var lastBlock = 0;
        //var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForAwareness", $identity_nid);
        lastBlock = 0;//phObj.lastBlock;
        var rememberz_nids = [];

        while (maxloops>0) {
            maxloops=maxloops-1;
            var relons_to_process = await ANNE.API.queryForAwareness($identity_nid, lastBlock, offset);
            if (relons_to_process.length) {
                var aNewRelons = await ANNE.MEMP.processNeuronsFromRelons(relons_to_process, true, true, "relons", "getAllMyAwarnesses");
                ANNE.TALK.printRelons(aNewRelons);
                //await HISTORY.processResultsForNewPullHistory(relons_to_process, phObj);
                console.log(relons_to_process);
                if (relons_to_process.length!=max_results) {
                    final_relons = final_relons.concat(aNewRelons);
                    break;
                }
                final_relons = final_relons.concat(aNewRelons);
                offset = offset+max_results;
            }
        }

        for (var f=0; f < final_relons.length; f++) {
            if (final_relons[f].reln==MEM.rememberz.nid) {
                rememberz_nids.push(final_relons[f].to_s);
            }
        }
        if (rememberz_nids.length) {
            await this.lookupMinList(rememberz_nids, [0,1,2]);
        }

        return final_relons;
    }



    /**
     * No garden or wv here..
     * It thus uses an offset if max results are reached (1000 max)
     */
     queryForInsWithRelnTo = async($type, $reln, $to, $to_type, $offset=0)=>{
        return await ANNE.API.queryForInsWithTo($type, $reln, $to, $to_type, $offset);
    }


    /**
     * 
     * @param {*} $nidOrTxid 
     * @param {*} $rObj 
     * @returns 
     */
     getAwarenessSummary = async ($nidOrTxid, $rObj=null) => {
        alert("(CHASE.getAwarenessSummary) do we still use this?");
        console.log("(CHASE.getAwarenessSummary) nidOrTxid:"+$nidOrTxid);
        var nObj = null;

        //feelz on a txid.
        if ($rObj!=null) {
            nObj = MEMG.getNeuronByTempNid($rObj.t); //emulated neuron (for the results, but use txid to get data)
            var id = $rObj.txid;
            var aRez = await ANNE.API.___getAwarenessSummary(id);
            var awarenessMongoDoc = {empty:true};
            if (aRez.length) {
                awarenessMongoDoc = aRez[0];
                var objkeys = Object.keys(awarenessMongoDoc);
                var total = 0;
                for (var k=0; k < objkeys.length; k++) {
                    if (objkeys[k]=="pos") total+=awarenessMongoDoc[ objkeys[k] ];
                    else if (objkeys[k]=="neg") total+=awarenessMongoDoc[ objkeys[k] ];
                    else if (objkeys[k]=="neu") total+=awarenessMongoDoc[ objkeys[k] ];
                }
                awarenessMongoDoc.total=total;
                if ($rObj) $rObj.awarenessMongoDoc = awarenessMongoDoc;
            }
            else {
                if ($rObj) $rObj.awarenessMongoDoc = {empty:true};
            }
            return awarenessMongoDoc;
        }

        //standard neuron feelz
        else {
            nObj = MEMG.getOrMakeIfNew($nidOrTxid);
            if (nObj.ghost==true ||  nObj.emulated==true || nObj.preloaded==true || nObj.temp==true) {
                return {empty:true};
            }
            var id=$nidOrTxid;
            var aRez = await ANNE.API.___getAwarenessSummary(id);
            var awarenessMongoDoc = {empty:true};

            if (aRez.length) {
                awarenessMongoDoc = aRez[0];
                var objkeys = Object.keys(awarenessMongoDoc);
                var total = 0;
                for (var k=0; k < objkeys.length; k++) {
                    if (objkeys[k]=="pos") total+=awarenessMongoDoc[ objkeys[k] ];
                    else if (objkeys[k]=="neg") total+=awarenessMongoDoc[ objkeys[k] ];
                    else if (objkeys[k]=="neu") total+=awarenessMongoDoc[ objkeys[k] ];
                }
                awarenessMongoDoc.total=total;

                if (nObj) nObj.awarenessMongoDoc = awarenessMongoDoc;
            }
            else {
                if (nObj) nObj.awarenessMongoDoc = {empty:true};
            }
            return awarenessMongoDoc;
        }
    }


    /**
     * 
     * @param {*} $to_nid 
     * @param {*} $relns 
     * @returns 
     */
    getSpecificAwarenesses=async($to_nid, $relns)=>{
        var nOBJ = MEMG.getNeuronByNid($to_nid);
        if (nOBJ.ghost==true) return [];
        var aRelons = await ANNE.API.getSpecificAwarenesses($to_nid, $relns);
        return aRelons;
    }


    /**
     * 
     * **** REDO THIS FOR ANNODE CHANGES ****
     * 
     * 
     * We are looking for a specifc type of inbound relon here..           
     * Used in dynamic tree generation when looking for a non-standard      
     * child.. like a Ford150 which is a model that has a TYPE_BE of make to
     * ford.                                                               
     * We cannot use our ins type optimizer since this is qualified by the 
     * reln.  We must depend upon our HISTORY tracker to optimize this type
     * of function.  If, however, we've already chased the ins for these  
     * types, then it is presumed we'd already have the reln for that    
     * schmema too.  SO, we can throttle back on queries but not alter the  
     * ins tracker from it.                                                 
     */
     chaseInsWithReln=async($nObj, $reln, $intype, $chasetypes=[0,1,2], $desiredParentNOBJ=null)=>{
        var intypes = [parseInt($intype)];
        if ($nObj.trusted_adopt==null) $nObj.trusted_adopt = [];
        var aRelons = await ANNE.API.chaseInsWithRelnGWV($nObj.nid, $reln, $nObj.trusted_adopt, intypes, $chasetypes, $desiredParentNOBJ);
        return aRelons;
    } 


    /**
     * @author {Mr.Scatman}
     */
    searchTMDBByString=async($searchVal, $cb=null, $validMatchTypes=["person"])=> {
        $searchVal = $searchVal.trim();
        //var validMatchTypes = ["person"]; //"company" 
        var matches = [];
        console.log("search moviedb for "+$searchVal);
        this.searchTMDBMulti($searchVal, function($rez) {
            console.log($rez);
            if ($rez.length==0) {
                if ($cb!=null) $cb([], $searchVal);
            }
            else {
                var aMatches = [];
                for (var r=0; r <$rez.length; r++) {
                    var rez = $rez[r];
                    if ($validMatchTypes.indexOf(rez.media_type)!==-1) {
                        aMatches.push(rez);
                    }
                }
                if ($cb!=null) $cb(aMatches, $searchVal);
            }
        });
    }



    /**
     * @author {Mr.Scatman}
     */
    searchTMDBMulti = async ($textString, $callback, $append=null)=> {
        var params = {
            "lang": "en",
            "timezone": "America/New_York",
            "adult" : false
        };
        console.log($textString);
        params.query=$textString;
        this.tmdb.call("/search/multi", params, 
            function(data) {if($callback) $callback(data.results);}, 
            function (err) {if($callback) $callback(data);}
        );
    }


    /**
     * @author {Mr.Scatman}
     */
    getMovieDBPerson= async ($personID, $callback, $appended=null) => {
        console.log("(getMovieDBPerson) person:"+$personID);
        var params = {
            "lang": "en",
            "timezone": "America/New_York",
            "adult" : false
        };
        if ($appended==null) {
            var append = ['movie_credits', 'tv_credits', 'combined_credits', 'external_ids', 'images', 'tagged_images'];
            params['append_to_response'] = [];
            params['append_to_response'] = append;
        }
        else params['append_to_response'] = $appended;
        this.tmdb.call("/person/"+$personID, params, 
            function(data) {if($callback) $callback(data, $personID, "person");}, 
            function (err) {if($callback) $callback(err, $personID, "person");}
        );
    }


    //tmdb caller.
    //"f6c6c0a4bff37a87365e3099e7caf7ff"  (ANNE's key)
    tmdb = {
        "api_key": "enter-your-tmdb-key",
        "base_uri": "https://api.themoviedb.org/3",
        "images_uri": "https://image.tmdb.org/t/p",
        "timeout": 7000,
        call: function($url, $params, $success, $error){
            var params_str ="api_key=f6c6c0a4bff37a87365e3099e7caf7ff";
            for (var key in $params) {
                if ($params.hasOwnProperty(key)) {
                    params_str+="&"+key+"="+encodeURIComponent($params[key]);
                }
            }
            var xhr = new XMLHttpRequest();
            xhr.timeout = 7000;
            xhr.ontimeout = function () {
                console.log("Request timed out: " + $url +" "+ params_str);
                console.log("plz try again");
                if ($error) $error({success:false, error_msg:"request timed out, please try again"});
            };
            var url = "https://api.themoviedb.org/3" + $url + "?" + params_str;
            console.log("url:"+url);
            xhr.open("GET", url, true);
            xhr.setRequestHeader('Accept', 'application/json');
            xhr.responseType = "text";
            xhr.onreadystatechange = function () {
                if (this.readyState === 4) {
                    if (this.status === 200){
                        if (typeof $success == "function") {
                            $success(JSON.parse(this.response));
                        }	
                        else {
                            console.log("success, but no callback");

                            return;
                        }
                    }
                    else {
                        var errorstr = this.response.toString();
                        if (typeof $error == "function") {    
                            if (this.status==404) {
                                $error({success:false, error_msg:"bad internet connection to source data"});
                            }
                            else if (this.status==404) {
                                $error({success:false, error_msg:"bad internet connection to source data"});
                            }
                            else $error(errorstr);
                        }
                        else {
                            console.log("no callback, error in api call");
                            return;
                        }
                    }
                }
            };
            xhr.send();
        }
    }
    

    /**
     * 
     * @param {*} $url 
     * @param {*} $method 
     * @returns 
     */
    simpleExternalCaller = async($url, $method="GET")=> {
        return new Promise((resolve, reject) => {
            var request = new XMLHttpRequest();
            request.open($method, $url, true);
            request.onload = function () {
                var data = JSON.parse(this.response)
                if (request.status >= 200 && request.status < 400) {
                    return resolve(data);
                } 
                else {
                    return resolve("");
                }
            }
            request.onerror = function() {
                console.log("there has been an error in the send to url:"+$url);
            }
            request.send()
        });
    }
}

const CHASE = new anne_chase ();
CHASE.is_init(CHASE);